/* Admin.cc: Admin connection handler class. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Admin.h"
#include "Array.h"
#include "Connection.h"
#include "Spool.h"
#include "Config.h"

extern "C" {
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
}

static int compare_groups_sort(const void *p, const void *q);
static int compare_groups_search(const void *p, const void *q);

Admin::Admin(Spool *spool_arg, Connection *conn_arg, Config *config_arg)
{
    spool = spool_arg;
    conn = conn_arg;
    config = config_arg;
    running = 1;
}

Admin::~Admin()
{
    delete conn;
}

void Admin::run()
{
    String line;
    const char *s, *p, *q;

    if (!config->get_admin_access(conn->name(), conn->ip())) {
	conn->write("502 You have no permission to talk.  Goodbye.\r\n");
	return;
    }

    conn->write("200 Threaded news server admin interface ready.\r\n");
    while (running && conn->read_line(&line) >= 0) {

	/* Strip trailing whitespace from the line. */
	s = line.chars();
	p = s + line.length() - 1;
	while (p >= s && isspace(*p))
	    p--;
	line.truncate(p + 1 - s);

	/* Ignore blank lines. */
	if (line.length() == 0)
	    continue;

	/* Find command (s to p) and remaining text (q). */
	s = find_word(line.chars());
	p = word_end(s);
	q = find_word(p);

	/* Find command associated with the starting word. */
	if (handle_command(s, p - s, q) < 0)
	    conn->write("500 Command not recognized.\r\n");

	line.clear();
    }
}

int Admin::handle_command(const char *s, int len, const char *rest)
{
    static struct entry {
	char *name;
	void (Admin::*func)(const char *s);
    } tab[] = {
	{ "newgroup",	Admin::newgroup },
	{ "takegroups",	Admin::takegroups },
	{ "quit",	Admin::quit }
    };
    int i;

    for (i = 0; i < sizeof(tab) / sizeof(struct entry); i++) {
	if (strlen(tab[i].name) == len && strnccmp(tab[i].name, s, len) == 0) {
	    (this->*tab[i].func)(rest);
	    return 0;
	}
    }

    return -1;
}

void Admin::newgroup(const char *s)
{
    const char *p, *q, *r, *t, *colon;

    /* Get first two words and start of the third (s to p, q to r, t). */
    t = find_word(r = word_end(q = find_word(p = word_end(s))));

    /* Check for bad characters in the newsgroup name. */
    colon = strchr(s, ':');
    if (colon && colon < p) {
	conn->write("501 Bad character in newsgroup name.\r\n");
	return;
    }

    /* Create the group. */
    if (*q == 0) {
	q = "y";
	r = q + 1;
    }
    spool->create_group(s, p - s, q, r - q, t, strlen(t));
    conn->writef("200 Group %C created.\r\n", s, p - s);
}

void Admin::takegroups(const char *s)
{
    const char *p, *q, *r;
    FILE *fp = NULL;
    int fd, ngroups = 0, i, do_merge;
    String name, buf;
    Array<Remote_group, 9, 4> groups;
    Remote_group *group;
    time_t now;

    /* Get first two words (s to p, q to r). */
    r = word_end(q = find_word(p = word_end(s)));

    /* Must be one or two words. */
    if (!*s || *r) {
	conn->write("501 TAKEGROUPS needs one or two arguments.");
	return;
    }

    do_merge = (*q) ? 0 : 1;
    fp = (do_merge) ? tmpfile() : fopen(q, "w");
    if (!fp) {
	conn->writef("480 Can't open output file: %s.\r\n", strerror(errno));
	return;
    }

    name.append(s, p - s);
    fd = tcp_connect(name.chars(), "nntp", 119, &buf);
    if (fd < 0) {
	conn->writef("480 %S\r\n", &buf);
	fclose(fp);
	return;
    }

    /* Read the welcome line from the remote server. */
    Connection remote(fd);
    if (remote.read_line_clear(&buf) < 0 || *buf.chars() != '2') {
	conn->writef("480 Access refused to %S.\r\n", &name);
	fclose(fp);
	return;
    }

    /* Get a list of newsgroups using LIST. */
    remote.write("LIST\r\n");
    remote.read_line_clear(&buf);
    while (remote.read_line_clear(&buf) >= 0) {
	/* Stop if we find a . by itself on a line. */
	if (buf.length() == 1 && *buf.chars() == '.')
	    break;

	/* Find the group name (s to p) and the flags (q to r). */
	p = word_end(s = buf.chars());
	q = find_word(word_end(find_word(word_end(find_word(p)))));
	r = word_end(q);

	/* Verify that the line was four words and that the flags were okay. */
	if (*r) {
	    warn("TAKEGROUPS skipped invalid list line: %s", s);
	    continue;
	}
	if (check_flags(q, r - q) < 0) {
	    warn("TAKEGROUPS invalid flags \"%s\" changed to \"y\"", q);
	    q = "y";
	}

	/* Make a new groups entry. */
	groups.ensure(ngroups);
	groups[ngroups].name = strduplicate(s, p - s);
	groups[ngroups].flags = strduplicate(q, strlen(q));
	groups[ngroups].desc = strduplicate("");
	ngroups++;
    }

    /* Sort the groups. */
    qsort(groups.base(), ngroups, sizeof(Remote_group), compare_groups_sort);

    /* Use LIST NEWSGROUPS to get descriptions. */
    remote.write("LIST NEWSGROUPS\r\n");
    if (remote.read_line_clear(&buf) >= 0 && *buf.chars() == '2') {
	while (remote.read_line_clear(&buf) >= 0) {
	    /* Stop if we find a . by itself on a line. */
	    if (buf.length() == 1 && *buf.chars() == '.')
		break;

	    /* Look up the group name. */
	    p = word_end(s = buf.chars());
	    name.clear();
	    name.append(s, p - s);
	    group = (Remote_group *) bsearch(name.chars(), groups.base(),
					     ngroups, sizeof(Remote_group),
					     compare_groups_search);
	    if (!group)
		continue;

	    /* Set the description of the group. */
	    delete[] group->desc;
	    group->desc = strduplicate(find_word(p));
	}
    }

    remote.write("QUIT\r\n");
    remote.read_line_clear(&buf);

    /* Write out the resulting newsgroup lines. */
    time(&now);
    for (i = 0; i < ngroups; i++) {
	fprintf(fp, "%s %ld %s %s\n", groups[i].name, (long) now,
		groups[i].flags, groups[i].desc);
	delete[] groups[i].name;
	delete[] groups[i].flags;
	delete[] groups[i].desc;
    }

    /* If this is a merge, do it now. */
    if (do_merge) {
	rewind(fp);
	spool->merge_groups(fp);
    }

    fclose(fp);
    conn->writef("200 TAKEGROUPS processed %d groups.\r\n", ngroups);
}

void Admin::quit(const char *s)
{
    conn->write("205 Closing connection.\r\n");
    running = 0;
}

static int compare_groups_sort(const void *p, const void *q)
{
    return strcmp(((Remote_group *) p)->name, ((Remote_group *) q)->name);
}

static int compare_groups_search(const void *p, const void *q)
{
    return strcmp((char *) p, ((Remote_group *) q)->name);
}

