/* Bucket.cc: An article bucket. */

#include "threads.h"
#include "error.h"
#include "util.h"
#include "platform.h"

#include "Bucket.h"
#include "Article.h"
#include "Spool.h"
#include "Headers.h"

extern "C" {
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>

/* Non-standard functions not generally prototyped.  Declare them without
 * a prototype, to avoid breaking on existing prototypes. */
int ftruncate();
#ifdef HAVE_FSYNC
int fsync();
#else
#define fsync(fd)
#endif
}

Bucket::Bucket(FILE *file, const char *name, int partition, time_t number)
{
    fp = file;
    filename.append(name);
    part = partition;
    bucket_num = number;
    list = NULL;
    fseek(fp, 0, SEEK_END);
}

Bucket::~Bucket()
{
    fclose(fp);
}

Article *Bucket::get_article(long offset)
{
    FILE *copy;
    Article_list *entry;
    int c;

    /* Open a file pointing to the bucket. */
    copy = fopen(filename.chars(), "r");
    if (!copy) {
	error("Can't re-open bucket file: %m");
	return NULL;
    }
    if (fseek(copy, offset, SEEK_SET) < 0) {
	error("Can't seek to location %l in bucket: %m", offset);
	return NULL;
    }

    /* Read in one character, to make sure there is an article here. */
    c = getc(copy);
    if (c == EOF) {
	fclose(copy);
	return NULL;
    }
    ungetc(c, copy);

    /* Make a new article, chain it into the list, and return it. */
    entry = new Article_list(this, copy, offset);
    lock();
    entry->next = list;
    list = entry;
    unlock();
    return &entry->article;
}

void Bucket::release_article(Article *article)
{
    Article_list **entry_ptr, *entry;

    lock();

    /* Find the article in the list. */
    for (entry_ptr = &list; *entry_ptr; entry_ptr = &(*entry_ptr)->next) {
	entry = *entry_ptr;
	if (&entry->article == article)
	    break;
    }
    if (!*entry_ptr) {
	crash("Released article <%d, %l, %d> is not in the article list.",
	      part, (long) bucket_num, article->offset());
    }

    entry->refs--;
    if (entry->refs == 0) {
	*entry_ptr = entry->next;
	unlock();
	delete entry;
    } else {
	unlock();
    }
}

int Bucket::find_header(long offset, const char *name, int len, String *buf)
{
    String line;
    FILE *copy;
    const char *s;

    copy = fopen(filename.chars(), "r");
    if (!copy) {
	error("Can't re-open bucket file: %m");
	return -1;
    }
    if (fseek(copy, offset, SEEK_SET) < 0) {
	error("Can't seek to location %l in bucket <%d, %l>: %m",
	      offset, part, (long) bucket_num);
	return -1;
    }

    while (read_line_clear(&line, copy) >= 0) {
	s = line.chars();
	if (!*s || (*s == '.' && !s[1])) {
	    fclose(copy);
	    return -1;
	}
	if (strnccmp(s, name, len) == 0 && *find_word(s + len) == ':') {
	    /* We found a match.  Look for continuations. */
	    buf->append(find_word(find_word(s + len) + 1));
	    while (read_line_clear(&line, copy) >= 0) {
		if (!isspace(*line.chars()))
		    break;
		buf->append(' ');
		buf->append(&line);
	    }
	    fclose(copy);
	    return 0;
	}
    }

    fclose(copy);
    error("Corrupt article in bucket %l at offset %l.", bucket_num, offset);
    return -1;
}

long Bucket::write_article(Headers *headers, const char *body, int len)
{
    const char *p;
    long offset;

    offset = ftell(fp);
    for (p = headers->first_header(); p; p = headers->next_header()) {
	fwrite(p, sizeof(char), strlen(p), fp);
	putc('\n', fp);
    }
    putc('\n', fp);
    fwrite(body, sizeof(char), len, fp);
    fputs(".\n", fp);
    fflush(fp);
    fsync(fileno(fp));
    return offset;
}

long Bucket::current_offset() const
{
    return ftell(fp);
}

int Bucket::consistency_check(Spool *spool, long offset)
{
    Headers headers;
    long new_offset;
    FILE *copy;
    String line;
    int num_lines;

    /* Open a file pointing to the bucket. */
    copy = fopen(filename.chars(), "r");
    if (!copy) {
	error("Consistency check: can't re-open bucket file: %m");
	return -1;
    }
    if (fseek(copy, offset, SEEK_SET) < 0) {
	error("Consistency check: can't seek to %l: %m", offset);
	return -1;
    }

    /* Re-index untilwe find an incomplete article or the end. */
    while (1) {
	/* Get the headers. */
	num_lines = headers.add_headers(copy);
	if (num_lines < 0)
	    break;

	/* Search for the terminating line. */
	while (read_line_clear(&line, copy) == 0) {
	    num_lines++;
	    if (line.length() == 1 && *line.chars() == '.') {
		new_offset = ftell(copy);
		if (spool->reindex_article(num_lines, new_offset - offset,
					   part, bucket_num, offset,
					   &headers) < 0) {
		    fclose(copy);
		    return -1;
		}
		offset = ftell(copy);
		break;
	    }
	}

	headers.clear();
    }

    /* Get rid of any partial article at the end. */
    fclose(copy);
    ftruncate(fileno(fp), (off_t) offset);

    return 0;
}

Article_list::Article_list(Bucket *bucket, FILE *fp, long offset)
: article(bucket, fp, offset)
{
    refs = 1;
}

