/* Config.cc: The configuration file. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Config.h"
#include "String.h"
#include "Wildcard_list.h"
#include "Newsfeed.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
int gethostname();
}

static void delete_chars(char *s);

Config::Config(const char *config_filename)
: moderators(delete_chars)
{
    char buf[128];

    /* Set reasonable defaults. */
    readpost = 0;
    if (gethostname(buf, sizeof(buf)) == 0)
	server_name.append(buf);
    add_mime_headers = 0;
    path_newsgroups_file.append(ETCDIR "/newsgroups");
    path_index_dir.append("/var/spool/news/indices/");
    path_groups_dir.append("/var/spool/news/groups/");
    path_pool_dir.append("/var/spool/news/pool/");
    path_feeds_dir.append("/var/spool/news/feeds/");
    schema = stringlist_empty();
    expire_incr = 86400;
    expire_keep_for = 14;
    cancel_keep_for = 14;
    default_expire_time.min = 1;
    default_expire_time.def = 10;
    default_expire_time.max = 28;
    default_moderator = strduplicate("%s@uunet.uu.net");
    feeds_bidirectional = 1;
    feed_sites = stringlist_empty();

    /* Read the config file. */
    read_config_file(config_filename);
}

Config::~Config()
{
    stringlist_free(schema);
    stringlist_free(feed_sites);
    delete[] default_moderator;
}

void Config::get_server_name(String *name) const
{
    name->append(&server_name);
}

int Config::any_distribution_accepted(const char *dist) const
{
    const char *p;
    String tmp;

    /* If no distributions were specified, assume we don't limit. */
    if (dists.num_entries() == 0)
	return 1;

    /* Check list of distributions against our table. */
    p = strchr(dist, ',');
    while (p) {
	tmp.clear();
	tmp.append(dist, p - dist);
	if (dists.match(tmp.chars(), 0) == 1)
	    return 1;
	p = strchr(dist = p + 1, ',');
    }
    return dists.match(dist, 0);
}

void Config::get_default_organization(String *buf) const
{
    buf->append(&organization);
}

bool Config::get_add_mime_headers_flag() const
{
    return add_mime_headers;
}

void Config::get_newsgroups_file(String *filename) const
{
    filename->append(&path_newsgroups_file);
}

void Config::get_index_directory(String *dirname) const
{
    dirname->append(&path_index_dir);
}

void Config::get_groups_directory(String *dirname) const
{
    dirname->append(&path_groups_dir);
}

void Config::get_pool_directory(String *dirname) const
{
    dirname->append(&path_pool_dir);
}

void Config::get_feeds_directory(String *dirname) const
{
    dirname->append(&path_feeds_dir);
}

time_t Config::decide_bucket(const char *newsgroups, time_t expires_header)
    const
{
    String name;
    Expire_spec spec;
    time_t result, t, now;
    const char *p = newsgroups;

    /* Check each newsgroup, and select the maximum result. */
    time(&now);
    result = constrain_time(expires_header, now, &default_expire_time);
    while (*p) {
	p = get_group_name(&name, p);
	spec = expire_times.match(name.chars(), default_expire_time);
	t = constrain_time(expires_header, now, &spec);
	if (t > result)
	    result = t;
    }
    return result;
}

int Config::get_expire_keep_for() const
{
    return expire_keep_for * expire_incr;
}

int Config::get_cancel_keep_for() const
{
    return cancel_keep_for * expire_incr;
}

int Config::find_in_schema(const char *name, int namelen) const
{
    char **p;

    for (p = schema; *p; p++) {
	if (strlen(*p) == namelen && strnccmp(name, *p, namelen) == 0)
	    return p - schema;
    }
    return -1;
}

const char *const *Config::get_schema() const
{
    return schema;
}

const char *const *Config::get_feed_sites() const
{
    return feed_sites;
}

Wildcard_list<bool> *Config::get_feed_groups(const char *name) const
{
    const char *const strings[] = { name, NULL };

    return feed_groups.sublist(strings);
}

Wildcard_list<bool> *Config::get_feed_dists(const char *name) const
{
    const char *const strings[] = { name, NULL };

    return feed_dists.sublist(strings);
}

Wildcard_list<bool> *Config::get_feed_paths(const char *name) const
{
    const char *const strings[] = { name, NULL };

    return feed_paths.sublist(strings);
}

void Config::get_moderator(const char *group, String *buf) const
{
    String mod, dash_group;

    dash_group.append(group);
    dash_group.substitute(".", "-");
    mod.append(moderators.match(group, default_moderator));
    mod.substitute("%s", dash_group.chars());
    buf->append(&mod);
}

bool Config::get_newgroup_allow(const char *group, int grouplen,
				const char *from) const
{
    String tmp(group, grouplen);
    const char *const strings[] = { tmp.chars(), from, NULL };

    return newgroup_allow.match(strings, 0);
}

Wildcard_list<bool> *Config::get_post_access(const char *name,
					     const char *ip) const
{
    const char *const strings[] = { name, ip, NULL };

    if (readpost)
	return read_access.sublist(strings);
    else
	return post_access.sublist(strings);
}

Wildcard_list<bool> *Config::get_read_access(const char *name,
					     const char *ip) const
{
    const char *const strings[] = { name, ip, NULL };

    return read_access.sublist(strings);
}

Wildcard_list<bool> *Config::get_transfer_access(const char *name,
						 const char *ip) const
{
    const char *const strings[] = { name, ip, NULL };

    return transfer_access.sublist(strings);
}

bool Config::get_admin_access(const char *name, const char *ip) const
{
    const char *const strings[] = { name, ip, NULL };

    return admin_access.match(strings, 0);
}

void Config::read_config_file(const char *filename)
{
    static struct {
	char *name;
	void (Config::*func)(const char *s);
    } sections[] = {
	{ "access",		Config::section_access },
	{ "expire",		Config::section_expire },
	{ "feeds",		Config::section_feeds },
	{ "general",		Config::section_general },
	{ "moderators",		Config::section_moderators },
	{ "newgroup",		Config::section_newgroup },
	{ "paths",		Config::section_paths },
	{ "schema",		Config::section_schema },
    };
    const int NUM_SECTIONS = (sizeof(sections) / sizeof(*sections));

    FILE *fp;
    String line;
    int section = -1, i;
    const char *p, *q;

    /* Open the config file; if there is none, just run with the defaults. */
    fp = fopen(filename, "r");
    if (!fp) {
	warn("Can't read config file %s", filename);
	return;
    }

    while (read_line_clear(&line, fp) == 0) {
	line.strip_trailing_whitespace();

	/* Find first non-whitespace character. */
	p = line.chars();
	for (q = p; isspace(*q); q++);

	/* If it's not a blank line or a comment, process it. */
	if (*q && *q != '#') {

	    if (q == p) {
		/* There's no leading whitespace; it's a new section header. */
		for (i = 0; i < NUM_SECTIONS; i++) {
		    if (strccmp(sections[i].name, p) == 0)
			break;
		}
		if (i < NUM_SECTIONS)
		    section = i;
		else
		    warn("Invalid section header %s in config file.", p);
	    } else {
		/* There's leading whitespace; call the section function. */
		if (section != -1)
		    (this->*sections[section].func)(q);
		else
		    warn("No initial section header in config file.");
	    }
	}
    }

    fclose(fp);
}

/* Commands:	read=post
 * 		[dis]allow <host-pattern> admin
 *		[dis]allow <host-pattern> post <newsgroup-pattern> ...
 *		[dis]allow <host-pattern> read <newsgroup-pattern> ...
 *		[dis]allow <host-pattern> transfer <newsgroup-pattern> ... */
void Config::section_access(const char *s)
{
    const char *p, *q, *r, *t, *u;
    bool whether;
    Wildcard_list<bool> *list;
    String pattern;

    /* Get first three words (s to p, q to r, t to u). */
    u = word_end(t = find_word(r = word_end(q = find_word(p = word_end(s)))));

    /* Check if one word, "read=post". */
    if (p - s == 9 && strnccmp(s, "read=post", 9) == 0 && *p == 0) {
	readpost = 1;
	return;
    }

    /* At least three words, first word is "allow" or "disallow". */
    if (p - s == 5 && strnccmp(s, "allow", 5) == 0 && *t != 0) {
	whether = 1;
    } else if (p - s == 8 && strnccmp(s, "disallow", 8) == 0 && *t != 0) {
	whether = 0;
    } else {
	warn("Invalid reader-access line in config file: %s", s);
	return;
    }

    if (u - t == 5 && strnccmp(t, "admin", 5) == 0 && *u == 0) {
	String host(q, r - q);
	const char *const as_name[] = { host.chars(), "*", NULL };
	const char *const as_ip[] = { "*", host.chars(), NULL };

	/* Add the host pattern as a name and as an IP address. */
	admin_access.add_pattern(as_name, whether);
	admin_access.add_pattern(as_ip, whether);
	return;
    } else if (u - t == 4 && strnccmp(t, "post", 4) == 0) {
	list = &post_access;
    } else if (u - t == 4 && strnccmp(t, "read", 4) == 0) {
	list = &read_access;
    } else if (u - t == 8 && strnccmp(t, "transfer", 8) == 0) {
	list = &transfer_access;
    } else {
	warn("Invalid reader-access line in config file: %s", s);
	return;
    }

    /* Add an entry for each word starting at the fourth. */
    String host(q, r - q);
    for (u = word_end(t = find_word(u)); *t; u = word_end(t = find_word(u))) {
	String group(t, u - t);
	const char *const as_name[] = { host.chars(), "*", group.chars(),
					NULL };
	const char *const as_ip[] = { "*", host.chars(), group.chars(), NULL };

	/* Allow host pattern to be either a name or an IP address. */
	list->add_pattern(as_name, whether);
	list->add_pattern(as_ip, whether);
    }
}

/* Commands:	<pattern> <min> <default> <max>
 *		default <min> <default> <max>
 *		increment <seconds>
 *		keep cancel <incrs>
 *		keep expire <incrs> */
void Config::section_expire(const char *s)
{
    const char *p, *q, *r, *t, *u, *v, *w;
    Expire_spec spec;

    /* Get four words (s to p, q to r, t to u, v to w). */
    u = word_end(t = find_word(r = word_end(q = find_word(p = word_end(s)))));
    w = word_end(v = find_word(u));

    /* Check if two words "increment <seconds>". */
    if (p - s == 9 && strnccmp(s, "increment", 9) == 0 && *q != 0 && *r == 0) {
	expire_incr = atoi(q);
	return;
    }

    /* Check if three words beginning with "keep". */
    if (p - s == 4 && strnccmp(s, "keep", 4) == 0 && *t != 0 && *u == 0) {
	if (r - q == 6 && strnccmp(q, "cancel", 6) == 0)
	    cancel_keep_for = atoi(t);
	else if (r - q == 6 && strnccmp(q, "expire", 6) == 0)
	    expire_keep_for = atoi(t);
	else
	    warn("Invalid expires line in config file: %s", s);
	return;
    }

    /* Otherwise, line must be four words. */
    if (*v == 0 || *w != 0) {
	warn("Invalid expires line in config file: %s", s);
	return;
    }

    /* Set the default specification or add one to the wildcard list. */
    spec.min = atoi(q);
    spec.def = atoi(t);
    spec.max = atoi(v);
    if (p - s == 7 && strnccmp(s, "default", 7) == 0) {
	default_expire_time = spec;
    } else {
	String pattern(s, p - s);
	expire_times.add_pattern(pattern.chars(), spec);
    }
}

/* Commands:	outgoing
 *		bidirectional
 *		site <hostname>
 *		[don't|only] feed <site-pattern> group <group-pattern> ...
 *		[don't|only] feed <site-pattern> dist <dist-pattern> ...
 *		[don't|only] feed <site-pattern> path <path-pattern> ...
 */
void Config::section_feeds(const char *s)
{
    const char *p, *q, *r, *t, *u, *line = s;
    bool whether = 1, only = 0;
    Wildcard_list<bool> *list;

    /* Get first word, check for "don't" or "only" followed by "feed". */
    p = word_end(s);
    if (p - s == 5 && strnccmp(s, "don't", 5) == 0) {
	whether = 0;
	s = find_word(p);
    } else if (p - s == 4 && strnccmp(s, "only", 4) == 0) {
	only = 1;
	s = find_word(p);
    }
    if ((only || !whether) && strnccmp(s, "feed", 4) != 0) {
	warn("Invalid feeds line: %s", line);
	return;
    }

    /* Get first three words (s to p, q to r, t to u). */
    u = word_end(t = find_word(r = word_end(q = find_word(p = word_end(s)))));

    /* Check if "outgoing" or "bidirectional". */
    if (p - s == 8 && strnccmp(s, "outgoing", 8) == 0 && !*p)
	feeds_bidirectional = 0;
    else if (p - s == 13 && strnccmp(s, "bidirectional", 13) == 0 && !*p)
	feeds_bidirectional = 1;

    /* Check if two words "site <hostname>". */
    if (p - s == 4 && strnccmp(s, "site", 4) == 0 && *q && !*r) {

	if (stringlist_contains(feed_sites, q, r - q)) {
	    warn("Duplicate feed site \"%s\"", q);
	} else {
	    const char *const patterns[] = { q, q, NULL };

	    /* Add a site to the feed sites list and add a pattern to
	     * the feed paths to exlude a path component equal to the machine
	     * name. */
	    feed_sites = stringlist_append(feed_sites, q, r - q);
	    feed_paths.add_pattern(patterns, 0);
	}

	/* If the user hasn't set "outgoing", add this host to the transfer
	 * access list as both a hostname and an IP address. */
	if (feeds_bidirectional) {
	    const char *const as_name[] = { q, "*", "*", NULL };
	    const char *const as_ip[] = { "*", q, "*", NULL };

	    transfer_access.add_pattern(as_name, 1);
	    transfer_access.add_pattern(as_ip, 1);
	}
	return;
    }

    if (p - s != 4 || strnccmp(s, "feed", 4) != 0) {
	warn("Invalid feeds line: %s", line);
	return;
    }

    if (u - t == 5 && strnccmp(t, "group", 5) == 0) {
	list = &feed_groups;
    } else if (u - t == 4 && strnccmp(t, "dist", 4) == 0) {
	list = &feed_dists;
    } else if (u - t == 4 && strnccmp(t, "path", 4) == 0) {
	list = &feed_paths;
    } else {
	warn("Invalid feed line: %s", line);
	return;
    }

    /* Add an entry for each group (each word starting from the fourth). */
    String site(q, r - q);
    for (u = word_end(t = find_word(u)); *t; u = word_end(t = find_word(u))) {
	String group(t, u - t);
	const char *const patterns[] = { site.chars(), group.chars(), NULL };
	list->add_pattern(patterns, whether);
    }

    if (only) {
	/* Go back and add a negative entry for site "*" for each group. */
	for (u = word_end(t = find_word(u)); *t;
	     u = word_end(t = find_word(u))) {
	    String group(t, u - t);
	    const char *const patterns[] = { "*", group.chars(), NULL };
	    list->add_pattern(patterns, 0);
	}
    }
}

/* Commands:	name <server name>
 *		mime-headers {yes|no}
 *		distributions <pattern> ...
 *		organization <organization>
 */
void Config::section_general(const char *s)
{
    const char *p, *q, *r;
    String tmp;

    /* Get first two words (s to p, q to r). */
    r = word_end(q = find_word(p = word_end(s)));

    if (p - s == 4 && strnccmp(s, "Name", 4) == 0 && *q != 0 && *r == 0) {
	/* "Name" followed by one word; set the server name. */
	server_name.clear();
	server_name.append(q, r - q);
    } else if (p - s == 12 && strnccmp(s, "mime-headers", 12) == 0
	       && *q != 0 && *r == 0) {
	add_mime_headers = (strccmp(q, "y") == 0 || strccmp(q, "yes") == 0);
    } else if (p - s == 13 && strnccmp(s, "distributions", 13) == 0
	       && *q != 0) {
	while (*q) {
	    tmp.clear();
	    tmp.append(q, r - q);
	    dists.add_pattern(tmp.chars(), 1);
	    r = word_end(q = find_word(r));
	}
    } else if (p - s == 12 && strnccmp(s, "organization", 12) == 0
	       && *q != 0 && *r == 0) {
	organization.clear();
	organization.append(q, r - q);
    } else {
	warn("Unknown general command in config file: %s", s);
    }
}

/* Commands:	<pattern> <moderator>
 *		default <moderator> */
void Config::section_moderators(const char *s)
{
    const char *p, *q, *r;
    String tmp;

    /* Get first two words (s to p, q to r). */
    r = word_end(q = find_word(p = word_end(s)));

    /* Must be two words. */
    if (*q == 0 || *r != 0) {
	warn("Invalid moderators line in config file: %s", s);
	return;
    }

    if (p - s == 7 && strnccmp(s, "default", 7) == 0) {
	delete[] default_moderator;
	default_moderator = strduplicate(q, r - q);
    } else {
	tmp.append(s, p - s);
	moderators.add_pattern(tmp.chars(), strduplicate(q, r - q));
    }
}

/* Commands:	allow <email-pattern> <group-pattern> ...
 *		disallow <email-pattern> <group-pattern> ... */
void Config::section_newgroup(const char *s)
{
    const char *p, *q, *r, *t, *u;
    bool whether;
    String pattern;

    /* Get first three words (s to p, q to r, t to u). */
    u = word_end(t = find_word(r = word_end(q = find_word(p = word_end(s)))));

    /* Figure out whether we want to allow or disallow. */
    if (p - s == 5 && strnccmp(s, "allow", 5) == 0 && *r != 0) {
	whether = 1;
    } else if (p - s == 8 && strnccmp(s, "disallow", 8) == 0 && *r != 0) {
	whether = 0;
    } else {
	warn("Invalid newgroup line in config file: %s", s);
	return;
    }

    /* Add an entry for each word starting at the third. */
    String email(q, r - q);
    for (; *t != 0; u = word_end(t = find_word(u))) {
	String group(t, u - t);
	const char *const patterns[] = { email.chars(), group.chars(), NULL };

	newgroup_allow.add_pattern(patterns, whether);
    }
}

/* Commands:	index <pathname>
 *		groups <pathname>
 *		pool <pathname>
 *		grouplist <pathname> */
void Config::section_paths(const char *s)
{
    struct {
	char *name;
	String *str;
	int dir_flag;
    } paths[] = {
	{ "newsgroups",	&path_newsgroups_file,	0 },
	{ "index",	&path_index_dir,	1 },
	{ "groups",	&path_groups_dir,	1 },
	{ "pool",	&path_pool_dir,		1 },
	{ "feeds",	&path_feeds_dir,	1 }
    };
    const int NUM_PATHS = (sizeof(paths) / sizeof(*paths));

    String *str;
    const char *p, *q, *r, *name;
    int i;

    /* The line must be two words (s to p, q to r). */
    p = word_end(s);
    q = find_word(p);
    r = word_end(q);
    if (*q == 0 || *r != 0) {
	warn("Invalid paths line in config file: %s", s);
	return;
    }

    for (i = 0; i < NUM_PATHS; i++) {
	name = paths[i].name;
	if (p - s == strlen(name) && strnccmp(name, s, p - s) == 0)
	    break;
    }
    if (i < NUM_PATHS) {
	str = paths[i].str;
	str->clear();
	str->append(q, r - q);
	if (paths[i].dir_flag && str->chars()[str->length() - 1] != '/')
	    str->append('/');
    } else {
	warn("Unknown path in config file: %s", s);
    }
}

/* Commands:	<header>
 *		<header> full */
void Config::section_schema(const char *s)
{
    const char *p, *q, *r;
    char *entry;
    int full = 0;

    /* Get first two words (s to p, q to r). */
    r = word_end(q = find_word(p = word_end(s)));

    /* See if we want a full header. */
    if (*r == 0 && strccmp(q, "full") == 0) {
	full = 1;
    } else if (*p != 0) {
	warn("Unknown schema command in config file: %s", s);
	return;
    }

    /* Add the new schema entry to the existing schemas.  If this is a
     * full entry, put a colon (':') at the beginning. */
    schema = stringlist_append(schema, s, p - s + full);
    if (full) {
	entry = schema[stringlist_length(schema) - 1];
	memmove(entry + 1, entry, p - s);
	*entry = ':';
    }
}

time_t Config::constrain_time(time_t t, time_t now, const Expire_spec *spec)
    const
{
    time_t min, def, max;

    min = ((now / expire_incr) + spec->min) * expire_incr;
    def = ((now / expire_incr) + spec->def) * expire_incr;
    max = ((now / expire_incr) + spec->max) * expire_incr;
    return (t == -1) ? def : (t < min) ? min : (t > max) ? max : t;
}

static void delete_chars(char *s)
{
    delete[] s;
}

int Expire_spec::operator ==(const Expire_spec &spec)
{
    return (min == spec.min && def == spec.def && max == spec.max);
}

