/* connection.cc: Handle client connections. */

#include "threads.h"
#include "error.h"

#include "Connection.h"
#include "String.h"
#include "util.h"

extern "C" {
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>

int getpeername();
}

const int OUTBUF_MAX_LEN = 4000;

Connection::Connection(int fd)
{
    struct hostent *hostent, hostbuf;
    char buf[4096];
    struct sockaddr_in peer_addr;
    int len = sizeof(peer_addr), herr;

    client_socket = fd;
    flags = fcntl(fd, F_GETFL);
    if (getpeername(fd, (struct sockaddr *) &peer_addr, &len) >= 0) {
	inet_ntoa_r(peer_addr.sin_addr, peer_ip, sizeof(peer_ip));
	hostent = gethostbyaddr_r((char *) &peer_addr.sin_addr,
				  sizeof(struct in_addr), AF_INET, &hostbuf,
				  buf, sizeof(buf), &herr);
	if (hostent) {
	    peer_name = strduplicate(hostent->h_name);
	    hostent = gethostbyname_r(peer_name, &hostbuf, buf, sizeof(buf),
				      &herr);
	    if (hostent && memcmp(&hostent->h_addr, &peer_addr, len) == 0)
		peer_name_status = NAME_OK;
	    else
		peer_name_status = NAME_FORGED;
	} else {
	    peer_name = strduplicate(peer_ip);
	    peer_name_status = NAME_NONE;
	}
    } else {
	peer_name_status = NAME_NOTSOCKET;
	peer_name = strduplicate("<not socket>");
	strcpy(peer_ip, "<not socket>");
    }
    /* Temporary. */
    log("Connection received from %s (%s).", peer_name, peer_ip);
}

Connection::~Connection()
{
    flush_outbuf();
    close(client_socket);
    delete[] peer_name;
}

int Connection::read_line(String *line)
{
    const char *newline, *s, *p, *q;
    char buf[4000];
    int count;

    while (1) {
	/* Scan inbuf for newlines. */
	s = inbuf.chars();
	newline = strchr(s, '\n');
	if (newline)
	    break;

	/* Read more text into inbuf and try again.  First, flush the output
	 * buffer. */
	flush_outbuf();
	count = read(buf, sizeof(buf));
	if (count <= 0)
	    return -1;
	inbuf.append(buf, count);
    }

    /* Copy from s to newline into line, stripping non-printable characters. */
    p = s;
    while (p < newline) {
	while (p < newline && !(isprint(*p) || *p == '\t' || *p == '\f'))
	    p++;
	q = p;
	while (q < newline && (isprint(*q) || *q == '\t' || *q == '\f'))
	    q++;
	line->append(p, q - p);
	p = q + 1;
    }

    inbuf.shift(newline + 1 - s);
    return 0;
}

int Connection::read_line_clear(String *line)
{
    line->clear();
    return read_line(line);
}

void Connection::write(const char *s, int len)
{
    int space_left;

    /* Make sure the buffer length doesn't go above BUFFER_MAX_LEN. */
    while (outbuf.length() + len > OUTBUF_MAX_LEN) {
	space_left = OUTBUF_MAX_LEN - outbuf.length();
	outbuf.append(s, space_left);
	flush_outbuf();
	s += space_left;
	len -= space_left;
    }

    outbuf.append(s, len);
}

void Connection::writef(const char *fmt, ...)
{
    va_list args;
    String str;

    va_start(args, fmt);
    vformat(&str, fmt, args);
    va_end(args);
    write(&str);
}

void Connection::flush_outbuf()
{
    int count;
    const char *s = outbuf.chars();
    int len = outbuf.length();

    while (len) {
	count = ::write(client_socket, s, len);
	if (count < 0)
	    break;
	s += count;
	len -= count;
    }

    outbuf.clear();
}

int Connection::read(char *buf, int buf_size)
{
    int count;

    /* Set the file descriptor non-blocking to allow fast streaming. */
    if (outbuf.length()) {
	fcntl(client_socket, F_SETFL, flags & O_NONBLOCK);
	count = ::read(client_socket, buf, buf_size);
	fcntl(client_socket, F_SETFL, flags);
	if (count != -1 || errno != EAGAIN)
	    return count;
	flush_outbuf();
    }
    return ::read(client_socket, buf, buf_size);
}

