/* Groups.cc: Managing the hierarchy of group files. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Groups.h"
#include "Atomic_file.h"
#include "Newsgroup.h"
#include "String.h"
#include "Config.h"
#include "Connection.h"
#include "Headers.h"
#include "Wildcard_list.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
}

/* Locking policy: the hash table and all of the ->next and ->in_file fields
 * of the Newsgroup_list structures are under control of the quiescing
 * read/write lock on the Groups object.  All other fields on the
 * Newsgroup_list objects are under control of the lock on that structure.
 * To write the bounds file or newsgroups file, you need a lock on update_lock
 * as well as a read lock on the Groups object, OR a write lock on the Groups
 * object.  (This prevents deadlock while ensuring that only one thread tries
 * to update files at a time.) */

const int MIN_POWER = 9;
const int DELTA = 2;
const int AVERAGE_LINE_LENGTH = 60;

static int check_file(FILE *fp, int partition, time_t bucket_num,
		      long offset, Newsgroup_list *entry);
static int eliminate_bucket_in_file(time_t bucket_num, const char *name,
				    Newsgroup_list *entry);
static int match_distribution(const char *name, const char *dist);
static FILE *open_group_file(const char *name, int check);
static void delete_group_file(const char *name, int dirnamelen);
static void *periodic_check_newsgroups(void *arg);

Groups::Groups(const Config *config_arg)
{
    String bounds_filename;
    int i;
    FILE *fp;

    config = config_arg;
    config->get_groups_directory(&dirname);
    config->get_newsgroups_file(&newsgroups_filename);
    newsgroups_file = new Atomic_file(newsgroups_filename.chars());
    format(&bounds_filename, "%S/bounds.list", &dirname);
    bounds_file = new Atomic_file(bounds_filename.chars());

    make_directories_in_path(dirname.chars());
    make_directories_in_path(bounds_filename.chars());

    size = adjust_size(0, MIN_POWER, DELTA);
    hashtab = new Newsgroup_list *[size];
    for (i = 0; i < size; i++)
	hashtab[i] = NULL;
    num_groups = 0;

    fp = newsgroups_file->open_read();
    if (fp) {
	read_newsgroups_file(fp);
	fclose(fp);
    } else {
	die("Can't read newsgroups file.");
    }
    read_bounds_file();

    start_thread(periodic_check_newsgroups, this);
}

Groups::~Groups()
{
    int i;
    Newsgroup_list *entry, *next;

    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = next) {
	    next = entry->next;
	    delete[] entry->name;
	    delete[] entry->flags;
	    delete[] entry->desc;
	    if (entry->newsgroup)
		delete entry->newsgroup;
	    delete entry;
	}
    }
}

Newsgroup *Groups::find_group(const char *name)
{
    FILE *fp;
    Newsgroup_list *entry;
    Newsgroup *group;
    String filename;

    /* Find the entry for the group in the hash table. */
    entry = follow_aliases(name);
    if (!entry)
	return NULL;

    /* If there's already a newsgroup object for this entry, bump up its
     * reference count and return it. */
    if (entry->newsgroup) {
	group = entry->newsgroup;
	entry->refs++;
	entry->unlock();
	return group;
    }

    /* We have to create the newsgroup object.  Construct its name. */
    construct_group_name_clear(&filename, name);

    /* Open the newsgroup file.  If it's not there, the group is empty. */
    fp = fopen(filename.chars(), "r");

    /* Create a new newsgroup, note it here, and return it. */
    group = new Newsgroup(name, fp, config, entry->last);
    entry->newsgroup = group;
    entry->refs = 1;
    entry->unlock();
    return group;
}

void Groups::release_group(Newsgroup *group)
{
    Newsgroup_list *entry;

    /* Find the newsgroup in the hash table. */
    entry = follow_aliases(group->name());
    if (!entry)
	crash("Released newsgroup %s not found.", group->name());

    entry->refs--;
    if (entry->refs == 0) {
	entry->newsgroup = NULL;
	delete group;
    }
    entry->unlock();
}

void Groups::reserve_positions(const char *groups,
			       const Wildcard_list<bool> *access, String *xref)
{
    String group_name, xref_buf;
    Newsgroup_list *entry;

    while (*groups) {

	/* Get the next group which accepts articles. */
	groups = get_group_name(&group_name, groups);
	if (access && !access->match(group_name.chars(), 0))
	    continue;
	entry = follow_aliases(group_name.chars());
	if (!entry)
	    continue;
	if (*entry->flags == 'x' || *entry->flags == 'j' ||
	    (access && !access->match(entry->name, 0))) {
	    entry->unlock();
	    continue;
	}

	/* Check that it's not a duplicate group. */
	group_name.append(':');
	if (strcstr(xref_buf.chars(), group_name.chars()) == NULL) {
	    /* Reserve an article number in the group and add an Xref entry. */
	    if (entry->first_unreserved <= entry->last)
		entry->first_unreserved = entry->last + 1;
	    format(&xref_buf, " %S%d", &group_name, entry->first_unreserved++);
	}
	entry->unlock();
    }

    xref->append(&xref_buf);
}

/* check is set to 1 if we're refiling an article, 0 if we're adding one.
 * Refiling takes into account various possible inconsistencies in the
 * grouplist file and is much slower. */
int Groups::add_to_groups(const Headers *headers, int num_lines, int len,
			  int partition, time_t bucket_num, long offset,
			  int check)
{
    const char *p, *q, *const *schema, *field;
    FILE *fp;
    Newsgroup_list *entry;
    String group_name, over_buf, index_buf, filename;
    int index;

    /* Build group line. */
    for (schema = config->get_schema(); *schema; schema++) {
	field = *schema;
	if (*field == ':') {
	    field++;
	    over_buf.append(field);
	    over_buf.append(": ", 2);
	}
	if (strccmp(field, "Lines") == 0) {
	    format(&over_buf, "%d", num_lines);
	} else if (strccmp(field, "Bytes") == 0) {
	    format(&over_buf, "%d", len);
	} else {
	    p = (*headers)[field];
	    if (p)
		over_buf.append(p);
	}
	over_buf.append('\001');
    }
    over_buf.substitute("\n\t\001", "  \t");

    p = strchr((*headers)["Xref"], ' ');
    while (p) {

	/* Get a group name and find the hash table entry for it.  If it
	 * is not in our list, then it might have been rmgrouped, so just
	 * ignore it. */
	q = strchr(p + 1, ':');
	group_name.clear();
	group_name.append(p + 1, q - (p + 1));
	index = atoi(q + 1);
	p = strchr(q, ' ');
	entry = follow_aliases(group_name.chars());
	if (!entry)
	    continue;

	/* Open the group file.  If necessary, check to make sure we don't
	 * already have an entry for this article. */
	construct_group_name_clear(&filename, group_name.chars());
	fp = open_group_file(filename.chars(), check);
	if (!fp) {
	    error("Can't open group file %s: %m", filename.chars());
	    entry->unlock();
	    return -1;
	}
	if (check) {
	    /* See if the article is already there; also count the number
	     * of articles and check the last article index, since they
	     * may be inconsistent. */
	    if (check_file(fp, partition, bucket_num, offset, entry)) {
		fclose(fp);
		entry->unlock();
		continue;
	    }
	} else {
	    /* Update the last index and number fields for this article. */
	    if (index > entry->last)
		entry->last = index;
	    entry->num++;
	}

	/* Write out a line to the file. */
	format_clear(&index_buf, "%l\t%d\t%l\t%l\t",
		     index, partition, bucket_num, offset);
	fputs(index_buf.chars(), fp);
	fputs(over_buf.chars(), fp);
	putc('\n', fp);
	fclose(fp);

        /* If someone has the newsgroup open, update that list. */
	if (entry->newsgroup) {
	    entry->newsgroup->add_entry(index, partition, bucket_num,
					offset, over_buf.chars());
	}

	entry->unlock();
    }

    return 0;
}

int Groups::write_new_bounds_file()
{
    FILE *fp;
    int i, count, status;
    Newsgroup_list *entry;
    String buf;

    fp = bounds_file->open_new();
    if (!fp)
	return -1;

    /* Write out the new file. */
    read_lock();
    update_lock.lock();
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next) {
	    format_clear(&buf, "%s %d %d %d\n",
			 entry->name, entry->first, entry->last, entry->num);
	    count = fwrite(buf.chars(), sizeof(char), buf.length(), fp);
	    if (count != buf.length()) {
		error("Can't write data to bounds file: %s");
		fclose(fp);
		update_lock.unlock();
		return -1;
	    }
	}
    }
    read_done();
    fclose(fp);

    status = bounds_file->update();
    update_lock.unlock();
    return status;
}

void Groups::send_grouplist(Connection *conn,
			    const Wildcard_list<bool> *access)
{
    int i;
    Newsgroup_list *entry;
    const char *flags;

    read_lock();
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next) {
	    if (!access || access->match(entry->name, 0)) {
		/* For moderated group, omit name of moderator. */
		flags = (*entry->flags == 'm') ? "m" : entry->flags;
		conn->writef("%s %N %N %s\r\n", entry->name, entry->last,
			     entry->first, flags);
	    }
	}
    }
    read_done();
}

void Groups::send_group_descriptions(Connection *conn,
				     const Wildcard_list<bool> *access)
{
    int i;
    Newsgroup_list *entry;

    read_lock();
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next) {
	    if (!access || access->match(entry->name, 0))
		conn->writef("%s %s\n", entry->name, entry->desc);
	}
    }
    read_done();
}

void Groups::send_new_groups(Connection *conn, time_t mintime,
			     const char *dist,
			     const Wildcard_list<bool> *access)
{
    int i;
    Newsgroup_list *entry;
    const char *flags;

    read_lock();
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next) {
	    if (entry->creation < mintime)
		continue;
	    if (*dist && !match_distribution(entry->name, dist))
		continue;
	    if (access && !access->match(entry->name, 0))
		continue;
	    flags = (*entry->flags == 'm') ? "m" : entry->flags;
	    conn->writef("%s %N %N %s\r\n", entry->name, entry->last,
			 entry->first, flags);
	}
    }
    read_done();
}

int Groups::eliminate_bucket(time_t bucket_num)
{
    int i;
    Newsgroup_list *entry;
    String buf;

    read_lock();
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next) {
	    entry->lock();

	    /* Update the newsgroup, if there is one. */
	    if (entry->newsgroup)
		entry->newsgroup->eliminate_bucket(bucket_num);

	    /* Update the file. */
	    construct_group_name_clear(&buf, entry->name);
	    if (eliminate_bucket_in_file(bucket_num, buf.chars(), entry) < 0) {
		entry->unlock();
		read_done();
		return -1;
	    }
	    entry->unlock();
	}
    }
    read_done();

    return write_new_bounds_file();
}

int Groups::verify_post_access(const char *groups,
			       const Wildcard_list<bool> *access,
			       String *which, String *mod)
{
    String group_name;
    Newsgroup_list *entry;

    while (*groups) {
	groups = get_group_name(&group_name, groups);
	entry = follow_aliases(group_name.chars());
	if (!entry || !strchr("ymj", *entry->flags) ||
	    !access->match(group_name.chars(), 0) ||
	    !access->match(entry->name, 0)) {
	    if (entry)
		entry->unlock();
	    which->append(&group_name);
	    return -1;
	}
	if (*entry->flags == 'm' && mod) {
	    if (entry->flags[1])
		mod->append(entry->flags + 1);
	    else
		config->get_moderator(group_name.chars(), mod);
	}
	entry->unlock();
    }
    return 0;
}

int Groups::any_groups_moderated(const char *groups)
{
    String group_name;
    Newsgroup_list *entry;
    bool moderated;

    while (*groups) {
	groups = get_group_name(&group_name, groups);
	entry = follow_aliases(group_name.chars());
	moderated = (entry && *entry->flags == 'm');
	if (entry)
	    entry->unlock();
	if (moderated)
	    return 1;
    }
    return 0;
}

void Groups::add_group(const char *group, int grouplen,
		       const char *flags, int flagslen,
		       const char *desc, int desclen,
		       time_t creation)
{
    String desc_str(desc, desclen), line;
    FILE *fp;

    check_newsgroups_file();
    write_lock();
    update_lock.lock();
    if (add_group_hashtable_entry(group, grouplen, flags, flagslen,
				  desc_str.chars(), creation)) {
	/* Entry already existed, update the long way. */
	update_newsgroups_file();
    } else {
	/* Append the new entry to the end. */
	fp = fopen(newsgroups_filename.chars(), "a");
	if (fp) {
	    fprintf(fp, "%*s %ld %*s %*s\n", grouplen, group, (long) creation,
		    flagslen, flags, desclen, desc);
	    fclose(fp);
	} else {
	    error("Can't open newsgroups file to append group %*s.\r\n",
		  grouplen, group);
	}
    }
    update_lock.unlock();
    write_done();
}

void Groups::merge_groups(FILE *fp)
{
    check_newsgroups_file();
    write_lock();
    read_newsgroups_file(fp);
    update_newsgroups_file();
    write_done();
}

int Groups::cancel_article(const char *xref)
{
    const char *p, *q;
    String group_name, filename, line;
    int index, found = 0;
    FILE *fp;
    fpos_t pos;
    Newsgroup_list *entry;

    p = strchr(xref, ' ');
    while (p) {

	/* Get a group name and find the hash table entry for it.  If there's
	 * no hash table, the group might have been rmgrouped, so ignore it. */
	q = strchr(p + 1, ':');
	group_name.clear();
	group_name.append(p + 1, q - (p + 1));
	index = atoi(q + 1);
	p = strchr(q, ' ');
	entry = follow_aliases(group_name.chars());
	if (!entry)
	    continue;

	/* Construct the group file name and open it. */
	construct_group_name_clear(&filename, group_name.chars());
	fp = fopen(filename.chars(), "r+");
	if (!fp) {
	    error("Can't open group file to cancel article %s/%d",
		  group_name.chars(), index);
	    entry->unlock();
	    return -1;
	}

	fgetpos(fp, &pos);
	while (read_line_clear(&line, fp) >= 0) {
	    if (atoi(line.chars()) == index) {
		fsetpos(fp, &pos);
		putc('C', fp);
		found = 1;
		break;
	    }
	    fgetpos(fp, &pos);
	}
	fclose(fp);

	if (found) {
	    if (entry->newsgroup)
		entry->newsgroup->eliminate_index(index);
	    entry->num--;
	}

	entry->unlock();
    }

    return 0;
}

/* Don't call this with a read or write lock; it gets a write lock. */
void Groups::check_newsgroups_file()
{
    struct stat statbuf;
    Newsgroup_list *entry;
    FILE *fp;
    int i;

    /* Get information about the newsgroups file. */
    if (stat(newsgroups_filename.chars(), &statbuf) < 0) {
	error("Can't stat newsgroups file.");
	return;
    }

    /* If the newsgroups file hasn't been modified since we last read it,
     * we don't need to read it. */
    if (statbuf.st_mtime < newsgroups_up_to_date)
	return;

    write_lock();

    /* Set in_file marks to 0. */
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next)
	    entry->in_file = 0;
    }

    fp = newsgroups_file->open_read();
    if (fp) {
	/* Read the newsgroups file, setting in_file marks to 1. */
	read_newsgroups_file(fp);
	fclose(fp);

	/* Delete any groups which disappeared. */
	for (i = 0; i < size; i++) {
	    for (entry = hashtab[i]; entry; entry = entry->next) {
		if (entry->in_file == 0)
		    ;
	    }
	}
    }

    write_done();
}

/* Assumes write lock or initialization phase. */
void Groups::read_newsgroups_file(FILE *fp)
{
    String buf;
    int flen;
    const char *p, *q, *r, *flags, *desc, *space;
    time_t creation, now;

    time(&now);
    while (read_line_clear(&buf, fp) >= 0) {

	/* Ignore comments and blank lines. */
	if (!buf.length() || *buf.chars() == '#')
	    continue;

	/* Read in name, creation time, flags, description. */
	r = q = word_end(p = buf.chars());
	creation = atoi(next_word(&r));
	flags = next_word(&r);
	space = strchr(flags, ' ');
	flen = (space) ? space - flags : strlen(flags);
	desc = find_word(flags + flen);

	/* Check for the validity of flags, and add the group. */
	if (check_flags(flags, flen) < 0) {
	    warn("Invalid newsgroups line: %s\n", buf.chars());
	    continue;
	}
	add_group_hashtable_entry(p, q - p, flags, flen, desc, creation);
    }

    newsgroups_up_to_date = now;
}

void Groups::read_bounds_file()
{
    FILE *fp;
    const char *p;
    Newsgroup_list *entry;
    String buf, name;

    fp = bounds_file->open_read();
    if (!fp)
	return;

    while (read_line_clear(&buf, fp) >= 0) {

	/* Find the end of the newsgroup name. */
	p = word_end(buf.chars());

	entry = find_table_entry(buf.chars(), p - buf.chars());
	if (!entry) {
	    construct_group_name_clear(&name, buf.chars());
	    delete_group_file(name.chars(), dirname.length());
	    continue;
	}

	/* Read in the first and last indices and the number of articles. */
	entry->first = atoi(next_word(&p));
	entry->last = atoi(next_word(&p));
	entry->num = atoi(next_word(&p));
	entry->unlock();
    }

    fclose(fp);
}

/* Internal method--assumes write lock or initialization phase. */
int Groups::add_group_hashtable_entry(const char *name, int len,
				      const char *flags, int flagslen,
				      const char *desc, time_t creation)
{
    Newsgroup_list *entry;
    int hashval;

    entry = find_table_entry(name, len);
    if (entry) {

	/* Replace fields in the current entry. */
	delete[] entry->name;
	delete[] entry->flags;
	delete[] entry->desc;
	entry->name = strduplicate(name, len);
	entry->flags = strduplicate(flags, flagslen);
	entry->desc = strduplicate(desc);
	entry->in_file = 1;
	entry->unlock();
	return 1;

    } else {

	/* Expand the hash table if it's getting too big. */
	if (num_groups >= size * 4)
	    expand_hash_table();

	/* Make a new newsgroup entry. */
	entry = new Newsgroup_list;
	entry->name = strduplicate(name, len);
	entry->first = 1;
	entry->last = 0;
	entry->num = 0;
	entry->first_unreserved = 1;
	entry->creation = creation;
	entry->flags = strduplicate(flags, flagslen);
	entry->desc = strduplicate(desc);
	entry->in_file = 1;
	entry->newsgroup = NULL;

	/* Insert it into the hash table. */
	hashval = hash_case(entry->name) % size;
	entry->next = hashtab[hashval];
	hashtab[hashval] = entry;

	num_groups++;
	return 0;
    }
}

/* Internal method--assumes lock or initialization phase. */
void Groups::expand_hash_table()
{
    int new_size, i, hashval;
    Newsgroup_list **new_tab, *entry, *next;

    /* Make space for a new hash table. */
    new_size = adjust_size(size + 1, MIN_POWER, DELTA);
    new_tab = new Newsgroup_list *[new_size];
    for (i = 0; i < new_size; i++)
	new_tab[i] = NULL;

    /* Rehash each entry in the old table. */
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = next) {
	    next = entry->next;
	    hashval = hash_case(entry->name) % new_size;
	    entry->next = new_tab[hashval];
	    new_tab[hashval] = entry;
	}
    }

    /* Replace the old table with the new one. */
    delete[] hashtab;
    hashtab = new_tab;
    size = new_size;
}

Newsgroup_list *Groups::find_table_entry(const char *name, int len)
{
    Newsgroup_list *entry;
    int hashval;

    hashval = hash_case(name, len) % size;
    read_lock();
    for (entry = hashtab[hashval]; entry; entry = entry->next) {
	if (strnccmp(entry->name, name, len) == 0 && entry->name[len] == 0)
	    break;
    }
    if (entry)
	entry->lock();
    read_done();
    return entry;
}

Newsgroup_list *Groups::follow_aliases(const char *name)
{
    Newsgroup_list *entry, *alias;

    /* Find the table entry.  If the flags don't start with '=', we're done. */
    entry = find_table_entry(name, strlen(name));
    if (!entry || *entry->flags != '=')
	return entry;

    /* We have an alias; look up the real group name. */
    alias = follow_aliases(entry->flags + 1);
    entry->unlock();
    return alias;
}

/* Assumes update lock and write lock. */
void Groups::update_newsgroups_file()
{
    FILE *old_fp, *new_fp;
    String buf;
    Newsgroup_list *entry;
    const char *s, *p, *q, *r, *t, *u, *v;
    int i;

    /* Open file handles for the old and new newsgroups files. */
    old_fp = newsgroups_file->open_read();
    if (!old_fp) {
	error("Can't open old newsgroups file for update.");
	return;
    }
    new_fp = newsgroups_file->open_new();
    if (!new_fp) {
	fclose(old_fp);
	error("Can't open new newsgroups file for update.");
	return;
    }

    /* Set all in_file fields to 0. */
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next)
	    entry->in_file = 0;
    }

    /* For each newsgroup in old_fp, write a line in new_fp and set the
     * newsgroups's in_file field to 1. */
    while (read_line_clear(&buf, old_fp) >= 0) {

	if (!buf.length() || *buf.chars() == '#') {
	    fprintf(new_fp, "%s\n", buf.chars());
	    continue;
	}

	/* Get first four words (s to p, q to r, t to u, v). */
	r = word_end(q = find_word(p = word_end(s = buf.chars())));
	v = find_word(u = word_end(t = find_word(r)));

	/* Make sure all white space is non-empty. */
	p = (*p) ? p : " ";
	r = (*r) ? r : " ";
	u = (*t) ? t : " ";

	/* Write out the entry, using the same white space as in the original
	 * newsgroups file. */
	entry = find_table_entry(s, p - s);
	if (entry) {
	    entry->in_file = 1;
	    fprintf(new_fp, "%s%*s%ld%*s%s%*s%s\n", entry->name, q - p, p,
		    (long) entry->creation, t - r, r, entry->flags, v - u, u,
		    entry->desc);
	}
    }

    /* Look for new newsgroups and write them out. */
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next) {
	    if (!entry->in_file) {
		fprintf(new_fp, "%s %ld %s %s\n", entry->name,
			(long) entry->creation, entry->flags, entry->desc);
	    }
	}
    }

    fclose(old_fp);
    fclose(new_fp);
    newsgroups_file->update();
}

void Groups::construct_group_name_clear(String *buf, const char *name)
{
    const char *p;

    buf->clear();
    buf->append(&dirname);
    while (1) {
	p = strchr(name, '.');
	if (p) {
	    buf->append(name, p - name);
	    buf->append('/');
	    name = p + 1;
	} else {
	    buf->append(name);
	    buf->append(".info", 5);
	    return;
	}
    }
}

/* Read through a newsgroup file checking for a bucket number and offset.
 * This only happens during re-indexing of articles.  In this case, though,
 * we also don't know whether the number of articles or the last article
 * index in the group file are correct, and we need to update the entry,
 * which is passed to us locked. */
static int check_file(FILE *fp, int partition, time_t bucket_num,
		      long offset,  Newsgroup_list *entry)
{
    String buf;
    int part;
    time_t bucket;
    long off;
    int index, found = 0;
    const char *p;

    entry->num = entry->last = 0;
    while (read_line_clear(&buf, fp) >= NULL) {
	p = buf.chars();
	index = atoi(p);
	part = atoi(next_word(&p));
	bucket = atol(next_word(&p));
	off = atol(next_word(&p));
	if (part == partition && bucket == bucket_num && off == offset)
	    found = 1;
	if (index > entry->last)
	    entry->last = index;
	entry->num++;
    }

    return found;
}

/* Assumes lock on entry. */
static int eliminate_bucket_in_file(time_t bucket_num, const char *name,
				    Newsgroup_list *entry)
{
    Atomic_file groupfile(name);
    FILE *in_fp, *out_fp;
    String line;
    int index, num = 0, first = 0, partition;
    time_t bucket;
    const char *p;

    /* Open the current and new group files. */
    in_fp = groupfile.open_read();
    if (!in_fp)
	return 0;
    out_fp = groupfile.open_new();
    if (!out_fp) {
	fclose(in_fp);
	return -1;
    }

    /* Copy the group file, filtering lines with the bucket in question. */
    while (read_line_clear(&line, in_fp) >= 0) {
	p = line.chars();
	index = atol(p);
	partition = atol(next_word(&p));
	bucket = atol(next_word(&p));
	if (bucket == bucket_num)
	    continue;
	if (fputs(line.chars(), out_fp) == EOF || putc('\n', out_fp) == EOF) {
	    fclose(in_fp);
	    fclose(out_fp);
	    error("Can't write new group file for %s: %m", name);
	    return -1;
	}
	num++;
	if (!first || index < first)
	    first = index;
    }
    fclose(in_fp);
    fclose(out_fp);

    /* Atomically update the group file. */
    if (groupfile.update() < 0)
	return -1;

    /* Update the fields in entry and return success. */
    entry->num = num;
    entry->first = (first > 0) ? first : entry->last + 1;
    return 0;
}

/* Dist is guaranteed to be of the form "<distributions>". */
static int match_distribution(const char *name, const char *dist)
{
    int name_dist_len;
    const char *p;

    /* Find length of first component in name. */
    p = strchr(name, '.');
    name_dist_len = (p) ? p - name : strlen(name);

    /* Move past '<' in distributions, check for empty list. */
    dist++;
    if (*dist == '>')
	return 0;

    /* Look for matches. */
    while (*dist) {
	p = strpbrk(dist, ",>");
	if (p - dist == name_dist_len && strnccmp(dist, name, p - dist) == 0)
	    return 1;
	dist = p + 1;
    }

    /* Didn't find a match. */
    return 0;
}

static FILE *open_group_file(const char *name, int check)
{
    FILE *fp;

    /* Try common cases (existing file, or file in existing directory). */
    fp = fopen(name, (check) ? "r+" : "a+");
    if (fp)
	return fp;
    fp = fopen(name, (check) ? "w+" : "w");
    if (fp)
	return fp;

    /* Make all the directories leading up to the file. */
    make_directories_in_path(name);
    return fopen(name, (check) ? "w+" : "w");
}

static void delete_group_file(const char *name, int dirnamelen)
{
    String buf;
    const char *p;
    DIR *dir;
    struct dirent d, *dp;
    int found;

    /* Remove the file itself. */
    if (unlink(name) < 0 && errno != ENOENT) {
	error("Can't unlink group file \"%s\": %m", name);
	return;
    }

    /* Remove any empty directories created by removing the file. */
    buf.append(name);
    while (1) {
	/* Lop off the trailing path component. */
	p = strrchr(buf.chars(), '/');
	buf.truncate(p - buf.chars());

	/* Quit if we're down to dirname. */
	if (buf.length() <= dirnamelen)
	    break;

	dir = opendir(buf.chars());
	if (!dir)
	    break;
	while (readdir_r(dir, &d, &dp) >= 0 && dp) {
	    if (strcmp(d.d_name, ".") != 0 && strcmp(d.d_name, "..") != 0) {
		found = 1;
		break;
	    }
	}
	closedir(dir);
	if (found)
	    break;
	rmdir(buf.chars());
    }
}

static void *periodic_check_newsgroups(void *arg)
{
    while (1) {
	sleep(60);
	((Groups *) arg)->check_newsgroups_file();
    }
}

