/* Headers.cc: Data structure for storing headers. */

#include "threads.h"
#include "util.h"

#include "Headers.h"
#include "String.h"

extern "C" {
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
}

struct Field {
    Field *hash_next;
    Field *next;
    int name_len;
    const char *value;
    char *line;
};

Headers::Headers()
{
    int i;

    for (i = 0; i < HEADER_HASHTAB_SIZE; i++)
	hashtab[i] = NULL;
    first = NULL;
    last = NULL;
    current = NULL;
}

Headers::~Headers()
{
    Field *next;

    while (first) {
	next = first->next;
	delete[] first->line;
	delete first;
	first = next;
    }
}

int Headers::add_header(const char *line)
{
    return add_header(line, strlen(line));
}

/* Returns -1 if header is malformatted, 0 if a new header, 1 if replacing
 * an existing header. */
int Headers::add_header(const char *line, int len)
{
    int hashval, i, name_len;
    Field *field;
    const char *p;

    /* Find separating colon. */
    if (len < 2 || isspace(*line) || *line == ':')
	return -1;
    p = (const char *) memchr(line + 1, ':', len - 1);
    if (!p)
	return -1;
    i = p - line;

    /* Find the name length and field value. */
    name_len = i;
    while (isspace(line[name_len - 1])) {
	if (--name_len <= 0)
	    return -1;
    }
    i++;
    while (i < len && isspace(line[i]))
	i++;

    field = find_field(line, name_len);
    if (field != NULL) {
	delete[] field->line;
	field->line = strduplicate(line, len);
	field->value = field->line + i;
	return 1;
    }

    /* Create a new field for this header. */
    field = new Field;
    field->line = strduplicate(line, len);
    field->name_len = name_len;
    field->value = field->line + i;

    /* Insert the field into the linked list and hash table. */
    field->next = NULL;
    if (last)
	last->next = field;
    else
	first = field;
    last = field;
    hashval = hash_case(line, name_len) % HEADER_HASHTAB_SIZE;
    field->hash_next = hashtab[hashval];
    hashtab[hashval] = field;

    return 0;
}

int Headers::add_headers(FILE *fp)
{
    String line;
    int c, num_lines = 0;

    while (read_line_clear(&line, fp) == 0) {
	num_lines++;
	if (line.length() == 0)
	    return num_lines;
	while (1) {
	    c = getc(fp);
	    if (c == EOF)
		return -1;
	    ungetc(c, fp);
	    if (c == ' ' || c == '\t') {
		line.append('\n');
		if (read_line(&line, fp) < 0)
		    return -1;
	    } else {
		break;
	    }
	}
	if (line.length() == 1 && *line.chars() == '.')
	    return -1;
	if (add_header(line.chars(), line.length()) != 0)
	    return -1;
    }
    return -1;
}

const char *Headers::add_headers(const char *lines)
{
    const char *p;

    while (1) {
	if (*lines == '\n')
	    return lines + 1;
	p = lines + 1;
	while (1) {
	    p = strchr(p, '\n');
	    if (!p)
		return NULL;
	    if (p[1] != ' ' && p[1] != '\t')
		break;
	    p++;
	}
	if (add_header(lines, p - lines) != 0)
	    return NULL;
	lines = p + 1;
    }
}

const char *Headers::find_value(const char *name) const
{
    return find_value(name, strlen(name));
}

const char *Headers::find_value(const char *name, int len) const
{
    Field *field;

    field = find_field(name, len);
    return (field) ? field->value : NULL;
}

const char *Headers::first_header()
{
    current = first;
    return (current) ? current->line : NULL;
}

const char *Headers::next_header()
{
    if (!current)
	return NULL;
    current = current->next;
    return (current) ? current->line : NULL;
}

void Headers::clear()
{
    Field *next;
    int i;

    while (first) {
	next = first->next;
	delete[] first->line;
	delete first;
	first = next;
    }
    for (i = 0; i < HEADER_HASHTAB_SIZE; i++)
	hashtab[i] = NULL;
    last = NULL;
    current = NULL;
}

Field *Headers::find_field(const char *name, int len) const
{
    int hashval;
    Field *field;

    hashval = hash_case(name, len) % HEADER_HASHTAB_SIZE;
    for (field = hashtab[hashval]; field; field = field->hash_next) {
	if (field->name_len == len && strnccmp(field->line, name, len) == 0)
	    return field;
    }
    return NULL;
}

