/* Index.cc: Maintaining the message-ID index. */

#include "threads.h"
#include "error.h"
#include "util.h"

#include "Index.h"
#include "Config.h"
#include "Dbz.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
}

/* Make configurable. */
const int MODE = S_IRUSR | S_IWUSR | S_IROTH;

Index::Index(const Config *config_arg)
{
    struct stat statbuf;

    config = config_arg;

    /* Construct the name of the index. */
    config->get_index_directory(&index_name);
    index_name.append("articles");

    make_directories_in_path(index_name.chars());

    /* Open the database. */
    db = new Dbz();
    db->set_incore(1);
    if (db->init(index_name.chars()) < 0) {
	if (db->fresh(index_name.chars(), 0, '\t', 'C', 1))
	    die("Can't open dbz database.");
    }

    /* Open the flat file. */
    flatfile_writer = open(index_name.chars(), O_WRONLY | O_APPEND | O_CREAT,
			   MODE);
    if (flatfile_writer < 0)
	die("Can't open index flat file: %m");
    flatfile_rover = open(index_name.chars(), O_RDWR, MODE);
    if (flatfile_rover < 0)
	die("Can't open index flat file: %m");

    /* stat the flat file to determine its initial size. */
    if (fstat(flatfile_rover, &statbuf) < 0)
	die("Can't stat flat file: %m");
    position = statbuf.st_size;
}

Index::~Index()
{
    db->close();
    close(flatfile_rover);
    close(flatfile_writer);
}

/* Returns -1 if id maps to a dead entry. */
int Index::lookup(const char *id, int *partition, time_t *bucket_num,
		  long *offset)
{
    char buf[100];
    const char *p = buf;

    if (get_line(id, buf, sizeof(buf), NULL) < 0)
	return -1;

    if (*buf == 'd')
	return -1;

    /* Read in the values from the buffer. */
    *partition = atoi(p);
    *bucket_num = atol(next_word(&p));
    *offset = atol(next_word(&p));
    return 0;
}

/* Returns true if any entry exists for id, even if it's a dead entry. */
int Index::present(const char *id)
{
    return (find_entry(id) >= 0);
}

int Index::store(const char *id, int partition, time_t bucket_num,
		 long offset)
{
    String line;
    int result;

    format(&line, "%s\t%d %l %l\n", id, partition, (long) bucket_num, offset);
    lock();
    result = store_line(id, line.chars(), line.length());
    unlock();
    return result;
}

int Index::store_dead_entry(const char *id)
{
    String line;
    int result;

    format(&line, "%s d%l\n",
	   id, (long)(time(NULL) + config->get_cancel_keep_for()));
    lock();
    result = store_line(id, line.chars(), line.length());
    unlock();
    return result;
}

int Index::overwrite_with_dead_entry(const char *id)
{
    char line[100], buf[20];
    long offset;
    int buflen, linelen;

    /* Get the existing line for id. */
    lock();
    if (get_line(id, line, sizeof(line), &offset) < 0) {
	unlock();
	return -1;
    }

    /* If it's already a dead entry, leave it alone. */
    if (*line == 'd') {
	unlock();
	return 0;
    }

    /* Overwrite the line with a dead entry. */
    linelen = strlen(line);
    sprintf(buf, "d%ld", (long)(time(NULL) + config->get_cancel_keep_for()));
    buflen = strlen(buf);
    memcpy(line, buf, buflen);
    memset(line + buflen, ' ', linelen - buflen - 1);

    /* Write the new line to the flatfile. */
    if (lseek(flatfile_rover, offset, SEEK_SET) < 0) {
	unlock();
	error("Can't seek to %d for Message-ID %s in index flat file: %m",
	      offset, id);
	return -1;
    }
    if (write(flatfile_rover, line, linelen) != linelen) {
	unlock();
	error("Can't rewrite index entry for Message-ID %s: %m", id);
	return -1;
    }

    unlock();
    return 0;
}

int Index::sync()
{
    int result;

    lock();
    result = db->sync();
    unlock();
    return result;
}

/* No threads should be performing write operations while this is taking place.
 * Read operations are okay. */
int Index::eliminate_bucket(time_t bucket_num)
{
    String new_name, old_name;
    String dirname, new_dirname, old_dirname;
    String pagname, new_pagname, old_pagname;
    String line;
    Dbz *new_db;
    FILE *in_fp = NULL, *out_fp = NULL;
    const char *p;
    time_t bucket, now = time(NULL);
    int rover = -1, writer = -1, result;
    datum key, value;
    long offset;

    /* Open file handles to the old and new flat files. */
    new_name.append(&index_name);
    new_name.append(".new");
    in_fp = fopen(index_name.chars(), "r");
    if (!in_fp) {
	error("Can't reopen index flatfile %s: %m", index_name.chars());
	return -1;
    }
    out_fp = fopen(new_name.chars(), "w");
    if (!out_fp) {
	error("Can't open new index flatfile %s: %m", new_name.chars());
	fclose(in_fp);
	return -1;
    }

    /* Copy from in_fp to out_fp, leaving out lines in the expired bucket. */
    while (read_line_clear(&line, in_fp) >= 0) {
	p = strchr(line.chars(), '\t');
	if (!p)
	    continue;
	if (p[1] == 'd') {
	    /* Dead entry.  Expire it if its expiration time is earlier than
	     * the current time.  If not, chop it off after the expiration
	     * time, since it might have some extra stuff on the end from
	     * overwrite_with_dead_entry(). */
	    if (atol(p + 2) < now)
		continue;
	    else
		line.truncate(word_end(p + 2) - line.chars());
	} else {
	    bucket = atol(find_word(word_end(p + 1)));
	    if (bucket == bucket_num) {
		if (expire_keep_for == 0)
		    continue;
		line.truncate(p + 1 - line.chars());
		format(&line, "d%l",
		       (long)(bucket_num + config->get_expire_keep_for()));
	    }
	}
	if (fputs(line.chars(), out_fp) == EOF || putc('\n', out_fp) == EOF) {
	    fclose(in_fp);
	    fclose(out_fp);
	    return -1;
	}
    }
    fclose(in_fp);
    fclose(out_fp);

    /* Open a file handle to the new flatfile and create a new database. */
    in_fp = fopen(new_name.chars(), "r");
    if (!in_fp) {
	error("Can't reopen new flatfile %s: %m", new_name.chars());
	return -1;
    }
    new_db = new Dbz();
    new_db->set_incore(1);
    if (new_db->again(new_name.chars(), index_name.chars()) < 0) {
	error("Can't open new DBZ database: %m");
	fclose(in_fp);
	delete new_db;
	return -1;
    }

    /* Build the DBZ index. */
    offset = 0;
    value.dsize = sizeof(offset);
    while (read_line_clear(&line, in_fp) >= 0) {
	p = strchr(line.chars(), '\t');
	if (!p)
	    continue;
	key.dptr = (char *) line.chars();
	key.dsize = p - line.chars();
	value.dptr = (char *) &offset;
	if (new_db->store(key, value) < 0) {
	    warn("Can't write new DBZ database: %m");
	    fclose(in_fp);
	    new_db->close();
	    delete new_db;
	    return -1;
	}
	offset += line.length() + 1;
    }
    fclose(in_fp);

    /* Sync the new database before switching over to it, out of paranoia. */
    if (new_db->sync() < 0) {
	warn("Can't sync new DBZ database: %m");
	new_db->close();
	delete new_db;
	return -1;
    }

    /* Open rover and writer descriptors to the new flatfile. */
    writer = open(new_name.chars(), O_WRONLY | O_APPEND | O_CREAT, MODE);
    rover = open(new_name.chars(), O_RDWR, MODE);
    if (writer < 0 || rover < 0) {
	error("Can't reopen new index flatfile %s: %m", new_name.chars());
	new_db->close();
	delete new_db;
	if (writer >= 0)
	    close(writer);
	if (rover >= 0)
	    close(rover);
	return -1;
    }

    /* Rename the database files.  This isn't totally atomic, but it's rare. */
    old_name.append(&index_name);
    old_name.append(".old");
    dirname.append(&index_name);
    dirname.append(".dir");
    old_dirname.append(&index_name);
    old_dirname.append(".old.dir");
    new_dirname.append(&index_name);
    new_dirname.append(".new.dir");
    pagname.append(&index_name);
    pagname.append(".pag");
    old_pagname.append(&index_name);
    old_pagname.append(".old.pag");
    new_pagname.append(&index_name);
    new_pagname.append(".new.pag");
    unlink(old_name.chars());
    unlink(old_dirname.chars());
    unlink(old_pagname.chars());
    result = rename(index_name.chars(), old_name.chars());
    result |= rename(dirname.chars(), old_dirname.chars());
    result |= rename(pagname.chars(), old_pagname.chars());
    unlink(index_name.chars());
    unlink(dirname.chars());
    unlink(pagname.chars());
    result |= rename(new_name.chars(), index_name.chars());
    result |= rename(new_dirname.chars(), dirname.chars());
    result |= rename(new_pagname.chars(), pagname.chars());
    unlink(old_name.chars());
    unlink(old_dirname.chars());
    unlink(old_pagname.chars());
    if (result != 0)
	die("Couldn't rename article index files: %m");

    /* Switch over to the new database. */
    lock();
    close(flatfile_rover);
    close(flatfile_writer);
    db->close();
    delete db;
    flatfile_rover = rover;
    flatfile_writer = writer;
    position = offset;
    db = new_db;
    unlock();

    return 0;
}

/* Assumes lock. */
int Index::get_line(const char *id, char *buf, int bufsize, long *offset)
{
    long off;
    char *p;

    if (!offset)
	offset = &off;

    /* Look up the offset in the index. */
    *offset = find_entry(id);
    if (*offset < 0)
	return -1;

    /* Skip past the message ID and read the data. */
    *offset += strlen(id) + 1;
    if (lseek(flatfile_rover, *offset, SEEK_SET) < 0) {
	error("Can't seek to %d for Message-ID %s in index flat file: %m",
	      *offset, id);
	return -1;
    }
    if (read(flatfile_rover, buf, bufsize - 1) <= 0) {
	error("Can't read at %d for Message-ID %s in index flat file: %m",
	      *offset, id);
	return -1;
    }
    unlock();

    buf[bufsize - 1] = 0;
    p = strchr(buf, '\n');
    if (p)
	p[1] = 0;
    return 0;
}

/* Assumes lock. */
int Index::store_line(const char *id, const char *line, int len)
{
    datum key, value;
    int result;

    /* Write the bucket and offset into the flat file. */
    result = write(flatfile_writer, line, len);
    if (result != len) {
	warn("Can't write index entry for Message-ID %s: %m", id);
	return -1;
    }

    /* Store the association in the dbz database. */
    key.dptr = (char *) id;
    key.dsize = strlen(id);
    if (key.dsize > DBZMAXKEY)
	key.dsize = DBZMAXKEY;
    value.dptr = (char *) &position;
    value.dsize = sizeof(position);
    if (db->store_case(key, value) < 0) {
	warn("Can't store pair for Message-ID %s.", id);
	return -1;
    }

    /* Update the position and return. */
    position += len;
    return 0;
}

/* Assumes lock. */
long Index::find_entry(const char *id)
{
    datum key;
    datum value;
    long offset;

    /* Look up the message ID in the database. */
    key.dptr = (char *) id;
    key.dsize = strlen(id);
    if (key.dsize > DBZMAXKEY)
	key.dsize = DBZMAXKEY;
    value = db->fetch_case(key);
    if (value.dptr == NULL || value.dsize != sizeof(offset))
	return -1;

    /* Copy the value to an aligned spot and return it. */
    memcpy(&offset, value.dptr, sizeof(offset));
    return offset;
}

