/* Newsfeed.cc: Handler for feeding articles to another site. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Newsfeed.h"
#include "Config.h"
#include "Headers.h"
#include "Connection.h"
#include "Article.h"
#include "Spool.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
}

static void *run_feed(void *arg);

Newsfeed::Newsfeed(const char *name_arg, Spool *spool_arg,
		   const Config *config)
{
    struct stat statbuf;

    name = strduplicate(name_arg);
    spool = spool_arg;
    groups_list = config->get_feed_groups(name);
    dists_list = config->get_feed_dists(name);
    paths_list = config->get_feed_paths(name);
    config->get_feeds_directory(&dirname);
    format(&queue_filename, "%S/%s.queue", &dirname, name);
    format(&work_filename, "%S/%s.work", &dirname, name);
    queue_file = fopen(queue_filename.chars(), "a+");
    if (!queue_file)
	die("Can't open queue file %S for output.", &queue_filename);

    if (stat(queue_filename.chars(), &statbuf) < 0)
	die("Can't stat queue file %s.", queue_filename.chars());
    work_flag = (statbuf.st_size > 0);

    start_thread(run_feed, this);
}

Newsfeed::~Newsfeed()
{
    delete[] name;
    delete groups_list;
    delete dists_list;
    delete paths_list;
}

void Newsfeed::feed_articles()
{
    FILE *fp;

    /* Check for a work file left by a crash. */
    fp = fopen(work_filename.chars(), "r");
    if (fp) {
	feed_articles_in_file(fp);
	fclose(fp);
    }

    /* Loop, waiting for work and processing the queue file. */
    while (1) {

	/* Wait until the queue file is non-empty. */
	file_lock.lock();
	while (!work_flag)
	    work.wait(&file_lock);

	/* Turn the queue file into the work file.  XXX error-handling. */
	fclose(queue_file);
	rename(queue_filename.chars(), work_filename.chars());
	queue_file = fopen(queue_filename.chars(), "a+");
	work_flag = 0;
	file_lock.unlock();

	/* Process the work file. */
	fp = fopen(work_filename.chars(), "r");
	feed_articles_in_file(fp);
	fclose(fp);
    }
}

int Newsfeed::propagate_article(int partition, time_t bucket_num, long offset,
				const Headers *headers, const char *body)
{
    if (!wanted(headers))
	return 0;

    return queue_article(partition, bucket_num, offset);
}

void Newsfeed::feed_articles_in_file(FILE *fp)
{
    char buf[100];
    const char *p;
    time_t bucket_num;
    long offset;
    int partition;
    Connection *conn;

    conn = make_connection();
    while (!conn) {
	sleep(300);
	conn = make_connection();
    }
    while (fgets(buf, sizeof(buf), fp)) {
	partition = atoi(buf);
	bucket_num = atol(p = find_word(word_end(buf)));
	offset = atol(find_word(word_end(p)));
	if (feed_article(conn, partition, bucket_num, offset) < 0) {
	    queue_article(partition, bucket_num, offset);
	    while (fgets(buf, sizeof(buf), fp)) {
		partition = atoi(buf);
		bucket_num = atol(p = find_word(word_end(buf)));
		offset = atol(find_word(word_end(p)));
		queue_article(partition, bucket_num, offset);
	    }
	    break;
	}
    }
    delete conn;
}

int Newsfeed::feed_article(Connection *conn, int partition, time_t bucket_num,
			   int offset)
{
    Article *art;
    String buf;

    /* Get a handle to the article from the spool. */
    art = spool->get_article(partition, bucket_num, offset);
    if (!art)
	return 0;

    /* Ask to send the article with an IHAVE. */
    art->get_message_id(&buf);
    conn->writef("IHAVE %S\r\n", &buf);
    if (conn->read_line_clear(&buf) < 0) {
	spool->release_article(art);
	return -1;
    }

    /* See if the remote server wants the article. */
    if (*buf.chars() == '4') {
	spool->release_article(art);
	return 0;
    }
    if (*buf.chars() != '3') {
	spool->release_article(art);
	return -1;
    }

    /* Send the article. */
    art->write_head(conn);
    conn->write("\r\n");
    art->write_body(conn);
    conn->write(".\r\n");

    /* See if we need to requeue the article. */
    if (conn->read_line_clear(&buf) < 0) {
	spool->release_article(art);
	return -1;
    }
    if (strncmp(buf.chars(), "436", 3) == 0)
	queue_article(partition, bucket_num, offset);

    spool->release_article(art);
    return 0;
}

Connection *Newsfeed::make_connection()
{
    String buf;
    int fd;
    Connection *conn;

    /* Make a TCP connection to the remote host. */
    fd = tcp_connect(name, "nntp", 119, &buf);
    if (fd < 0) {
	warn("%s", buf.chars());
	return NULL;
    }

    conn = new Connection(fd);

    /* Read the first line from the remote host. */
    if (conn->read_line(&buf) < 0) {
	warn("Couldn't read initial line from %s.", name);
	delete conn;
	return NULL;
    }

    /* See if we have read access to the host. */
    if (strncmp(buf.chars(), "200", 3) != 0) {
	warn("No transfer permission to %s: %s", name, buf.chars());
	if (*buf.chars() == '2') {
	    /* Shut down the connection gracefully. */
	    conn->write("QUIT\r\n");
	    conn->read_line_clear(&buf);
	}
	delete conn;
	return NULL;
    }

    return conn;
}

int Newsfeed::wanted(const Headers *headers)
{
    const char *path, *groups, *dists, *p;
    String tmp;

    /* Don't propagate if there is an excluded path component. */
    path = (*headers)["Path"];
    for (p = strchr(path, '!'); p; p = strchr(path = p + 1, '!')) {
	tmp.clear();
	tmp.append(path, p - path);
	if (paths_list->match(tmp.chars(), 1) == 0)
	    return 0;
    }
    if (paths_list->match(path, 1) == 0)
	return 0;

    /* Don't propagate if none of the newwsgroups are wanted. */
    groups = (*headers)["Newsgroups"];
    for (p = strchr(groups, ','); p; p = strchr(groups = p + 1, ',')) {
	tmp.clear();
	tmp.append(groups, p - groups);
	if (groups_list->match(tmp.chars(), 0) == 1)
	    break;
    }
    if (!p && groups_list->match(groups, 0) == 0)
	return 0;

    /* Don't propagate if distributions are specified and none are wanted. */
    dists = (*headers)["Distributions"];
    if (dists) {
	for (p = strchr(dists, ','); p; p = strchr(dists = p + 1, ',')) {
	    tmp.clear();
	    tmp.append(dists, p - dists);
	    if (dists_list->match(tmp.chars(), 1) == 1)
		break;
	}
	if (!p && dists_list->match(dists, 1) == 0)
	    return 0;
    }

    /* Propagate the article. */
    return 1;
}

int Newsfeed::queue_article(int partition, time_t bucket_num, long offset)
{
    char buf[100];

    /* Add the article to the queue file. */
    sprintf(buf, "%d\t%ld\t%ld\n", partition, (long) bucket_num, offset);
    file_lock.lock();
    if (fputs(buf, queue_file) < 0)
	return -1;
    work_flag = 1;
    work.signal();
    file_lock.unlock();
    return 0;
}

static void *run_feed(void *arg)
{
    ((Newsfeed *) arg)->feed_articles();
    return NULL;
}

