/* Newsgroup.cc: A newsgroup file. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Newsgroup.h"
#include "Connection.h"
#include "Config.h"
#include "Spool.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
}

const int MIN_POWER = 5;
const int DELTA = 2;
const int AVERAGE_LINE_LENGTH = 225; /* A heuristic. */

const char *find_overview_field(const char *overdata, int fnum, int *len);
static void send_header(Connection *conn, Index_entry *entry, int fnum,
			Spool *spool, const char *name, int namelen);

/* fp may be NULL, in which case the newsgroup is assumed to be empty. */
Newsgroup::Newsgroup(const char *name, FILE *fp, const Config *config_arg,
		     int last_arg)
{
    struct stat statbuf;
    int approx_num_entries = 0, index;
    String buf;
    int partition;
    time_t bucket_num;
    long offset;
    const char *overdata, *p;

    config = config_arg;
    last = last_arg;
    group_name.append(name);

    /* Estimate the number of entries in the group file. */
    if (fp && fstat(fileno(fp), &statbuf) >= 0)
	approx_num_entries = statbuf.st_size / AVERAGE_LINE_LENGTH;
    else if (fp)
	warn("Can't stat newsgroup file %s: %m", name);

    /* Initialize the index entries table. */
    entries_size = adjust_size(approx_num_entries, MIN_POWER, DELTA);
    entries = new Index_entry[entries_size];
    num_entries = 0;

    /* Read in the group file and close it, if we have one. */
    if (fp) {
	while (read_line_clear(&buf, fp) >= NULL) {
	    p = buf.chars();

	    /* Check if it's a cancelled line. */
	    if (*p == 'C')
		continue;

	    /* Extract the index, bucket number, and offset, and add an index
	     * entry to the end of the list. */
	    index = atol(p);
	    partition = atoi(next_word(&p));
	    bucket_num = atol(next_word(&p));
	    offset = atol(next_word(&p));
	    overdata = next_word(&p);
	    add_index_entry(index, partition, bucket_num, offset, overdata);
	}
	fclose(fp);
    }
}

Newsgroup::~Newsgroup()
{
    int i;

    for (i = 0; i < num_entries; i++)
	delete[] entries[i].overdata;
    delete[] entries;
}

int Newsgroup::num_articles()
{
    return num_entries;
}

int Newsgroup::first_index()
{
    int ind;

    lock();
    ind = (num_entries > 0) ? entries[0].index : last + 1;
    unlock();
    return ind;
}

int Newsgroup::last_index()
{
    return last;
}

int Newsgroup::lookup(int index, int *partition, time_t *bucket_num,
		      long *offset)
{
    int i;

    /* Look for the index in the table. */
    lock();
    i = find_entry(index);
    if (i < 0) {
	unlock();
	return -1;
    }

    *partition = entries[i].partition;
    *bucket_num = entries[i].bucket_num;
    *offset = entries[i].offset;
    unlock();
    return 0;
}

int Newsgroup::next(int *index, int *partition, time_t *bucket_num,
		    long *offset)
{
    int i;

    /* Look for the index in the table. */ 
    lock();
    i = find_entry(*index);
    if (i < 0 || i == num_entries - 1) {
	unlock();
	return -1;
    }

    /* Copy the next entry into the arguments, and cache the entry
     * number. */
    i++;
    *index = entries[i].index;
    *partition = entries[i].partition;
    *bucket_num = entries[i].bucket_num;
    *offset = entries[i].offset;
    unlock();
    return 0;
}

int Newsgroup::prev(int *index, int *partition, time_t *bucket_num,
		    long *offset)
{
    int i;

    /* Look for the index in the table. */
    lock();
    i = find_entry(*index);
    if (i <= 0) {
	unlock();
	return -1;
    }

    /* Copy the previous entry into the arguments, and cache the entry
     * number. */
    i--;
    *index = entries[i].index;
    *partition = entries[i].partition;
    *bucket_num = entries[i].bucket_num;
    *offset = entries[i].offset;
    unlock();
    return 0;
}

int Newsgroup::first(int *index, int *partition, time_t *bucket_num,
		     long *offset)
{
    /* Return the first entry, unless the group is empty. */
    lock();
    if (!num_entries) {
	unlock();
	return -1;
    }
    *index = entries[0].index;
    *partition = entries[0].partition;
    *bucket_num = entries[0].bucket_num;
    *offset = entries[0].offset;
    unlock();
    return 0;
}

void Newsgroup::add_entry(int index, int partition, time_t bucket_num,
			  long offset, const char *overdata)
{
    lock();
    add_index_entry(index, partition, bucket_num, offset, overdata);
    unlock();
}

void Newsgroup::eliminate_bucket(time_t bucket_num)
{
    int i, j;

    /* Eliminate the entries in the newsgroup whose bucket is equal to
     * bucket_num. */
    lock();
    j = 0;
    for (i = 0; i < num_entries; i++) {
	if (entries[i].bucket_num != bucket_num)
	    entries[j++] = entries[i];
    }
    num_entries = j;
    unlock();
}

void Newsgroup::eliminate_index(int index)
{
    int i;

    /* Eliminate the entry in the newsgroup whose index is index. */
    lock();
    for (i = 0; i < num_entries; i++) {
	if (entries[i].index == index)
	    break;
    }
    for (; i < num_entries - 1; i++)
	entries[i] = entries[i + 1];
    num_entries--;
    unlock();
}

void Newsgroup::send_overview(Connection *conn, int low, int high)
{
    int i, ind;

    lock();
    if (low == high) {
	/* If it's a single article, use a binary search to do it quickly. */
	i = find_entry(low);
	if (i >= 0)
	    conn->writef("%d\t%s\r\n", low, entries[i].overdata);
    } else {
	/* If it's anything else, just iterate through the whole list. */
	for (i = 0; i < num_entries; i++) {
	    ind = entries[i].index;
	    if (ind >= low && (high == -1 || ind <= high))
		conn->writef("%d\t%s\r\n", ind, entries[i].overdata);
	}
    }
    unlock();
}

void Newsgroup::send_headers(Connection *conn, const char *name, int namelen,
			     int low, int high, Spool *spool)
{
    int fnum, ind, i;

    lock();
    fnum = config->find_in_schema(name, namelen);
    if (low == high) {
	/* Find a single article quickly with a binary search. */
	i = find_entry(low);
	if (i >= 0)
	    send_header(conn, &entries[i], fnum, spool, name, namelen);
    } else {
	/* Otherwise, just iterate through all of the entries. */
	for (i = 0; i < num_entries; i++) {
	    ind = entries[i].index;
	    if (ind >= low && (high == -1 || ind <= high))
		send_header(conn, &entries[i], fnum, spool, name, namelen);
	}
    }
    unlock();
}

/* Assumes lock. */
void Newsgroup::add_index_entry(int index, int partition, time_t bucket_num,
				long offset, const char *overdata)
{
    Index_entry *new_entries, *entry;
    int new_size;

    /* Expand the table if we have to. */
    if (num_entries >= entries_size) {
	new_size = adjust_size(num_entries + 1, MIN_POWER, DELTA);
	new_entries = new Index_entry[new_size];
	memcpy(new_entries, entries, num_entries * sizeof(Index_entry));
	delete entries;
	entries = new_entries;
	entries_size = new_size;
    }

    /* Entries will usually be added sequentially, but we need to make sure. */
    entry = entries + num_entries;
    while (entry > entries && (entry - 1)->index > index) {
	*entry = *(entry - 1);
	entry--;
    }

    entry->index = index;
    entry->partition = partition;
    entry->bucket_num = bucket_num;
    entry->offset = offset;
    entry->overdata = strduplicate(overdata);
    num_entries++;
    if (index > last)
	last = index;
}

/* Assumes read lock. */
int Newsgroup::find_entry(int index)
{
    int low, high, mid;

    /* Perform a simple binary search.  Note that there's no equality
     * test in the loop; this is somewhat faster on average.  Be
     * careful of infinite loops if you modify this search, though;
     * make sure to consider the case where high == low + 1. */
    low = 0;
    high = num_entries - 1;
    while (high > low) {
	mid = (low + high) / 2;
	if (index > entries[mid].index)
	    low = mid + 1;
	else
	    high = mid;
    }

    return (entries[low].index == index) ? low : -1;
}

const char *find_overview_field(const char *overdata, int fnum, int *len)
{
    const char *p = overdata, *q;

    while (fnum) {
	p = strchr(p, '\t');
	p = (p) ? p + 1 : "";
	fnum--;
    }
    q = strchr(p, '\t');
    *len = (q) ? q - p : strlen(p);
    return p;
}

static void send_header(Connection *conn, Index_entry *entry, int fnum,
			Spool *spool, const char *name, int namelen)
{
    String buf;
    const char *field;
    int len;

    if (fnum >= 0) {
	field = find_overview_field(entry->overdata, fnum, &len);
	conn->writef("%d %C\r\n", entry->index, field, len);
    } else if (spool->find_header(entry->partition, entry->bucket_num,
				  entry->offset, name, namelen, &buf) >= 0) {
	conn->writef("%d %S\r\n", entry->index, &buf);
    }
}

