/* Nntp.c: Implementation of the Nntp protocol-handler class. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Nntp.h"
#include "Connection.h"
#include "String.h"
#include "Spool.h"
#include "Article.h"
#include "Newsgroup.h"
#include "Config.h"
#include "Wildcard_list.h"

extern "C" {
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
}

typedef void (Nntp::*Handler)(const char *s);

static int get_range(const char *s, int *low, int *high);

Nntp::Nntp(Spool *spool_arg, Connection *conn_arg, Config *config_arg)
{
    spool = spool_arg;
    conn = conn_arg;
    config = config_arg;
    cur_group = NULL;
    cur_article = NULL;
    cur_index = -1;
    running = 1;
    post_access = config->get_post_access(conn->name(), conn->ip());
    read_access = config->get_read_access(conn->name(), conn->ip());
    transfer_access = config->get_transfer_access(conn->name(), conn->ip());
}

Nntp::~Nntp()
{
    if (cur_group)
	spool->release_group(cur_group);
    if (cur_article)
	spool->release_article(cur_article);
    delete conn;
    delete post_access;
    delete read_access;
    delete transfer_access;
}

void Nntp::run()
{
    String line;
    const char *s, *p, *q;
    bool whether;

    /* If you're not a cop, you're little people. */
    if (read_access->is_degenerate(0, &whether) && !whether &&
	transfer_access->is_degenerate(0, &whether) && !whether) {
	conn->write("502 You have no permission to talk.  Goodbye.\r\n");
	return;
    }

    if (post_access->is_degenerate(0, &whether) && !whether)
	conn->write("201 Threaded news server ready (no post access)\r\n");
    else
	conn->write("200 Threaded news server ready\r\n");
    while (running && conn->read_line_clear(&line) >= 0) {

	/* Strip trailing whitespace from the line. */
	s = line.chars();
	p = s + line.length() - 1;
	while (p >= s && isspace(*p))
	    p--;
	line.truncate(p + 1 - s);

	log("NNTP %s: %s", conn->name(), s);

	/* Ignore blank lines. */
	if (line.length() == 0)
	    continue;

	/* Find command (s to p) and remaining text (q). */
	s = find_word(line.chars());
	p = word_end(s);
	q = find_word(p);

	/* Find command associated with the starting word. */
	if (handle_command(s, p - s, q) < 0)
	    conn->write("500 Command not recognized\r\n");
    }
}

int Nntp::handle_command(const char *s, int len, const char *rest)
{
    static struct entry {
	char *name;
	void (Nntp::*func)(const char *s);
    } tab[] = {
	{ "article",	Nntp::article },
	{ "body",	Nntp::body },
	{ "group",	Nntp::group },
	{ "head",	Nntp::head },
	{ "help",	Nntp::help },
	{ "ihave",	Nntp::ihave },
	{ "last",	Nntp::last },
	{ "list",	Nntp::list },
	{ "newgroups",	Nntp::newgroups },
	{ "newnews",	Nntp::newnews },
	{ "next",	Nntp::next },
	{ "post",	Nntp::post },
	{ "quit",	Nntp::quit },
	{ "slave",	Nntp::slave },
	{ "stat",	Nntp::stat },
	{ "xhdr",	Nntp::xhdr },
	{ "xover",	Nntp::xover }
    };
    int i;

    for (i = 0; i < sizeof(tab) / sizeof(struct entry); i++) {
	if (strlen(tab[i].name) == len && strnccmp(tab[i].name, s, len) == 0) {
	    (this->*tab[i].func)(rest);
	    return 0;
	}
    }

    return -1;
}

void Nntp::article(const char *s)
{
    fetch_article(s, 1, 1);
}

void Nntp::body(const char *s)
{
    fetch_article(s, 0, 1);
}

void Nntp::group(const char *s)
{
    Newsgroup *group;
    int num, first, last;

    /* Check reader access to the group. */
    if (!read_access->match(s, 0)) {
	conn->writef("502 No access to group \"%s\"\r\n", s);
	return;
    }

    /* Look up the group in the spool. */
    group = spool->find_group(s);
    if (!group) {
	conn->write("411 No such newsgroup\r\n");
	return;
    }

    /* Release the current article, if we have one. */
    if (cur_article) {
	spool->release_article(cur_article);
	cur_article = NULL;
	cur_index = -1;
    }

    /* Get information about the group. */
    num = group->num_articles();
    first = group->first_index();
    last = group->last_index();

    /* Set the current group to this group. */
    if (cur_group)
	spool->release_group(cur_group);
    cur_group = group;

    /* Respond with the number of articles in the group and the index
     * of the first and last article. */
    conn->writef("211 %d %d %d %s group selected\r\n", num, first, last, s);
}

void Nntp::head(const char *s)
{
    fetch_article(s, 1, 0);
}

void Nntp::help(const char *s)
{
    conn->write("100 Help text follows\r\n");
    conn->write("  ARTICLE [message-id|number]\r\n");
    conn->write("  BODY [message-id|number]\r\n");
    conn->write("  GROUP newsgroup\r\n");
    conn->write("  HELP\r\n");
    conn->write("  IHAVE messsage-id\r\n");
    conn->write("  LAST\r\n");
    conn->write("  LIST active|newsgroups|distributions|schema\r\n");
    conn->write("  NEWGROUPS date time [GMT] [<distributions>]\r\n");
    conn->write("  NEWNEWS newsgroups date time [GMT] [<distributions>]\r\n");
    conn->write("  NEXT\r\n");
    conn->write("  POST\r\n");
    conn->write("  QUIT\r\n");
    conn->write("  SLAVE\r\n");
    conn->write("  STAT [message-id|number]\r\n");
    conn->write("  XHDR field [message-id|range]\r\n");
    conn->write("  XOVER range\r\n");
    conn->write(".\r\n");
}

void Nntp::ihave(const char *s)
{
    /* NNTP status numbers for POST_OK, POST_REJECTED, and POST_FAILED. */
    static int statuses[] = { 235, 437, 437 };
    String line, article;
    Post_status result;

    if (transfer_access->num_entries() == 0) {
	conn->write("502 You don't have access to transfer articles.\r\n");
	return;
    }

    /* Check the message ID for syntactic correctness. */
    if (verify_message_id(s) < 0) {
	conn->write("501 message ID not correct\r\n");
	return;
    }

    if (!spool->article_wanted(s)) {
	conn->write("435 article not wanted - do not send it\r\n");
	return;
    }

    conn->write("335 send article to be transferred."
		"  End with <CR-LF>.<CR-LF>\r\n");

    /* Read the article from the client. */
    while (conn->read_line(&line) >= 0) {
	/* If the line is simply a period, the article is done. */
	if (line.length() == 1 && *line.chars() == '.')
	    break;

	article.append(&line);
        article.append('\n');
	line.clear();
    }

    spool_message.clear();
    result = spool->ihave_article(s, &article, transfer_access,
				  &spool_message);

    conn->writef("%d %s\r\n", statuses[result], spool_message.chars());
}

void Nntp::last(const char *s)
{
    Article *article;

    /* Make sure there's a current group and article. */
    if (!cur_group) {
	conn->write("412 No newsgroup selected\r\n");
	return;
    }
    if (!cur_article && select_first_article() < 0) {
	conn->write("420 No current article has been selected.\r\n");
	return;
    }

    /* Get the next article in the current group. */
    article = spool->prev_article(cur_group, &cur_index);
    if (!article) {
	conn->write("422 No previous article in this group.\r\n");
	return;
    }

    /* Set the current article and index. */
    spool->release_article(cur_article);
    cur_article = article;

    /* Return success to the client. */
    message_id.clear();
    article->get_message_id(&message_id);
    conn->writef("223 %d %s article retrieved - request text separately.\r\n",
		 cur_index, message_id.chars());
}

void Nntp::list(const char *s)
{
    Wildcard_list<bool> *access = NULL;
    bool whether;

    /* Sites with transfer access can find out about all groups. */
    if (transfer_access->is_degenerate(0, &whether) && !whether)
	access = read_access;

    if (!*s || strccmp(s, "active") == 0) {
	conn->write("215 Newsgroups in form \"group high low flags\"\r\n");
	spool->send_grouplist(conn, access);
	conn->write(".\r\n");
    } else if (strccmp(s, "newsgroups") == 0) {
	conn->write("215 Descriptions in form \"group description\"\r\n");
	spool->send_group_descriptions(conn, access);
	conn->write(".\r\n");
    } else if (strccmp(s, "distributions") == 0) {
	conn->write("500 Command not implemented\r\n");
    } else if (strccmp(s, "distrib.pats") == 0) {
	conn->write("500 What's a distrib.pats file?\r\n");
    } else if (strccmp(s, "schema") == 0 ||
	       strccmp(s, "overview.fmt") == 0) {
	conn->write("215 Order of fields in overview database\r\n");
	spool->send_schema(conn);
	conn->write(".\r\n");
    } else {
	conn->write("501 Bad command use\r\n");
    }
}

void Nntp::newgroups(const char *s)
{
    const char *date, *end_date, *time, *end_time;
    const char *gmt, *end_gmt, *dist, *end_dist;
    time_t t;
    Wildcard_list<bool> *access = NULL;
    bool whether;

    /* Sites with transfer access can find out about all groups. */
    if (transfer_access->is_degenerate(0, &whether) && !whether)
	access = read_access;

    /* Find date, time, optional GMT, and optional distribution. */
    end_date = word_end(date = s);
    end_time = word_end(time = find_word(end_date));
    end_gmt  = word_end(gmt  = find_word(end_time));
    end_dist = word_end(dist = find_word(end_gmt));
    if (*end_gmt == 0 && *gmt == '<') {
	dist = gmt;
	gmt = end_gmt;
    }

    /* First two words must be six characters each, gmt must be "GMT" if
     * specified, distributions must begin with a < and end with a >. */
    if (*end_dist != 0 || end_date - date != 6 || end_time - time != 6
	|| (*gmt && (end_gmt - gmt != 3 || strnccmp(gmt, "GMT", 3) != 0))
	|| (*dist && (*dist != '<' || end_dist[-1] != '>'))) {
	conn->write("501 Bad command use\r\n");
	return;
    }

    /* Parse date and time into Unix time. */
    t = parse_nntp_date(date, time, gmt);
    if (t == -1) {
	conn->write("501 Bad command use\r\n");
	return;
    }

    conn->write("231 Newsgroups in form \"group high low flags\"\r\n");
    spool->send_new_groups(conn, t, dist, access);
    conn->write(".\r\n");
}

void Nntp::newnews(const char *s)
{
    conn->write("400 Command not implemented\r\n");
}

void Nntp::next(const char *s)
{
    Article *article;

    /* Make sure there's a current group and article. */
    if (!cur_group) {
	conn->write("412 No newsgroup selected.\r\n");
	return;
    }
    if (!cur_article && select_first_article() < 0) {
	conn->write("420 No current article has been selected.\r\n");
	return;
    }

    /* Get the next article in the current group. */
    article = spool->next_article(cur_group, &cur_index);
    if (!article) {
	conn->write("421 No next article in this group.\r\n");
	return;
    }

    /* Set the current article and index. */
    spool->release_article(cur_article);
    cur_article = article;

    /* Return success to the client. */
    message_id.clear();
    article->get_message_id(&message_id);
    conn->writef("223 %d %s article retrieved - request text separately.\r\n",
		 cur_index, message_id.chars());
}

void Nntp::post(const char *s)
{
    /* NNTP status numbers for POST_OK, POST_REJECTED, and POST_FAILED. */
    String line, article;
    bool whether;

    if (post_access->is_degenerate(0, &whether) && !whether) {
	conn->write("440 Posting not allowed.\r\n");
	return;
    }

    conn->write("340 Send article to be posted."
		"  End with <CR-LF>.<CR-LF>\r\n");

    /* Read the article from the client. */
    while (conn->read_line(&line) >= 0) {
	/* If the line is simply a period, the article is done. */
	if (line.length() == 1 && *line.chars() == '.')
	    break;

	article.append(&line);
        article.append('\n');
	line.clear();
    }

    spool_message.clear();
    if (spool->post_article(&article, conn->name(), post_access,
			    &spool_message) < 0)
	conn->writef("441 %s\r\n", spool_message.chars());
    else
	conn->write("240 Article posted okay.\r\n");
}

void Nntp::quit(const char *s)
{
    conn->write("205 Closing connection.\r\n");
    running = 0;
}

void Nntp::slave(const char *s)
{
    conn->write("202 Slave status noted.\r\n");
}

void Nntp::stat(const char *s)
{
    /* Special case stat of a message-ID. */
    if (*s == '<') {
	if (spool->article_exists(s))
	    conn->writef("223 0 %s article exists in spool.\r\n", s);
	else
	    conn->write("430 No such article.\r\n");
    } else {
	fetch_article(s, 0, 0);
    }
}

void Nntp::xhdr(const char *s)
{
    int low, high;
    const char *p, *q;

    /* Find header name and range (s to p, low, high). */
    p = word_end(s);
    q = find_word(p);
    if (*q == '<') {
	spool_message.clear();
	if (spool->find_header(q, s, p - s, &spool_message) < 0) {
	    conn->write("430 No such article\r\n");
	} else {
	    conn->writef("221 0 %C header of article %s\r\n", s, p - s, q);
	    conn->writef("%s %S\r\n.\r\n", q, &spool_message);
	}
	return;
    }
    if (get_range(q, &low, &high) < 0) {
	conn->write("501 Incorrect range syntax\r\n");
    } else if (!cur_group) {
	conn->write("412 No newsgroup has been selected.\r\n");
    } else {
	if (low == -1 && high == -1) {
	    if (!cur_article && cur_group->num_articles() > 0)
		cur_article = spool->first_article(cur_group, &cur_index);
	    if (!cur_article) {
		conn->write("420 No current article has been selected.\r\n");
		return;
	    }
	    low = high = cur_index;
	}
	conn->writef("221 %C fields follow.\r\n", s, p - s);
	cur_group->send_headers(conn, s, p - s, low, high, spool);
	conn->write(".\r\n");
    }
}

void Nntp::xover(const char *s)
{
    int low, high;

    if (get_range(s, &low, &high) < 0) {
	conn->write("501 Incorrect range syntax\r\n");
    } else if (!cur_group) {
	conn->write("412 No newsgroup has been selected.\r\n");
    } else {
	if (low == -1 && high == -1) {
	    if (!cur_article && cur_group->num_articles() > 0)
		cur_article = spool->first_article(cur_group, &cur_index);
	    if (!cur_article) {
		conn->write("420 No current article has been selected.\r\n");
		return;
	    }
	    low = high = cur_index;
	}
	conn->write("224 data follows.\r\n");
	cur_group->send_overview(conn, low, high);
	conn->write(".\r\n");
    }
}

void Nntp::fetch_article(const char *s, int head, int body)
{
    static char *responses[2][2] = {
	{ "223 %d %s article retrieved - request text separately.\r\n",
	  "222 %d %s article retrieved - body follows.\r\n" },
	{ "221 %d %s article retrieved - head follows.\r\n",
	  "220 %d %s article retrieved - head and body follow.\r\n" }
    };
    String buf;
    Article *article;
    int index = 0;
    const char *message_id;
    bool whether;

    /* Sites with only transfer access can't do more than stat a message-ID. */
    if (read_access->is_degenerate(0, &whether) && !whether) {
	conn->write("502 Reading articles not allowed.\r\n");
	return;
    }

    /* Preemptively check a couple of error conditions. */
    if (*s != '<' && !cur_group) {
	conn->write("412 No newsgroup has been selected.\r\n");
	return;
    }
    if (!*s && !cur_article && select_first_article() < 0) {
	conn->write("420 No current article has been selected.\r\n");
	return;
    }

    if (*s == '<') {
	/* It's a message ID. */
	article = spool->find_article(s);
	if (!article) {
	    conn->write("430 No such article.\r\n");
	    return;
	}
    } else if (!*s) {
	article = cur_article;
	index = cur_index;
    } else {
	/* It's an article index in the current group. */
	index = atol(s);
	article = spool->find_article(cur_group, index);
	if (!article) {
	    conn->write("423 Bad article number\r\n");
	    return;
	}

	/* Set the current article and index to this article. */
	if (cur_article)
	    spool->release_article(cur_article);
	cur_article = article;
	cur_index = index;
    }

    if (*s != '<') {
	article->get_message_id(&buf);
	message_id = buf.chars();
    } else {
	message_id = s;
    }
	
    conn->writef(responses[head][body], index, message_id);
    if (head)
	article->write_head(conn);
    if (head && body)
	conn->write("\r\n", 2);
    if (body)
	article->write_body(conn);
    if (head || body)
	conn->write(".\r\n", 3);

    /* If the article was specified by message-ID, discard it. */
    if (*s == '<')
	spool->release_article(article);
}

int Nntp::select_first_article()
{
    if (cur_group->num_articles() > 0)
	cur_article = spool->first_article(cur_group, &cur_index);
    return (cur_article) ? 0 : -1;
}

/* Parse a range, returning 0 if successful, less than zero if not.  Accept
 * N, -N, N-, N-N, or a blank string, where N is a positive integer.  In the
 * returned bounds, -1 indicates an unspecified value. */
static int get_range(const char *s, int *low, int *high)
{
    if (!*s) {
	*low = *high = -1;
	return 0;
    }

    /* Get the low bound if it's specified. */
    if (isdigit(*s)) {
	*low = atoi(s);
	while (isdigit(*s))
	    s++;
    } else {
	*low = -1;
    }

    /* Get the upper bound if it's specified. */
    s = find_word(s);
    if (*s == '-') {
	s = find_word(s + 1);
	*high = (isdigit(*s)) ? atoi(s) : -1;
    } else {
	*high = *low;
    }

    /* If *low = *high = -1, the range was syntactically incorrect. */
    return (*low == -1 && *high == -1) ? -1 : 0;
}

