/* Bucket.cc: Managing the pool of article buckets. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Pool.h"
#include "Atomic_file.h"
#include "Bucket.h"
#include "Config.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
}

Pool::Pool(Spool *spool_arg, const Config *config)
{
    DIR *dir;
    struct dirent entry, *entp;
    String name;
    struct stat statbuf;

    spool = spool_arg;

    /* Get directory and checkpoint file pathnames. */
    config->get_pool_directory(&dirname);
    format(&name, "%Scheckpoint", &dirname);
    checkpoint = new Atomic_file(name.chars());
    make_directories_in_path(dirname.chars());

    /* Find all of the buckets in all of the partitions. */
    for (npart = 0; ; npart++) {
	format_clear(&name, "%S%d", &dirname, npart);
	if (stat(name.chars(), &statbuf) < 0 || !S_ISDIR(statbuf.st_mode))
	    break;
	partitions.ensure(npart);
	partitions[npart].buckets = NULL;
	dir = opendir(name.chars());
	while (readdir_r(dir, &entry, &entp) >= 0 && entp) {
	    format_clear(&name, "%l", atol(entry.d_name));
	    if (strcmp(name.chars(), entry.d_name) != 0)
		continue;
	    format_clear(&name, "%S%d/%s", &dirname, npart, entry.d_name);
	    open_and_add_bucket(npart, name.chars(), atol(entry.d_name));
	}
	closedir(dir);
    }

    /* Make sure there's at least one partition. */
    if (npart == 0) {
	format_clear(&name, "%S0", &dirname);
	if (mkdir(name.chars(), S_IRWXU | S_IRWXG | S_IROTH) < 0)
	    die("Can't create initial partition %s", name.chars());
	partitions.ensure(0);
	partitions[0].buckets = NULL;
	npart = 1;
    }
}

/* Doesn't unlink from the filesystem; all buckets should be released
 * before this destructor is used. */
Pool::~Pool()
{
    int i;
    Bucket_list *next;

    for (i = 0; i < npart; i++) {
	while (partitions[i].buckets) {
	    next = partitions[i].buckets->next;
	    delete partitions[i].buckets;
	    partitions[i].buckets = next;
	}
    }
}

Bucket *Pool::find_bucket(int partition, time_t bucket_number)
{
    Bucket_list *entry;
    Bucket *bucket;

    /* If we find the entry, bump its refcount so we don't throw away
     * the bucket, and return its bucket.  Otherwise, return NULL. */
    lock();
    entry = search_buckets(partition, bucket_number);
    if (entry) {
	bucket = &entry->bucket;
	entry->refs++;
    } else {
	bucket = NULL;
    }
    unlock();
    return bucket;
}

Bucket *Pool::find_or_create_bucket(int partition, time_t bucket_number)
{
    Bucket *bucket;
    Bucket_list *entry;
    String name;
    FILE *fp;

    /* See if the bucket already exists. */
    lock();
    entry = search_buckets(partition, bucket_number);
    if (entry) {
	bucket = &entry->bucket;
	entry->refs++;
	unlock();
	return bucket;
    }

    /* Create a new bucket file. */
    format(&name, "%S%d/%l", &dirname, partition, bucket_number);
    fp = fopen(name.chars(), "w+");
    if (!fp) {
	warn("Can't create bucket file %s: %m", name.chars());
	unlock();
	return NULL;
    }

    /* Chain the new bucket into the list. */
    entry = new Bucket_list(fp, name.chars(), partition, bucket_number);
    entry->next = partitions[partition].buckets;
    partitions[partition].buckets = entry;

    /* Return the new bucket. */
    entry->refs++;
    bucket = &entry->bucket;
    unlock();
    return bucket;
}

void Pool::release_bucket(Bucket *bucket)
{
    Bucket_list **entry_ptr, *entry;
    int i;

    lock();

    /* Find the entry whose bucket is bucket. */
    for (i = 0; i < npart; i++) {
	entry_ptr = &partitions[i].buckets;
	for (; *entry_ptr; entry_ptr = &(*entry_ptr)->next) {
	    entry = *entry_ptr;
	    if (&entry->bucket == bucket)
		break;
	}
	if (*entry_ptr)
	    break;
    }
    if (!*entry_ptr) {
	/* Something is amiss. */
	crash("Released bucket <%d, %l> is not in the bucket list.",
	      bucket->partition(), (long) bucket->number());
    }

    /* Release this list entry. */
    release_list_entry(entry_ptr);

    unlock();
}

/* Stub. */
int Pool::decide_partition(int num_bytes)
{
    return (rand() % npart);
}

/* No other thread should touch the pool while this is taking place. */
int Pool::write_checkpoint_record()
{
    FILE *fp;
    Bucket_list *entry;
    String line;
    int count, i;

    fp = checkpoint->open_new();
    if (!fp)
	return -1;

    /* Write out the new file. */
    for (i = 0; i < npart; i++) {
	for (entry = partitions[i].buckets; entry; entry = entry->next) {
	    line.clear();
	    format(&line, "%d %l %l\n", entry->bucket.partition(),
		   (long) entry->bucket.number(),
		   entry->bucket.current_offset());
	    count = fwrite(line.chars(), sizeof(char), line.length(), fp);
	    if (count != line.length()) {
		warn("Can't write data to checkpoint record: %m");
		fclose(fp);
		return -1;
	    }
	}
    }
    fclose(fp);

    return checkpoint->update();
}

/* No other threads should touch the pool while this is taking place. */
int Pool::consistency_check()
{
    Bucket_list *entry;
    FILE *fp;
    char buf[100];
    time_t bucket_num;
    long offset;
    const char *p;
    int i, partition;

    /* Start by setting all consistent_at offsets to 0, to account for
     * buckets with no checkpoint record. */
    for (i = 0; i < npart; i++) {
	for (entry = partitions[i].buckets; entry; entry = entry->next)
	    entry->consistent_at = 0;
    }

    /* Read in the checkpoint record and set the consistent_at offsets. */
    fp = checkpoint->open_read();
    if (fp) {
	while (fgets(buf, sizeof(buf), fp)) {
	    p = buf;
	    partition = atoi(p);
	    bucket_num = atol(next_word(&p));
	    offset = atol(next_word(&p));
	    entry = search_buckets(partition, bucket_num);
	    if (entry)
		entry->consistent_at = offset;
	}
	fclose(fp);
    }

    /* Re-index any articles past the consistent_at offsets. */
    for (i = 0; i < npart; i++) {
	for (entry = partitions[i].buckets; entry; entry = entry->next) {
	    if (entry->bucket.consistency_check(spool,
						entry->consistent_at) < 0)
		return -1;
	}
    }

    /* Write out a checkpoint record, since we're now consistent. */
    return write_checkpoint_record();
}

time_t Pool::first_bucket()
{
    time_t least = 0;
    Bucket_list *b;
    int i, found = 0;

    for (i = 0; i < npart; i++) {
	for (b = partitions[i].buckets; b; b = b->next) {
	    if (!found || b->bucket.number() < least) {
		least = b->bucket.number();
		found = 1;
	    }
	}
    }
    return (found) ? least : -1;
}

void Pool::expire_bucket(time_t bucket_num)
{
    Bucket_list **entry_ptr;
    int i;

    /* Release the list entries for any buckets numbered bucket_num. */
    lock();
    for (i = 0; i < npart; i++) {
	entry_ptr = &partitions[i].buckets;
	for (; *entry_ptr; entry_ptr = &(*entry_ptr)->next) {
	    if ((*entry_ptr)->bucket.number() == bucket_num) {
		release_list_entry(entry_ptr);
		break;
	    }
	}
    }
    unlock();
}

/* Stub. */
int Pool::mark_cancelled(int partition, time_t bucket_num, long offset)
{
    return 0;
}

/* Assumes lock. */
void Pool::release_list_entry(Bucket_list **entry_ptr)
{
    Bucket_list *entry = *entry_ptr;

    entry->refs--;
    if (entry->refs == 0) {
	String filename;

	/* The refcount hit zero, so this bucket is due to be removed.
	 * Unlink the bucket from the file system and delete it.  Later
	 * we need to worry about moving open articles from this bucket
	 * into temporary files. */
	format(&filename, "%S%d/%l", &dirname, entry->bucket.partition(),
	       (long) entry->bucket.number());
	unlink(filename.chars());

	/* Remove this entry from the list. */
	*entry_ptr = entry->next;
	delete entry;
    }
}

void Pool::open_and_add_bucket(int partition, const char *name,
			       time_t bucket_number)
{
    FILE *fp;
    Bucket_list *entry;

    fp = fopen(name, "r+");
    if (fp == NULL)
	die("Could not open bucket %s: %m", name);

    entry = new Bucket_list(fp, name, partition, bucket_number);
    entry->next = partitions[partition].buckets;
    partitions[partition].buckets = entry;
}

/* Assumes lock or initialization phase. */
Bucket_list *Pool::search_buckets(int partition, time_t bucket_number)
{
    Bucket_list *entry;

    if (partition < 0 || partition >= npart)
	return NULL;

    /* Find the entry whose bucket's number is bucket_number. */
    for (entry = partitions[partition].buckets; entry; entry = entry->next) {
	if (entry->bucket.number() == bucket_number)
	    return entry;
    }
    return NULL;
}

Bucket_list::Bucket_list(FILE *fp, const char *name,
			 int partition, time_t bucket_number)
: bucket(fp, name, partition, bucket_number)
{
    refs = 1;
}

