/* Spool.cc: Implementation of the news spool. */

/* See ../doc/INTERNALS for notes on the spool organization. */

#include "threads.h"
#include "util.h"
#include "platform.h"
#include "error.h"

#include "Spool.h"
#include "Index.h"
#include "Groups.h"
#include "Pool.h"
#include "Bucket.h"
#include "Newsgroup.h"
#include "Article.h"
#include "Config.h"
#include "Connection.h"

extern "C" {
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
}

static void *periodic_checkpoints(void *arg);
static void *periodic_expire(void *arg);

Spool::Spool(const Config *config_arg)
{
    config = config_arg;
    index = new Index(config);
    groups = new Groups(config);
    pool = new Pool(this, config);
    feeds = new Feeds(this, config);
    config->get_server_name(&server_name);
    throttle_message = NULL;

    /* Make sure the spool is consistent, or disallow writing. */
    if (consistency_check() < 0)
	throttle("Consistency check failed.");

    start_thread(periodic_checkpoints, this);
    start_thread(periodic_expire, this);
}

Spool::~Spool()
{
    delete index;
    delete groups;
    delete pool;
    delete feeds;
}

Newsgroup *Spool::find_group(const char *name)
{
    return groups->find_group(name);
}

Article *Spool::find_article(const char *id)
{
    int partition;
    time_t bucket_number;
    long offset;

    /* Find the bucket and offset of the article. */
    if (index->lookup(id, &partition, &bucket_number, &offset) < 0)
	return NULL;

    return get_article(partition, bucket_number, offset);
}

Article *Spool::find_article(Newsgroup *group, int index)
{
    int partition;
    time_t bucket_number;
    long offset;

    if (group->lookup(index, &partition, &bucket_number, &offset) < 0)
	return NULL;

    return get_article(partition, bucket_number, offset);
}

Article *Spool::next_article(Newsgroup *group, int *index)
{
    int partition;
    time_t bucket_number;
    long offset;
    Article *article;
    int index_tmp;

    index_tmp = *index;
    if (group->next(&index_tmp, &partition, &bucket_number, &offset) < 0)
	return NULL;

    article = get_article(partition, bucket_number, offset);
    if (article)
	*index = index_tmp;
    return article;
}

Article *Spool::prev_article(Newsgroup *group, int *index)
{
    int partition;
    time_t bucket_number;
    long offset;
    Article *article;
    int index_tmp;

    index_tmp = *index;
    if (group->prev(&index_tmp, &partition, &bucket_number, &offset) < 0)
	return NULL;

    article = get_article(partition, bucket_number, offset);
    if (article)
	*index = index_tmp;
    return article;
}

Article *Spool::first_article(Newsgroup *group, int *index)
{
    int partition;
    time_t bucket_number;
    long offset;
    Article *article;
    int index_tmp;

    index_tmp = *index;
    if (group->first(&index_tmp, &partition, &bucket_number, &offset) < 0)
	return NULL;

    article = get_article(partition, bucket_number, offset);
    if (article)
	*index = index_tmp;
    return article;
}

Article *Spool::get_article(int partition, time_t bucket_num, long offset)
{
    Bucket *bucket;
    Article *article;

    /* Find the bucket corresponding to the given bucket number. */
    bucket = pool->find_bucket(partition, bucket_num);
    if (bucket == NULL)
	return NULL;

    /* Get the article from the bucket, release the bucket, and return. */
    article = bucket->get_article(offset);
    pool->release_bucket(bucket);
    return article;
}

int Spool::article_exists(const char *id)
{
    int partition;
    time_t bucket_number;
    long offset;

    /* Perform an index lookup, return true if it succeeds. */
    return (index->lookup(id, &partition, &bucket_number, &offset) >= 0);
}

void Spool::release_group(Newsgroup *group)
{
    groups->release_group(group);
}

void Spool::release_article(Article *article)
{
    Bucket *bucket = article->bucket();

    bucket->release_article(article);
}

int Spool::article_wanted(const char *id)
{
    int partition;
    time_t bucket_number;
    long offset;

    /* Perform an index lookup, and return true if it's not there. */
    return (index->lookup(id, &partition, &bucket_number, &offset) < 0);
}

/* The ihave command has to:
 * - Verify that all required headers on the article exist.
 * - Verify that the Message-ID field is the same as the one in the ihave
 *   command.
 * - Verify that the Newsgroups line doesn't contain any invalid
 *   characters.
 * - Verify that we want at least one of the specified distributions.
 * - Store and propagate the article. */
Post_status Spool::ihave_article(const char *id, const String *article,
				 const Wildcard_list<bool> *access,
				 String *message)
{
    static const char *required[] = { "From", "Date", "Newsgroups", "Subject",
				      "Message-ID", "Path", NULL };
    Headers headers;
    int num_lines, num_bytes;
    const char **req, *body, *message_id, *newsgroups, *dist;

    /* Fill headers with header lines. */
    body = headers.add_headers(article->chars());
    if (!body) {
	message->append("Malformatted headers.");
	return POST_REJECTED;
    }

    /* Verify that all required headers exist. */
    for (req = required; *req; req++) {
	if (headers[*req] == NULL) {
	    format(message, "Missing header %s", req);
	    return POST_REJECTED;
	}
    }

    /* Verify that the message ID is equivalent to id. */
    message_id = headers["Message-ID"];
    if (!message_id_equal(message_id, id)) {
	message->append("Message-ID not equivalent to ihave ID.");
	return POST_REJECTED;
    }

    /* Verify that the newsgroup line contains no invalid characters. */
    newsgroups = headers["Newsgroups"];
    if (strpbrk(newsgroups, ":/ \f\t\n") != NULL) {
	message->append("Invalid characters in newsgroup line.");
	return POST_REJECTED;
    }

    /* Verify that we accept at least one distribution. */
    dist = headers["Distribution"];
    if (dist && config->any_distribution_accepted(dist) < 0) {
	format(message, "Server does not accept distributions \"%s\".", dist);
	return POST_REJECTED;
    }

    /* If any group is moderated, verify that there is an Approved header. */
    if (groups->any_groups_moderated(newsgroups) && !headers["Approved"]) {
	message->append("Unapproved article posted to moderated group.");
	return POST_REJECTED;
    }

    /* Count lines and bytes. */
    num_lines = count_chars(body, '\n');
    num_bytes = article->length() - (body - article->chars());

    return transfer_article(&headers, body, id, num_lines, num_bytes,
			    access, message);
}

/* The post command needs to:
 * - Error if From or Newsgroups doesn't exist, if neither of Subject
 *   and Control exist, or if NNTP-Posting-Host, Xref, Date-Received,
 *   Received, Posted, Posting-Version, or Relay-Version do exist.
 * - Verify the Date, Expires, Distribution, and Message-ID headers
 *   if they exist.
 * - Generate Date, Message-ID, Path, Organization, MIME-Version,
 *   Content-Type, Lines, and NNTP-Posting-Host headers if they don't
 *   exist.
 * - Override Subject with Control header if a Control header exists;
 *   otherwise, generate Control header from Subject if Subject begins
 *   with "cmsg".
 * - Verify that posting is allowed to all newsgroups in the Newsgroups
 *   header and, if it exists, the Followup-To header.
 * - Mail the article if it is to a moderated group.
 * - Otherwise, store and propagate the article. */
int Spool::post_article(const String *article, const char *posting_host,
			const Wildcard_list<bool> *access, String *message)
{
    Headers headers;
    const char *body, *newsgroups, *control, *subject, *date, *expires, *dist;
    const char *id, *follow;
    int num_lines, num_bytes;
    time_t t, now;
    String buf, mod;

    body = headers.add_headers(article->chars());
    if (!body) {
	message->append("Malformatted headers.");
	return -1;
    }
    num_bytes = article->length() - (body - article->chars());
    num_lines = count_chars(body, '\n');

    /* Pull out some newsgroup headers we want to examine. */
    newsgroups = headers["Newsgroups"];
    control = headers["Control"];
    subject = headers["Subject"];
    date = headers["Date"];
    expires = headers["Expires"];
    dist = headers["Distribution"];
    id = headers["Message-ID"];
    follow = headers["Followup-To"];

    /* Verify that Newsgroups and From headers exist, and that either Control
     * or Subject exists. */
    if (!newsgroups || (!control && !subject) || !headers["From"]) {
	message->append("Missing From, Newsgroups, or Subject header.");
	return -1;
    }

    /* Complain about system or obsolete headers. */
    if (headers["NNTP-Posting-Host"] || headers["Xref"]) {
	message->append("NNTP-Posting-Host or Xref set in headers.");
	return -1;
    }
    if (headers["Date-Received"] || headers["Received"] || headers["Posted"]
	|| headers["Posting-Version"] || headers["Relay-Version"]) {

	message->append("Obsolete article headers.");
	return -1;
    }

    /* Verify the Date header if it exists, or fill it in if it doesn't. */
    if (date) {
	t = parse_header_date(date);
	time(&now);
	if (t == -1 || t > now) {
	    message->append("Date header is invalid or in the future.");
	    return -1;
	}
    } else {
	buf.append("Date: ");
	generate_current_header_date(&buf);
	headers.add_header(buf.chars(), buf.length());
    }

    /* Verify the Expires header if it exists. */
    if (expires && parse_header_date(expires) == -1) {
	message->append("Expires header contains invalid date.");
	return -1;
    }

    /* Verify the Distribution header if it exists. */
    if (dist && config->any_distribution_accepted(dist) < 0) {
	format(message, "Server does not accept distributions \"%s\"", dist);
	return -1;
    }

    /* Verify that posting is allowed in Newsgroups and Followup-To:
     * headers. */
    if (strpbrk(newsgroups, ":/ \f\t\n") != NULL ||
	(follow && strpbrk(follow, ":/ \f\t\n") != NULL)) {
	message->append("Invalid Newsgroups or Followup-To header.");
	return -1;
    }
    buf.clear();
    if (groups->verify_post_access(newsgroups, access, &buf, &mod) < 0 ||
	(follow &&
	 groups->verify_post_access(follow, access, &buf, NULL) < 0)) {
	format(message, "Posting not allowed to \"%S\".", &buf);
	return -1;
    }

    /* Verify the Message-ID header if it exists, or generate one if not. */
    if (id && verify_message_id(id) < 0) {
	format(message, "Invalid Message-ID \"%s\"", id);
	return -1;
    } else if (!id) {
	buf.clear();
	buf.append("Message-ID: ");
	generate_message_id(&buf);
	headers.add_header(buf.chars(), buf.length());
	id = headers["Message-ID"];
    }

    /* Generate Path, Organization, MIME-Version, Content-Type, Lines, and
     * NNTP-Posting-Host headers if they don't exist. */
    if (!headers["Path"])
	headers.add_header("Path: usenet");
    if (!headers["Organization"]) {
	buf.clear();
	buf.append("Organization: ");
	config->get_default_organization(&buf);
	headers.add_header(buf.chars(), buf.length());
    }
    if (config->get_add_mime_headers_flag()) {
	if (!headers["MIME-Version"])
	    headers.add_header("MIME-Version: 1.0");
	if (!headers["Content-Type"])
	    headers.add_header("Content-Type: text/plain");
    }
    format_clear(&buf, "Lines: %d", num_lines);
    headers.add_header(buf.chars(), buf.length());
    format_clear(&buf, "NNTP-Posting-Host: %s", posting_host);
    headers.add_header(buf.chars(), buf.length());

    /* Make sure Subject and Control headers agree. */
    if (control) {
	format_clear(&buf, "Subject: cmsg %s", control);
	headers.add_header(buf.chars(), buf.length());
    } else if (strnccmp(subject, "cmsg ", 5) == 0) {
	format_clear(&buf, "Control: %s", subject + 5);
	headers.add_header(buf.chars(), buf.length());
    }

    if (*mod.chars()) {
	return mail_article(&headers, body, num_bytes, mod.chars(), message);
    } else {
	return (transfer_article(&headers, body, id, num_lines, num_bytes,
				 access, message) == POST_OK) ? 0 : -1;
    }
}

Post_status Spool::transfer_article(Headers *headers, const char *body,
				    const char *id, int num_lines,
				    int num_bytes,
				    const Wildcard_list<bool> *access,
				    String *message)
{
    int partition;
    time_t bucket_num, expire_time;
    Bucket *bucket;
    long offset;
    const char *expires;
    Post_status status;
    Id_entry *entry;
    String buf;
    const char *control = (*headers)["Control"];

    /* Add the name of this server to the path. */
    format(&buf, "Path: %S!%s", &server_name, (*headers)["Path"]);
    headers->add_header(buf.chars(), buf.length());

    /* Add an Xref line. */
    format_clear(&buf, "Xref: %S", &server_name);
    groups->reserve_positions((*headers)["Newsgroups"], access, &buf);
    headers->add_header(buf.chars(), buf.length());

    /* Record that we've started the post, or return an appropriate status
     * if we collided with another thread (POST_REJECTED if it committed
     * or rejected the article, POST_FAILED if it throttled trying to post
     * the article). */
    entry = transactions.beginning_post(id, &status);
    if (!entry) {
	if (status == POST_REJECTED)
	    message->append("Already have this article.");
	else if (status == POST_FAILED)
	    message->append("Can't write article to disk.");
	return status;
    }

    /* Reject the article if we already have it. */
    if (index->present(id)) {
	transactions.post_committed(entry);
	message->append("Already have this article.");
	return POST_REJECTED;
    }

    /* Accept the article later if we are throttled. */
    throttle_lock.lock();
    if (throttle_message) {
	transactions.post_failed(entry);
	message->append(throttle_message);
	return POST_FAILED;
    }
    throttle_lock.unlock();

    /* Add the article to the appropriate bucket. */
    expires = (*headers)["Expires"];
    expire_time = (expires) ? parse_header_date(expires) : -1;
    bucket_num = config->decide_bucket((*headers)["Newsgroups"], expire_time);
    partition = pool->decide_partition(num_bytes);
    bucket = pool->find_or_create_bucket(partition, bucket_num);
    if (bucket) {
	offset = bucket->write_article(headers, body, num_bytes);
	pool->release_bucket(bucket);
    }
    if (!bucket || offset < 0) {
	throttle("Can't create or write to bucket <%d, %l>.",
		 partition, (long) bucket_num);
	transactions.post_failed(entry);
	message->append("Can't write article to disk.");
	return POST_FAILED;
    }

    /* Add the article to the index. */
    if (index->store(id, partition, bucket_num, offset) < 0) {
	throttle("Can't write index entries.");
	transactions.post_committed(entry);
	message->append("Article stored, but not indexed or in groups.");
	return POST_REJECTED;
    }

    /* If this is a control message, process it; otherwise add the article
     * to its newsgroups. */
    if (control) {
	handle_control_message(headers, body);
    } else if (groups->add_to_groups(headers, num_lines, num_bytes,
				     partition, bucket_num, offset, 0) < 0) {
	throttle("Can't write to group file.");
	transactions.post_committed(entry);
	message->append("Article stored and indexed, but not in groups.");
	return POST_REJECTED;
    }

    /* Propagate the article.  If this fails, throttle, but report success
     * in transferring the article; we will repropagate the article during
     * the next consistency check. */
    if (feeds->propagate_article(partition, bucket_num, offset,
				 headers, body) == -1)
	throttle("Can't write propagation bookkeeping entry.");

    transactions.post_committed(entry);
    message->append("Article transferred ok.");
    return POST_OK;
}

void Spool::handle_control_message(Headers *headers, const char *body)
{
    const char *p, *q, *r, *groups = (*headers)["Newsgroups"];

    /* Get first words (p to q) and beginning of second (r). */
    r = find_word(q = word_end(p = (*headers)["Control"]));
    if (q - p == 6 && strnccmp(p, "cancel", 6) == 0 && *r != 0)
	control_cancel(r, (*headers)["From"], groups);
    else if (q - p == 5 && strnccmp(p, "ihave", 5) == 0)
	control_ihave(r, body, groups);
    else if (q - p == 6 && strnccmp(p, "sendme", 6) == 0)
	control_sendme(r, body, groups);
    else if (q - p == 8 && strnccmp(p, "newgroup", 8) == 0)
	control_newgroup(r, body, (*headers)["From"]);
    else if (q - p == 7 && strnccmp(p, "rmgroup", 7) == 0)
	control_rmgroup(r, (*headers)["From"]);
}

int Spool::mail_article(Headers *headers, const char *body, int num_bytes,
			const char *mod, String *message)
{
    FILE *mail;
    const char *line;
    int status;

    mail = popen(SENDMAIL_CMD, "w");
    if (!mail) {
	message->append("Can't start mailer to mail article to moderator.");
	return -1;
    }

    fprintf(mail, "To: %s\n", mod);
    for (line = headers->first_header(); line; line = headers->next_header()) {
	fputs(line, mail);
	putc('\n', mail);
    }

    putc('\n', mail);
    fwrite(body, sizeof(char), num_bytes, mail);
    status = ferror(mail);
    status = (pclose(mail) == 0) ? status : -1;
    if (status < 0) {
	message->append("Can't mail article to moderator.");
	return -1;
    }

    return 0;
}

/* Assumes no modification while this is taking place, and only one call
 * at a time. */
int Spool::reindex_article(int num_lines, int num_bytes, int partition,
			   time_t bucket_num, long offset,
			   const Headers *headers)
{
    const char *groupnames, *id;

    groupnames = (*headers)["Newsgroups"];
    id = (*headers)["Message-ID"];
    if (!groupnames || !id)
	return -1;

    if (!index->present(id) < 0) {
	if (index->store(id, partition, bucket_num, offset) < 0)
	    return -1;
    }

    if ((*headers)["Control"] == NULL) {
	if (groups->add_to_groups(headers, num_lines, num_bytes,
				  partition, bucket_num, offset, 1) < 0)
	    return -1;
    }

    if (feeds->propagate_article(partition, bucket_num, offset,
				 headers, NULL) < 0)
	return -1;

    /* XXX Unfinished control messages. */
    return 0;
}

void Spool::create_group(const char *group, int grouplen,
			 const char *flags, int flagslen,
			 const char *desc, int desclen)
{
    groups->add_group(group, grouplen, flags, flagslen, desc, desclen,
		      time(NULL));
}

void Spool::merge_groups(FILE *fp)
{
    groups->merge_groups(fp);
}

void Spool::send_grouplist(Connection *conn, const Wildcard_list<bool> *access)
{
    groups->send_grouplist(conn, access);
}

void Spool::send_group_descriptions(Connection *conn,
				    const Wildcard_list<bool> *access)
{
    groups->send_group_descriptions(conn, access);
}

void Spool::send_new_groups(Connection *conn, time_t mintime, const char *dist,
			    const Wildcard_list<bool> *access)
{
    groups->send_new_groups(conn, mintime, dist, access);
}

void Spool::send_schema(Connection *conn)
{
    const char *const *p;

    for (p = config->get_schema(); *p; p++) {
	if (**p == ':')
	    conn->writef("%s:full\r\n", *p + 1);
	else
	    conn->writef("%s:\r\n", *p);
    }
}

int Spool::find_header(int partition, time_t bucket_num, long offset,
		       const char *name, int len, String *buf)
{
    Bucket *bucket;
    int result;

    /* Find the bucket. */
    bucket = pool->find_bucket(partition, bucket_num);
    if (bucket == NULL)
	return -1;

    /* Look up the header, release the bucket, and return. */
    result = bucket->find_header(offset, name, len, buf);
    pool->release_bucket(bucket);
    return result;
}

int Spool::find_header(const char *id, const char *name, int len, String *buf)
{
    int partition;
    time_t bucket_num;
    long offset;

    if (index->lookup(id, &partition, &bucket_num, &offset) < 0)
	return -1;
    return find_header(partition, bucket_num, offset, name, len, buf);
}

void Spool::throttle(const char *message, ...)
{
    va_list args;

    throttle_lock.lock();
    if (throttle_message)
	throttle_message->clear();
    else
	throttle_message = new String;
    va_start(args, message);
    vformat(throttle_message, message, args);
    va_end(args);
    throttle_lock.unlock();
}

void Spool::unthrottle()
{
    if (consistency_check() < 0)
	return;

    throttle_lock.lock();
    delete throttle_message;
    throttle_message = NULL;
    throttle_lock.unlock();
}

void Spool::checkpoint()
{
    transactions.quiesce();
    if (!throttle_message) {
	if (groups->write_new_bounds_file() < 0)
	    throttle("Can't write updated grouplist.");
	else if (index->sync() < 0)
	    throttle("Can't synchronize article index.");
	else if (pool->write_checkpoint_record() < 0)
	    throttle("Can't write updated consistency record.");
    }
    transactions.quiesce_done();
}

void Spool::expire()
{
    time_t bucket_num, now;

    while (1) {
	bucket_num = pool->first_bucket();
	time(&now);
	if (bucket_num < 0 || bucket_num > now)
	    break;
	log("Expiring bucket %ld", (long) bucket_num);
	transactions.quiesce();
	if (index->eliminate_bucket(bucket_num) < 0)
	    throttle("Not enough space to rewrite index.");
	transactions.quiesce_done();
	if (groups->eliminate_bucket(bucket_num) < 0)
	    throttle("Not enough space to rewrite group files.");
	pool->expire_bucket(bucket_num);
    }
}

int Spool::consistency_check()
{
    int result;

    /* Quiesce all transactions.  Normally, there won't be any, since the
     * server is throttled or hasn't started accepting connections.  Since
     * posting is asynchronous, though, there may still be some stray
     * posting operations. */
    transactions.quiesce();
    result = pool->consistency_check();
    transactions.quiesce_done();
    return result;
}

/* Message ID format similar to INN's, representing numbers using a
 * 32-character alphabet to represent the process ID, the time, and a
 * 15-bit sequence number.  (INN doesn't use the sequence number.) */
void Spool::generate_message_id(String *buf)
{
    const time_t OFFSET = 673416000L;
    unsigned long pid, t, seq;

    pid = getpid();
    t = time(NULL) - OFFSET;
    seq = sequence.next();
    format(buf, "<%R$%R$%R@%S>", pid, t, seq, &server_name);
}

/* To cancel an article which has already been posted:
 * 	- Remove all references to the article from all group files it's in.
 *	- Rewrite the index entry to the article as a dead entry.
 *	- Mark the article as cancelled in the buckets.
 */
void Spool::control_cancel(const char *ids, const char *from,
			   const char *groups)
{
    const char *p;
    Id_entry *entry;

    for (p = word_end(ids); *ids; ids = find_word(p = word_end(ids))) {
	String id(ids, p - ids);

	entry = transactions.beginning_cancel(id.chars());
	cancel_article(id.chars());
	transactions.cancel_done(entry);
    }
}

/* Stub. */
void Spool::control_ihave(const char *ids, const char *body,
			  const char *groups)
{
    ;
}

/* Stub. */
void Spool::control_sendme(const char *ids, const char *body,
			   const char *groups)
{
    ;
}

void Spool::control_newgroup(const char *s, const char *body,
			     const char *from)
{
    const char *p, *q, *r, *flags = "y";

    /* Get first two words of control argument (s to p, q to r). */
    r = word_end(q = find_word(p = word_end(s)));

    /* If second word exists, it must be "moderated" or "unmoderated". */
    if (r - q == 9 && strnccmp(q, "moderated", 9) == 0 && *r == 0) {
	flags = "m";
    } else if (r - q == 11 && strnccmp(q, "unmoderated", 11) == 0 && *r == 0) {
	flags = "y";
    } else if (*q != 0) {
	log("Bad newgroup syntax: %s", s);
    }

    /* Check the newsgroup name for bad characters. */
    if (strchr(s, ':')) {
	log("Bad newgroup group name: %s", s);
	return;
    }

    /* Check permissions on creating a new group. */
    if (!config->get_newgroup_allow(s, p - s, from)) {
	log("Unapproved newgroup for from %s: %s", from, s);
	return;
    }

    /* Look for a description in the body. */
    q = strcstr(body, "\nFor your newsgroups file:\n");
    if (q) {
	q = find_word(word_end(find_word(p + 27)));
	r = strchr(p, '\n');
    }
    if (!q || !r)
	r = q = "";

    /* Create the group. */
    log("Creating group requested by %s: %s", from, s);
    create_group(s, p - s, flags, 1, q, r - q);

    /* For failure atomicity, flush the grouplist file. */
    groups->write_new_bounds_file();
}

/* Stub. */
void Spool::control_rmgroup(const char *group, const char *from)
{
    ;
}

/* Assumes a cancel transaction on id is in effect. */
void Spool::cancel_article(const char *id)
{
    String xref;
    int partition;
    time_t bucket_num;
    long offset;

    if (index->lookup(id, &partition, &bucket_num, &offset) < 0) {
	/* We don't have the article.  If there's not already a dead entry
	 * for the message ID, make one. */
	if (!index->present(id))
	    index->store_dead_entry(id);
    } else {
	if (find_header(partition, bucket_num, offset, "Xref", 4, &xref) < 0) {
	    error("Can't get Xref header for canceled article %s.", id);
	    return;
	}
	if (groups->cancel_article(xref.chars()) < 0)
	    return;
	if (index->overwrite_with_dead_entry(id) < 0)
	    return;
	pool->mark_cancelled(partition, bucket_num, offset);
    }
}

static void *periodic_checkpoints(void *arg)
{
    while (1) {
	sleep(60);
	((Spool *) arg)->checkpoint();
    }
}

static void *periodic_expire(void *arg)
{
    while (1) {
	sleep(60);
	((Spool *) arg)->expire();
    }
}

