/* Spool.h: Interface to the news spool (thread-safe). */

#ifndef SPOOL_H
#define SPOOL_H

#include "util.h"

#include "Transactions.h"
#include "String.h"
#include "Feeds.h"

extern "C" {
#include <sys/types.h>
}

class Config;
class Article;
class Newsgroup;
class Index;
class Groups;
class Pool;
class Feeds;
class Headers;
class Connection;
template<class Value> class Wildcard_list;

class Spool {
    const Config *config;
    Index *index;
    Groups *groups;
    Pool *pool;
    Feeds *feeds;
    Transactions transactions;
    String *throttle_message;
    Mutex throttle_lock;
    String server_name;
    Counter sequence;

    int consistency_check();
    time_t decide_bucket(const char *expires);
    Post_status transfer_article(Headers *headers, const char *body,
				 const char *id, int num_lines,
				 int num_bytes,
				 const Wildcard_list<bool> *access,
				 String *message);
    void Spool::handle_control_message(Headers *headers, const char *body);
    int mail_article(Headers *headers, const char *body, int num_bytes,
		     const char *mod, String *message);
    void generate_message_id(String *buf);
    void control_cancel(const char *ids, const char *from, const char *groups);
    void control_ihave(const char *ids, const char *body, const char *groups);
    void control_sendme(const char *ids, const char *body, const char *groups);
    void control_newgroup(const char *s, const char *body,
			  const char *from);
    void control_rmgroup(const char *group, const char *from);
    void cancel_article(const char *id);

  public:
    Spool(const Config *config_arg);
    ~Spool();

    /* Looking up newsgroups and articles. */
    Newsgroup *find_group(const char *name);
    Article *find_article(const char *id);
    Article *find_article(Newsgroup *group, int index);
    Article *next_article(Newsgroup *group, int *index);
    Article *prev_article(Newsgroup *group, int *index);
    Article *first_article(Newsgroup *group, int *index);
    Article *get_article(int partition, time_t bucket_number, long offset);
    int article_exists(const char *id);

    void release_group(Newsgroup *group);
    void release_article(Article *article);

    int article_wanted(const char *id);
    Post_status ihave_article(const char *id, const String *article,
			      const Wildcard_list<bool> *access,
			      String *message);
    int post_article(const String *article, const char *posting_host,
		     const Wildcard_list<bool> *access, String *message);
    int reindex_article(int num_lines, int num_byes, int partition,
			time_t bucket_num, long offset,
			const Headers *headers);
    void create_group(const char *group, int grouplen,
		      const char *flags, int flagslen,
		      const char *desc, int desclen);
    void merge_groups(FILE *fp);

    void send_grouplist(Connection *conn, const Wildcard_list<bool> *access);
    void send_group_descriptions(Connection *conn,
				 const Wildcard_list<bool> *access);
    void send_new_groups(Connection *conn, time_t mintime, const char *dist,
			 const Wildcard_list<bool> *access);
    void send_schema(Connection *conn);
    int find_header(int partition, time_t bucket_num, long offset,
		    const char *name, int len, String *buf);
    int find_header(const char *id, const char *name, int len, String *buf);

    void throttle(const char *message, ...);
    void unthrottle();

    void checkpoint();
    void expire();
};

#endif

