/* Transactions.cc: Keeping track of posts in progress. */

#include "threads.h"
#include "util.h"

#include "Transactions.h"

/* The transactions list keeps track of which threads are in the process
 * of posting what articles (this applies to articles which have already
 * been transmitted; the response to the IHAVE is just a hint).  In the
 * vast majority of cases, each thread simply creates a new entry in the
 * linked list, and deletes it when it is finished posting.  In the
 * rare case where more than one thread is trying to post the same article
 * at the same time, the first thread creates the list entry and the
 * remaining threads wait until the first thread is done.  The first
 * thread signals it is done either with a call to post_committed(),
 * indicating that the article has been accepted (and should now be
 * rejected) or with a call to post_failed(), indicating that the server
 * has throttled and the client should be asked to retransmit the article
 * later. */

Transactions::Transactions()
{
    head = NULL;
}

Transactions::~Transactions()
{
    Id_entry *entry;

    while (head) {
	entry = head->next;
	delete entry;
    }
}

/* id must be valid until the post_in_index() or post_failed() call. */
Id_entry *Transactions::beginning_post(const char *id, Post_status *status)
{
    Id_entry *entry;

    /* Look for an existing entry. */
    lock();
    if (detect_collision(id, status)) {
	unlock();
	return NULL;
    }

    /* Create a new entry. */
    entry = add_entry(id);
    unlock();
    post_lock.read_lock();
    return entry;
}

void Transactions::post_committed(Id_entry *entry)
{
    post_lock.read_done();
    lock();
    release_entry(entry, POST_REJECTED);
    unlock();
}

void Transactions::post_failed(Id_entry *entry)
{
    post_lock.read_done();
    lock();
    release_entry(entry, POST_FAILED);
    unlock();
}

/* id must be valid until the cancel_done() call. */
Id_entry *Transactions::beginning_cancel(const char *id)
{
    Id_entry *entry;

    lock();
    detect_collision(id, NULL);
    entry = add_entry(id);
    unlock();
    post_lock.read_lock();
    return entry;
}

void Transactions::cancel_done(Id_entry *entry)
{
    post_lock.read_done();
    lock();
    release_entry(entry, POST_REJECTED);
    unlock();
}

void Transactions::quiesce()
{
    post_lock.write_lock();
}

void Transactions::quiesce_done()
{
    post_lock.write_done();
}

/* Assumes lock. */
int Transactions::detect_collision(const char *id, Post_status *status)
{
    Id_entry *entry;

    for (entry = head; entry; entry = entry->next) {
	if (entry->id != NULL && message_id_equal(entry->id, id)) {
	    entry->collisions++;
	    entry->finished.wait(this);
	    if (status)
		*status = entry->status;
	    release_entry(entry, entry->status);
	    return 1;
	}
    }
    return 0;
}

#include <assert.h>

/* Assumes lock. */
Id_entry *Transactions::add_entry(const char *id)
{
    Id_entry *entry;

    assert(!head || head->prev_ptr == &head);
    entry = new Id_entry;
    entry->id = id;
    entry->collisions = 0;
    entry->next = head;
    if (head != NULL)
	head->prev_ptr = &entry->next;
    head = entry;
    entry->prev_ptr = &head;
    assert(!head || head->prev_ptr == &head);
    return entry;
}

/* Assumes lock. */
void Transactions::release_entry(Id_entry *entry, Post_status status)
{
    assert(!head || head->prev_ptr == &head);
    if (entry->collisions == 0) {
	*entry->prev_ptr = entry->next;
	if (entry->next != NULL)
	    entry->next->prev_ptr = entry->prev_ptr;
	delete entry;
    } else {
	entry->id = NULL;
	entry->status = status;
	entry->finished.signal();
    }
    assert(!head || head->prev_ptr == &head);
}

