/* Wildcard_list.cc: Matching against a list of wildcard patterns. */

#include "threads.h"
#include "util.h"

#include "Wildcard_list.h"

extern "C" {
#include <assert.h>
#include <string.h>
}

#pragma implementation

const int DELTA = 4;
const int MIN_POWER = 5;

template<class Value> struct Wildcard_entry {
    char **patterns;
    Value value;
};

static int patterns_always_match(const char *const *patterns);

template<class Value>
Wildcard_list<Value>::Wildcard_list(void (*destructor_arg)(Value value))
{
    entries = NULL;
    num = 0;
    size = 0;
    destructor = destructor_arg;
}

template<class Value> Wildcard_list<Value>::~Wildcard_list()
{
    int i;

    for (i = 0; i < num; i++) {
	stringlist_free(entries[i].patterns);
	if (destructor)
	    destructor(entries[i].value);
    }
    delete[] entries;
}

template<class Value>
void Wildcard_list<Value>::add_pattern(const char *const *patterns,
				       Value value)
{
    Wildcard_entry<Value> *new_entries;
    int new_size, i;

    /* Resize the table if necessary. */
    if (num == size) {
	new_size = adjust_size(num + 1, MIN_POWER, DELTA);
	new_entries = new Wildcard_entry<Value>[new_size];
	for (i = 0; i < size; i++)
	    new_entries[i] = entries[i];
	if (entries)
	    delete[] entries;
	entries = new_entries;
	size = new_size;
    }

    /* Fill in a new entry at the end. */
    entries[num].patterns = stringlist_copy(patterns);
    entries[num].value = value;
    num++;
}

template<class Value>
void Wildcard_list<Value>::add_pattern(const char *pattern, Value value)
{
    const char *const patterns[] = { pattern, NULL };

    add_pattern(patterns, value);
}

template<class Value>
Value Wildcard_list<Value>::match(const char *const *strings,
				  Value default_val) const
{
    int i;
    const char *const *p;
    const char *const *q;

    for (i = 0; i < num; i++) {
	for (p = entries[i].patterns, q = strings; *p && *q; p++, q++) {
	    if (!wildcard_match(*p, *q))
		break;
	}
	if (!*p && !*q)
	    return entries[i].value;
    }
    return default_val;
}

template<class Value>
Value Wildcard_list<Value>::match(const char *s, Value default_val)
    const
{
    const char *const strings[] = { s, NULL };

    return match(strings, default_val);
}

template<class Value> int Wildcard_list<Value>::num_entries()
{
    return num;
}

template<class Value>
int Wildcard_list<Value>::is_degenerate(Value default_value,
					Value *value) const
{
    int i;

    /* If there are no entries, the list always returns the default value. */
    if (num == 0) {
	*value = default_value;
	return 1;
    }

    /* See if we always return the first value. */
    *value = entries[0].value;
    if (patterns_always_match(entries[0].patterns))
	return 1;
    for (i = 1; i < num; i++) {
	if (!(entries[i].value == *value))
	    return 0;
	if (patterns_always_match(entries[i].patterns))
	    return 1;
    }

    /* Degenerate if all the values in the table are the default value. */
    return (*value == default_value);
}

template<class Value>
Wildcard_list<Value> *Wildcard_list<Value>::sublist(const char *const *strings)
    const
{
    Wildcard_list<Value> *sub = new Wildcard_list<Value>;
    const char *const *p;
    const char *const *q;
    int i;

    for (i = 0; i < num; i++) {
	for (p = entries[i].patterns, q = strings; *p && *q; p++, q++) {
	    if (!wildcard_match(*p, *q))
		break;
	}
	if (!*q)
	    sub->add_pattern(p, entries[i].value);
    }

    return sub;
}

static int patterns_always_match(const char *const *patterns)
{
    int len;

    for (; *patterns; patterns++) {
	len = strcspn(*patterns, "*");
	if (!len || (*patterns)[len])
	    return 0;
    }
    return 1;
}

/* Kludge to force instantiations of template. */
#include "Config.h"
template class Wildcard_list<Expire_spec>;
template class Wildcard_list<bool>;
template class Wildcard_list<char *>;

