/* main.cc: Main loop of TNS. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Server.h"
#include "Connection.h"
#include "Config.h"
#include "Spool.h"
#include "Nntp.h"
#include "Admin.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <signal.h>
#include <netinet/in.h>
}

struct Params {
    Spool *spool;
    Config *config;
};

static void *nntp(void *arg);
static void admin(Params *params);
static void *handle_nntp_connection(void *arg);
static void *handle_admin_connection(void *arg);

int debug = 0;

int main(int argc, char **argv)
{
    Params params;
    const char *config_file = DEFAULT_CONFIG_FILE;
    struct sigaction action;

    /* Initialize utility and error modules, ignore SIGPIPE. */
    action.sa_handler = SIG_IGN;
    sigemptyset(&action.sa_mask);
    action.sa_flags = 0;
    sigaction(SIGPIPE, &action, NULL);
    util_init();
    error_init(argv[0]);

    /* Process command-line flags. */
    for (argv++; *argv && **argv == '-'; argv++) {
	switch ((*argv)[1]) {
	  case 'd':
	    debug = 1;
	    break;
	  default:
	    die("Invalid command-line flag \"-%c\".", (*argv)[1]);
	}
    }

    if (*argv && argv[1])
	die("Too many command-line arguments.");
    if (*argv)
	config_file = *argv;

    if (debug)
	pthread_setschedparam(pthread_self(), SCHED_FIFO, NULL);

    log("Started.");
    params.config = new Config(config_file);
    params.spool = new Spool(params.config);

    start_thread(nntp, &params);
    admin(&params);

    delete params.spool;
    delete params.config;
    return 0;
}

static void *nntp(void *arg)
{
    Params *params = (Params *) arg;
    Server server("nntp", htons(119));
    Nntp *nntp;
    Connection *conn;

    log("Accepting NNTP connections.");
    while (1) {
	conn = server.accept();
	nntp = new Nntp(params->spool, conn, params->config);
	start_thread(handle_nntp_connection, nntp);
    }
    return NULL;
}

static void admin(Params *params)
{
    Server server(NULL, htons(20119));
    Connection *conn;
    Admin *admin;

    while (1) {
	conn = server.accept();
	admin = new Admin(params->spool, conn, params->config);
	start_thread(handle_admin_connection, admin);
    }
}

static void *handle_nntp_connection(void *arg)
{
    Nntp *nntp = (Nntp *) arg;

    nntp->run();
    delete nntp;
    return NULL;
}

static void *handle_admin_connection(void *arg)
{
    Admin *admin = (Admin *) arg;

    admin->run();
    delete admin;
    return NULL;
}

