/* util.cc: Miscellaneous utilities. */

#include "threads.h"
#include "util.h"

#include "String.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <resolv.h>
#include <errno.h>
}

/* Timezone error, must not be a valid differential. */
#define TZ_ERROR -10000
#define LEAP_YEAR(y) (((y) % 4 == 0 && (y) % 100 != 0) || ((y) % 400) == 0)

char lowercase[128];
char uppercase[128];

static const char *months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
				"Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

static int find_weekday(const char *s);
static int find_month(const char *s);
static int parse_timezone(const char *s);
static time_t gmktime(struct tm *tm);

void util_init(void)
{
    int i;

    for (i = 0; i < 128; i++) {
	lowercase[i] = (isupper(i) ? tolower(i) : i);
	uppercase[i] = (islower(i) ? toupper(i) : i);
    }
}

/* Algorithm by Peter J. Weinberger. */
unsigned long hash(const char *s)
{
    unsigned long hashval = 0, g;

    /* Algorithm by Peter J. Weinberger. */
    for (; *s; s++) {
	hashval = (hashval << 4) + *s;
	g = hashval & 0xf0000000;
	if (g) {
	    hashval ^= g >> 24;
	    hashval ^= g;
	}
    }
    return hashval;
}

/* Assumes ASCII.  Change "(*s & 0x5f)" to "tolower(*s)" to eliminate
 * this assumption, with a corresponding speed hit. */
unsigned long hash_case(const char *s)
{
    unsigned long hashval = 0, g;

    /* Algorithm by Peter J. Weinberger. */
    for (; *s; s++) {
	hashval = (hashval << 4) + (*s & 0x5f);
	g = hashval & 0xf0000000;
	if (g) {
	    hashval ^= g >> 24;
	    hashval ^= g;
	}
    }
    return hashval;
}

/* Assumes ASCII.  Change "(*s & 0x5f)" to "tolower(*s)" to eliminate
 * this assumption, with a corresponding speed hit. */
unsigned long hash_case(const char *s, int len)
{
    unsigned long hashval = 0, g;

    /* Algorithm by Peter J. Weinberger. */
    for (; len; s++, len--) {
	hashval = (hashval << 4) + (*s & 0x5f);
	g = hashval & 0xf0000000;
	if (g) {
	    hashval ^= g >> 24;
	    hashval ^= g;
	}
    }
    return hashval;
}

/* Compare two strings, ignoring case. */
int strccmp(const char *s1, const char *s2)
{
    while (*s1 && LCASE(*s1) == LCASE(*s2))
	s1++, s2++;
    return LCASE(*s1) - LCASE(*s2);
}

/* Compare two strings up to n characters, ignoring case. */
int strnccmp(const char *s1, const char *s2, int n)
{
    while (n-- && *s1 && LCASE(*s1) == LCASE(*s2))
	s1++, s2++;
    return (n >= 0) ? LCASE(*s1) - LCASE(*s2) : 0;
}

/* Look for substr in s, ignoring case. */
const char *strcstr(const char *s, const char *substr)
{
    int len = strlen(substr) - 1;

    while (*s) {
	if (LCASE(*s) == LCASE(*substr)) {
	    if (strnccmp(s + 1, substr + 1, len) == 0)
		return s;
	}
	s++;
    }
    return NULL;
}

char *strduplicate(const char *s)
{
    return strduplicate(s, strlen(s));
}

char *strduplicate(const char *s, int len)
{
    char *p;

    p = new char[len + 1];
    memcpy(p, s, len);
    p[len] = 0;
    return p;
}

void reverse_chars(char *s, int n)
{
    char t;

    while (n > 0) {
	t = *s;
	*s = s[n - 1];
	s[n - 1] = t;
	s++;
	n -= 2;
    }
}

const char *word_end(const char *s)
{
    while (*s && !isspace(*s))
	s++;
    return s;
}

const char *find_word(const char *s)
{
    while (isspace(*s))
	s++;
    return s;
}

const char *next_word(const char **sptr)
{
    const char *s = *sptr;

    while (*s && !isspace(*s))
	s++;
    while (isspace(*s))
	s++;
    *sptr = s;
    return s;
}

int count_chars(const char *s, int c)
{
    char *p;
    int count = 0;

    p = strchr(s, c);
    while (p) {
	count++;
	p = strchr(p + 1, c);
    }
    return count;
}

int count_words(const char *s)
{
    int count = 0;

    s = find_word(s);
    while (*s) {
	count++;
	next_word(&s);
    }
    return count;
}

int wildcard_match(const char *pattern, const char *s)
{
    const char *lastglob = NULL;

    /* Scan the input string updating pattern, backtracking pattern to lastglob
     * whenever we hit a mismatch.  This is optimized a bit for quickly
     * matching globs, such that we skip quickly over characters in the input
     * string that would cause us to immediately backtrack. */
    while (*s) {
	if (*pattern == '*') {
	    lastglob = pattern++;
	    if (!*pattern)
		return 1;
	    if (*pattern != '*' && *pattern != '?') {
		while (*s && LCASE(*s) != LCASE(*pattern))
		    s++;
	    }
	} else if (LCASE(*pattern) == LCASE(*s) || *pattern == '?') {
	    pattern++;
	    s++;
	} else if (!lastglob) {
	    return 0;
	} else {
	    pattern = lastglob;
	}
    }

    /* Return true if we've exhausted the pattern. */
    return (*pattern == 0);
}

char **stringlist_empty()
{
    char **list;

    list = new char *[1];
    list[0] = NULL;
    return list;
}

char **stringlist_append(char **list, const char *s, int len)
{
    int llen;
    char **newlist, **p, **q;

    llen = stringlist_length(list);
    newlist = new char *[llen + 2];
    for (p = list, q = newlist; *p; p++, q++)
	*q = *p;
    *q = strduplicate(s, len);
    *(q + 1) = NULL;
    delete[] list;
    return newlist;
}

char **stringlist_append_words(char **list, const char *words)
{
    int len;
    char **newlist, **p, **q;
    const char *end;

    len = stringlist_length(list) + count_words(words);
    newlist = new char *[len + 1];
    for (p = list, q = newlist; *p; p++, q++)
	*q = *p;
    while (*words) {
	end = word_end(words);
	*q++ = strduplicate(words, end - words);
	words = find_word(end);
    }
    delete[] list;
    return newlist;
}

int stringlist_length(const char *const *list)
{
    const char *const *p = list;

    while (*p)
	p++;
    return p - list;
}

int stringlist_contains(const char *const *list, const char *s, int len)
{
    for (; *list; list++) {
	if (strnccmp(*list, s, len) == 0 && (*list)[len] == 0)
	    return 1;
    }
    return 0;
}

char **stringlist_copy(const char *const *list)
{
    char **newlist, **p;

    newlist = new char *[stringlist_length(list)];
    for (p = newlist; *list; list++, p++)
	*p = strduplicate(*list);
    *p = NULL;
    return newlist;
}

void stringlist_free(char **list)
{
    char **p;

    for (p = list; *p; p++)
	delete[] p;
    delete[] list;
}

char *long_to_ascii(long num, Number_buf nbuf)
{
    char *p = nbuf + NUMBER_BUF_SIZE - 1;
    int sign = 0;

    *p-- = 0;
    if (num < 0) {
	sign = 1;
	num = -num;
    } else if (!num) {
	*p-- = '0';
    }
    while (num) {
	*p-- = num % 10 + '0';
	num /= 10;
    }
    if (sign)
	*p-- = '-';
    return p + 1;
}

char *long_to_radix32(unsigned long num, Number_buf nbuf)
{
    static char *alphabet = "0123456789abcdefghijklmnopqrstuv";
    char *p = nbuf + NUMBER_BUF_SIZE - 1;

    *p-- = 0;
    if (num) {
	while (num) {
	    *p-- = alphabet[num % 32];
	    num /= 32;
	}
    } else {
	*p-- = alphabet[0];
    }
    return p;
}

void format(String *buf, const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    vformat(buf, fmt, args);
    va_end(args);
}

void format_clear(String *buf, const char *fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    buf->clear();
    vformat(buf, fmt, args);
    va_end(args);
}

void vformat(String *buf, const char *fmt, va_list args)
{
    char *p, *s, zero_filler[] = "00000000000";
    Number_buf nbuf;
    String *str;
    int len;

    while (1) {

	/* Find % or end of string. */
	p = strchr(fmt, '%');
	if (!p || !p[1]) {
	    /* No more percents; copy rest and stop. */
	    buf->append(fmt);
	    break;
	}
	buf->append(fmt, p - fmt);

	switch (p[1]) {

	  case '%':
	    buf->append('%');
	    break;

	  case 's':
	    s = va_arg(args, char *);
	    buf->append(s);
	    break;

	  case 'S':
	    str = va_arg(args, String *);
	    buf->append(str);
	    break;

	  case 'd':
	    s = long_to_ascii(va_arg(args, int), nbuf);
	    buf->append(s);
	    break;

	  case 'l':
	    s = long_to_ascii(va_arg(args, long), nbuf);
	    buf->append(s);
	    break;

	  case 'c':
	    buf->append(va_arg(args, int));
	    break;

	  case 'C':
	    s = va_arg(args, char *);
	    len = va_arg(args, int);
	    buf->append(s, len);
	    break;

	  case 'N':
	    s = long_to_ascii(va_arg(args, int), nbuf);
	    len = NNTP_NUMBER_LEN - strlen(s);
	    if (*s == '-')
		buf->append(*s++);
	    buf->append(zero_filler, len);
	    buf->append(s);
	    break;

	  case 'R':
	    s = long_to_radix32(va_arg(args, unsigned long), nbuf);
	    buf->append(s);
	    break;

	  default:
	    buf->append('%');
	    buf->append(p[1]);
	}

	fmt = p + 2;
    }
}

/* Return the smallest number N such that N >= size and N = 2^B - delta for
 * some B >= min_bits. */
int adjust_size(int size, int min_power, int delta)
{
    int b;

    size += delta - 1;
    b = min_power;
    while ((size >> b) != 0)
	b++;
    return (1 << b) - delta;
}

int read_line(String *line, FILE *fp)
{
    char buf[1024];

    /* Read in the first chunk, and handle some trivial cases. */
    if (!fgets(buf, sizeof(buf), fp) || !*buf)
	return -1;
    if (*buf == '\n')
	return 0;

    /* Until we hit a newline, keep reading in text. */
    while (1) {
	line->append(buf);
	if (line->chars()[line->length() - 1] == '\n') {
	    line->truncate(line->length() - 1);
	    break;
	}
	if (!fgets(buf, sizeof(buf), fp))
	    break;
    }

    /* Strip off any trailing carriage return and return success. */
    if (line->chars()[line->length() - 1] == '\r')
	line->truncate(line->length() - 1);
    return 0;
}

int read_line_clear(String *line, FILE *fp)
{
    line->clear();
    return read_line(line, fp);
}

int tcp_connect(const char *name, const char *service,
		unsigned short fallback, String *error)
{
    struct hostent_answer hanswer;
    struct servent_answer sanswer;
    struct hostent *hostent;
    struct servent *servent;
    struct sockaddr_in sin;
    int fd;

    sin.sin_family = AF_INET;

    hostent = gethostbyname_r(name, &hanswer);
    if (!hostent) {
	format(error, "Can't resolve host %s: %s", name, hstrerror(h_errno));
	return -1;
    }
    memcpy(&sin.sin_addr, hostent->h_addr, sizeof(struct in_addr));

    servent = getservbyname_r(service, "tcp", &sanswer);
    sin.sin_port = (servent) ? servent->s_port : htons(fallback);

    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
	format(error, "Can't create socket for connection to %s: %s",
	       name, strerror(errno));
	return -1;
    }

    if (connect(fd, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
	format(error, "Can't connect to %s: %s", name, strerror(errno));
	close(fd);
	return -1;
    }

    return fd;
}

/* Compare two Message-IDs.  A robust comparison would require more
 * parsing (see RFCs 822 and 1036), but this works fine in news
 * articles. */
/* XXX I don't think this is right at all. */
int message_id_equal(const char *p, const char *q)
{
    while (*p && *p != '@' && *p == *q)
	p++, q++;
    while (*p && tolower(*p) == tolower(*q))
	p++, q++;
    return (*p) ? 0 : 1;
}

/* Verify that a Message-ID is valid.  A full check for RFC 822 compliance
 * would require more parsing, but this routine only checks whether the
 * Message-ID is composed of a string of non-blank characters other than
 * '>' between '<' and '>', and that it contains at least one '@'.  (RFC
 * 822 allows blanks in Message-IDs, but RFC 1036 doesn't.)  I don't see
 * any need to verify full RFC822 compliance. */
int verify_message_id(const char *id)
{
    if (*id != '<')
	return 0;
    id++;
    while (*id && *id != '>' && !isspace(*id))
	id++;
    return (*id == '>' && id[1] == 0) ? 0 : -1;
}

/* Parse a date according to RFC 822, allowing more than two characters
 * for the year.  Return -1 on rejection.  Grammar is:
 *
 * date = [ weekday "," ] D[D] month Y* HH:MM[:SS] zone
 * zone = ("+" / "-")HHMM / Military code / Supported zone name
 *
 * Example: Fri, 23 Sep 1994 14:42:50 GMT */
time_t parse_header_date(const char *date)
{
    struct tm tm;
    int differential;
    time_t t;

    /* Look for and discard a leading weekday and comma. */
    date = find_word(date);
    if (find_weekday(date) >= 0 && date[3] == ',')
	date = find_word(date + 4);

    /* One or two digits giving the day of the month. */
    if (!isdigit(*date))
	return -1;
    tm.tm_mday = *date - '0';
    date++;
    if (isdigit(*date)) {
	tm.tm_mday = (tm.tm_mday * 10) + (*date - '0');
	date++;
    }
    date = find_word(date);

    /* Name of month. */
    tm.tm_mon = find_month(date);
    if (tm.tm_mon < 0)
	return -1;
    date = find_word(date + 3);

    /* One or more digits giving the year. */
    if (!isdigit(*date))
	return -1;
    tm.tm_year = strtol(date, (char **) &date, 10);
    tm.tm_year = (tm.tm_year < 1000) ? tm.tm_year + 1900 : tm.tm_year;
    if (tm.tm_year < 1900)
	return -1;
    date = find_word(date);

    /* HH:MM */
    if (!isdigit(*date) || !isdigit(date[1]) || date[2] != ':' ||
	!isdigit(date[3]) || !isdigit(date[4]))
	return -1;
    tm.tm_hour = (*date - '0') * 10 + (date[1] - '0');
    tm.tm_min = (date[3] - '0') * 10 + (date[4] - '0');
    tm.tm_sec = 0;
    date += 5;
    /* Optional :SS */
    if (*date == ':' && isdigit(date[1]) && isdigit(date[2])) {
	tm.tm_sec = (date[1] - '0') * 10 + (date[2] - '0');
	date += 3;
    }
    date = find_word(date);

    /* Get time zone differential in minutes. */
    differential = parse_timezone(date);
    if (differential == TZ_ERROR)
	return -1;

    t = gmktime(&tm);
    return t - (differential * 60);
}

static int find_weekday(const char *s)
{
    static const char *weekdays[] = { "Mon", "Tue", "Wed", "Thu",
				      "Fri", "Sat", "Sun" };
    int i;

    for (i = 0; i < 7; i++) {
	if (strnccmp(weekdays[i], s, 3) == 0)
	    return i;
    }
    return -1;
}

static int find_month(const char *s)
{
    int i;

    for (i = 0; i < 12; i++) {
	if (strnccmp(months[i], s, 3) == 0)
	    return i;
    }
    return -1;
}

/* Returns differential in minutes, or TZ_ERROR for error. */
static int parse_timezone(const char *s)
{
    static struct {
	char *name;
	int differential;
    } named[] = {
	{ "UT",   0 },		{ "GMT",  0 },
	{ "EST", -5 * 60 },	{ "EDT", -4 * 60 },
	{ "CST", -6 * 60 },	{ "CDT", -5 * 60 },
	{ "MST", -7 * 60 },	{ "MDT", -6 * 60 },
	{ "PST", -8 * 60 },	{ "PDT", -7 * 60 }
    };
    const int NUM_NAMED_ZONES = sizeof(named) / sizeof(*named);
    static int milzones[] = { -1, -2, -3, -4, -5, -6, -7, -8, -9, 666,
			      -10, -11, -12, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
			      11, 12, 0 };
    int i;

    if (*s == '+' || *s == '-')
	return atoi(s);

    for (i = 0; i < NUM_NAMED_ZONES; i++) {
	if (strnccmp(named[i].name, s, strlen(named[i].name)) == 0)
	    return named[i].differential;
    }

    if (isalpha(*s) && *s != 'j' && *s != 'J' && !isalpha(s[1]))
	return milzones[LCASE(*s) - 'a'] * 60;

    return TZ_ERROR;
}

/* date and time are guaranteed to be six characters, gmt is either an empty
 * string or "GMT". */
time_t parse_nntp_date(const char *date, const char *timest, const char *gmt)
{
    struct tm tm, current;
    time_t t;
    const char *p;
    int century;

    /* Verify that date and time are digits. */
    for (p = date; p < date + 6; p++) {
	if (!isdigit(*p))
	    return -1;
    }
    for (p = timest; p < timest + 6; p++) {
	if (!isdigit(*p))
	    return -1;
    }

    /* Convert digits into numbers, reject invalid numbers. */
    tm.tm_year = (date[0] - '0') * 10 + (date[1] - '0');
    tm.tm_mon =  (date[2] - '0') * 10 + (date[3] - '0') - 1;
    tm.tm_mday = (date[4] - '0') * 10 + (date[5] - '0');
    tm.tm_hour = (timest[0] - '0') * 10 + (timest[1] - '0');
    tm.tm_min =  (timest[2] - '0') * 10 + (timest[3] - '0');
    tm.tm_sec =  (timest[4] - '0') * 10 + (timest[5] - '0');
    tm.tm_isdst = -1;
    if (tm.tm_mon > 11 || tm.tm_mday > 31 || tm.tm_hour > 23 ||
	tm.tm_min > 59 || tm.tm_sec > 59 || tm.tm_mon < 0 || tm.tm_mday < 1)
	return -1;

    /* Get current time to find century. */
    time(&t);
    if (localtime_r(&t, &current)) {
	century = (current.tm_year + 1900) / 100 * 100;
	if (current.tm_year - tm.tm_year >= 50)
	    century += 100;
	else if (tm.tm_year - current.tm_year >= 50)
	    century -= 100;
	tm.tm_year += century - 1900;
    }

    return (*gmt) ? gmktime(&tm) : mktime(&tm);
}

static time_t gmktime(struct tm *tm)
{
    static int mstart[] = { 0,  31,  59,  90, 120, 151,
			    181, 212, 243, 273, 304, 334, 365 };
    time_t t;

    /* Convert tm fields into seconds since midnight, Jan 1, 1970 GMT. */
    t = (tm->tm_year - 70) * (time_t)(365 * 86400);
    t += ((tm->tm_year - 69) / 4) * 86400;
    t += (mstart[tm->tm_mon] + tm->tm_mday - 1) * 86400;
    t += tm->tm_hour * 3600 + tm->tm_min * 60 + tm->tm_sec;
    if (LEAP_YEAR(tm->tm_year) && tm->tm_mon > 1)
	t += 86400;
    return t;
}

int generate_current_header_date(String *buf)
{
    char str[40];
    time_t now;
    struct tm tm;

    time(&now);
    if (gmtime_r(&now, &tm) == NULL)
	return -1;
    sprintf(str, "%d %s %d %02d:%02d:%02d GMT",
	    tm.tm_mday, months[tm.tm_mon], 1900 + tm.tm_year,
	    tm.tm_hour, tm.tm_min, tm.tm_sec);
    buf->append(str);
    return 0;
}

const char *get_group_name(String *group_name, const char *p)
{
    const char *q;

    q = strchr(p, ',');
    q = (q) ? q : p + strlen(p);
    group_name->clear();
    group_name->append(p, q - p);
    return (*q == ',') ? q + 1 : q;
}

void make_directories_in_path(const char *name)
{
    String partial;
    const char *p, *q;

    q = strchr(p = name, '/');
    while (q) {
	partial.append(p, q - p);
	mkdir(partial.chars(), S_IRWXU | S_IRWXG | S_IROTH);
	q = strchr((p = q) + 1, '/');
    }
}

int check_flags(const char *flags, int flen)
{
    if (!strchr("xjmyn=", *flags))
	return -1;
    if (flen != 1 && !(*flags == 'm' || *flags == '='))
	return -1;
    return 0;
}

