/* util.h: Declarations for miscellaneous utilities. */

#ifndef UTIL_H
#define UTIL_H

#include "threads.h"

const int NUMBER_BUF_SIZE = 32;
typedef char Number_buf[NUMBER_BUF_SIZE];

class String;

extern "C" {
#include <stdio.h>
#include <stdarg.h>
#include <sys/types.h>
}

#define LCASE(c) lowercase[(int) c]
#define UCASE(c) uppercase[(int) c]

/* Number of characters to use to represent a numbers using %N in
 * format().  Make sure zero_filler[] in format() is as long as
 * this number. */
#define NNTP_NUMBER_LEN 10

extern char lowercase[128];
extern char uppercase[128];

void util_init(void);

/* Operations on strings. */
unsigned long hash(const char *s);
unsigned long hash_case(const char *s);
unsigned long hash_case(const char *s, int len);
int strccmp(const char *s1, const char *s2);
int strnccmp(const char *s1, const char *s2, int n);
const char *strcstr(const char *s, const char *substr);
char *strduplicate(const char *s);
char *strduplicate(const char *s, int len);
void reverse_chars(char *s, int n);
const char *word_end(const char *s);
const char *find_word(const char *s);
const char *next_word(const char **sptr);
int count_chars(const char *s, int c);
int count_words(const char *s);
int wildcard_match(const char *pattern, const char *s);

/* Operations on lists of strings. */
char **stringlist_empty();
char **stringlist_append(char **list, const char *s, int len);
char **stringlist_append_words(char **list, const char *words);
int stringlist_length(const char *const *list);
int stringlist_contains(const char *const *list, const char *s, int len);
char **stringlist_copy(const char *const *list);
void stringlist_free(char **list);

/* Formatting operations. */
char *long_to_ascii(long num, Number_buf nbuf);
char *long_to_radix32(unsigned long num, Number_buf nbuf);
void format(String *buf, const char *fmt, ...);
void format_clear(String *buf, const char *fmt, ...);
void vformat(String *buf, const char *fmt, va_list args);

/* Miscellaneous. */
int adjust_size(int size, int min_power, int delta);
int read_line(String *line, FILE *fp);
int read_line_clear(String *line, FILE *fp);

/* Network operations. */
int tcp_connect(const char *name, const char *service,
		unsigned short fallback, String *error);

/* Specialized routines for NNTP, the spool, etc. */
int message_id_equal(const char *p, const char *q);
int verify_message_id(const char *id);
time_t parse_header_date(const char *date);
time_t parse_nntp_date(const char *date, const char *timest, const char *nntp);
int generate_current_header_date(String *buf);
const char *get_group_name(String *group_name, const char *p);
void make_directories_in_path(const char *name);
int check_flags(const char *flags, int flen);

class Counter : private Mutex {
    int n;

  public:
    Counter() { n = 0; }
    Counter(int start) { n = start; }
    int next() { int r; lock(); r = n++; unlock(); return r; }
};

#endif

