/* Admin.cc: Admin connection handler class. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Admin.h"
#include "Array.h"
#include "Connection.h"
#include "Spool.h"
#include "Config.h"

extern "C" {
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>
}

static int compare_groups_sort(const void *p, const void *q);
static int compare_groups_search(const void *p, const void *q);

Admin::Admin(Spool *spool_arg, Connection *conn_arg, Config *config_arg)
{
    spool = spool_arg;
    conn = conn_arg;
    config = config_arg;
    running = 1;
}

Admin::~Admin()
{
    delete conn;
}

void Admin::run()
{
    Stringbuf linebuf;
    String line, cmd, rest;

    if (!config->get_admin_access(conn->name(), conn->ip())) {
	conn->write("502 You have no permission to talk.  Goodbye.\r\n");
	return;
    }

    conn->write("200 Threaded news server admin interface ready.\r\n");
    while (running && conn->read_line_clear(&linebuf) >= 0) {
	line = linebuf.str();
	line.strip_whitespace();
	rest = line.get_words(&cmd, 1);

	/* Ignore blank lines. */
	if (line.len() == 0)
	    continue;

	/* Run command associated with the starting word. */
	if (handle_command(cmd, rest) < 0)
	    conn->write("500 Command not recognized.\r\n");
    }
}

int Admin::handle_command(String cmd, String rest)
{
    static struct entry {
	String name;
	void (Admin::*func)(String s);
    } tab[] = {
	{ String("drain"),		Admin::drain },
	{ String("expire"),		Admin::expire },
	{ String("getgroups"),		Admin::getgroups },
	{ String("help"),		Admin::help },
	{ String("list"),		Admin::list },
	{ String("mergegroups"),	Admin::mergegroups },
	{ String("newgroup"),		Admin::newgroup },
	{ String("shutdown"),		Admin::shutdown },
	{ String("throttle"),		Admin::throttle },
	{ String("quit"),		Admin::quit }
    };
    unsigned int i;

    for (i = 0; i < sizeof(tab) / sizeof(struct entry); i++) {
	if (compare_case(tab[i].name, cmd) == 0) {
	    (this->*tab[i].func)(rest);
	    return 0;
	}
    }

    return -1;
}

/* Syntax: drain <partition> <number> */
/* Stub. */
void Admin::drain(String s)
{
    String w[2], rest;
    Bucket_loc bloc;

    /* Check that we have two arguments, a partition and bucket number. */
    rest = s.get_words(w, 2);
    if (w[1].len() == 0 || rest.len() != 0) {
	conn->write("501 DRAIN needs two arguments.\r\n");
	return;
    }

    bloc.partition = w[0].tonum();
    bloc.bucket_num = w[1].tonum();
    if (spool->drain_bucket(&bloc) < 0)
	conn->write("480 Can't drain bucket.\r\n");
    else
	conn->write("200 Starting to drain bucket.\r\n");
}

/* Syntax: expire <partition> <number> */
/* Stub. */
void Admin::expire(String)
{
    ;
}

/* Syntax: getgroups <hostname> <output file> */
void Admin::getgroups(String s)
{
    String w[4], rest;
    FILE *fp = NULL;
    int fd, ngroups = 0, i;
    Stringbuf filename, buf;
    Array<Remote_group, 9, 4> groups;
    Remote_group *group;
    time_t now;

    rest = s.get_words(w, 2);

    /* Must be one or two words. */
    if (w[1].len() == 0 || rest.len() != 0) {
	conn->write("501 GETGROUPS needs two arguments.");
	return;
    }

    filename.store(w[1]);
    fp = fopen(filename.cstring(), "w");
    if (!fp) {
	conn->writef("480 Can't open output file: %m\r\n");
	return;
    }

    fd = tcp_connect(w[0], "nntp", 119, &buf);
    if (fd < 0) {
	conn->writef("480 %s\r\n", buf.str());
	fclose(fp);
	return;
    }

    /* Read the welcome line from the remote server. */
    Connection remote(fd);
    if (remote.read_line_clear(&buf) < 0 || !buf.len() || *buf.str() != '2') {
	conn->writef("480 Access refused to %s.\r\n", w[1]);
	fclose(fp);
	return;
    }

    /* Get a list of newsgroups using LIST. */
    remote.write("LIST\r\n");
    remote.read_line_clear(&buf);
    while (remote.read_line_clear(&buf) >= 0) {
	/* Stop if we find a . by itself on a line. */
	if (buf.len() == 1 && *buf.str() == '.')
	    break;

	/* Verify that the line is four words and that the flags were okay. */
	rest = buf.str().get_words(w, 4);
	if (w[3].len() == 0 || rest.len() != 0) {
	    warn("TAKEGROUPS skipped invalid list line: %s", buf.str());
	    continue;
	}
	if (check_flags(w[3]) < 0) {
	    warn("TAKEGROUPS invalid flags \"%s\" changed to \"y\"", w[3]);
	    w[3] = "y";
	}

	/* Make a new groups entry. */
	groups.ensure(ngroups);
	groups[ngroups].name.store(w[0]);
	groups[ngroups].flags.store(w[3]);
	ngroups++;
    }

    /* Sort the groups. */
    qsort(groups.base(), ngroups, sizeof(Remote_group), compare_groups_sort);

    /* Use LIST NEWSGROUPS to get descriptions. */
    remote.write("LIST NEWSGROUPS\r\n");
    if (remote.read_line_clear(&buf) >= 0 && buf.len() && *buf.str() == '2') {
	while (remote.read_line_clear(&buf) >= 0) {
	    /* Stop if we find a . by itself on a line. */
	    if (buf.len() == 1 && *buf.str() == '.')
		break;

	    /* Look up the group name. */
	    rest = buf.str().get_words(w, 1);
	    group = (Remote_group *) bsearch(&w[0], groups.base(),
					     ngroups, sizeof(Remote_group),
					     compare_groups_search);
	    if (!group)
		continue;

	    /* Set the description of the group. */
	    group->desc.store(rest);
	}
    }

    /* Gracefully terminate the NNTP session. */
    remote.write("QUIT\r\n");
    remote.read_line_clear(&buf);

    /* Write out the resulting newsgroup lines. */
    time(&now);
    for (i = 0; i < ngroups; i++) {
	buf.format_clear("%s %l %s %s\n", groups[i].name.str(), (long) now,
			 groups[i].flags.str(), groups[i].desc.str());
	fputs(buf.cstring(), fp);
    }

    fclose(fp);
    conn->writef("200 GETGROUPS retrieved %d groups.\r\n", ngroups);
}

/* Syntax: help */
void Admin::help(String)
{
    conn->write("100 Help text follows.\r\n");
    conn->write("  DRAIN partition bucket\r\n");
    conn->write("  EXPIRE partition bucket\r\n");
    conn->write("  GETGROUPS server name output file\r\n");
    conn->write("  HELP\r\n");
    conn->write("  LIST BUCKETS\r\n");
    conn->write("  MERGEGROUPS filename\r\n");
    conn->write("  NEWGROUP groupname flags description\r\n");
    conn->write("  SHUTDOWN\r\n");
    conn->write("  THROTTLE\r\n");
    conn->write("  QUIT\r\n");
    conn->write(".\r\n");
}

/* Syntax: list buckets */
void Admin::list(String s)
{
    String w, rest;

    rest = s.get_words(&w, 1);

    if (compare_case(w, "buckets") == 0 && rest.len() == 0) {
	conn->write("215 list of buckets follows.\r\n");
	spool->send_bucket_list(conn);
	conn->write(".\r\n ");
    } else {
	conn->write("501 Invalid arguments to LIST.\r\n");
    }
}

/* Syntax: mergegroups <filename> */
void Admin::mergegroups(String s)
{
    String w, rest;
    Stringbuf filename;
    FILE *fp;

    /* Verify that we have one argument, a filename to read from. */
    rest = s.get_words(&w, 1);
    if (w.len() == 0 || rest.len() != 0) {
	conn->write("501 MERGEGROUPS needs one filename argument.\r\n");
	return;
    }

    /* Open the file. */
    filename.store(w);
    fp = fopen(filename.cstring(), "r");
    if (!fp) {
	conn->write("480 Can't open input file: %m\r\n");
	return;
    }

    spool->merge_groups(fp);
    conn->write("200 MERGEGROUPS finished.\r\n");
}

/* Syntax: newgroup <groupname> <flags> <description> */
void Admin::newgroup(String s)
{
    String w[2], rest;

    rest = s.get_words(w, 2);

    /* Check for bad characters in the newsgroup name. */
    if (w[0].find(':') != -1) {
	conn->write("501 Bad character in newsgroup name.\r\n");
	return;
    }

    /* Create the group. */
    if (w[1].len() == 0)
	w[1] = "y";
    spool->create_group(w[0], w[1], w[2]);
    conn->writef("200 Group %s created.\r\n", w[0]);
}

/* Syntax: shutdown */
/* Stub. */
void Admin::shutdown(String)
{
    ;
}

/* Syntax: throttle */
/* Stub. */
void Admin::throttle(String)
{
    ;
}

void Admin::quit(String)
{
    conn->write("205 Closing connection.\r\n");
    running = 0;
}

static int compare_groups_sort(const void *p, const void *q)
{
    return compare_case(((Remote_group *) p)->name.str(),
			((Remote_group *) q)->name.str());
}

static int compare_groups_search(const void *p, const void *q)
{
    return compare_case(*((String *) p), ((Remote_group *) q)->name.str());
}

