/* Bucket.cc: An article bucket. */

#include "threads.h"
#include "error.h"
#include "util.h"
#include "platform.h"

#include "Bucket.h"
#include "Article.h"
#include "Spool.h"
#include "Headers.h"
#include "String.h"
#include "Stringbuf.h"
#include "Spool.h"

extern "C" {
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <sys/types.h>

/* Non-standard functions not generally prototyped.  Declare them without
 * a prototype, to avoid breaking on existing prototypes. */
int ftruncate();
#ifdef HAVE_FSYNC
int fsync();
#else
#define fsync(fd)
#endif
}

Bucket::Bucket(FILE *file, String name, const Bucket_loc *bloc,
	       Bucket_type type, time_t expire_time)
{
    fp = file;
    filename.store(name);
    this->bloc = *bloc;
    bucket_type = type;
    this->expire_time = expire_time;
    list = NULL;
    fseek(fp, 0, SEEK_END);
}

Bucket::~Bucket()
{
    fclose(fp);
}

Article *Bucket::get_article(long offset)
{
    FILE *copy;
    Article_list *entry;
    int c;

    /* Open a file pointing to the bucket. */
    copy = fopen(filename.cstring(), "r");
    if (!copy) {
	error("Can't re-open bucket file: %m");
	return NULL;
    }
    if (fseek(copy, offset, SEEK_SET) < 0) {
	error("Can't seek to location %l in bucket: %m", offset);
	return NULL;
    }

    /* Read in one character, to make sure there is an article here. */
    c = getc(copy);
    if (c == EOF) {
	fclose(copy);
	return NULL;
    }
    ungetc(c, copy);

    /* Make a new article, chain it into the list, and return it. */
    entry = new Article_list(this, copy, offset);
    lock();
    entry->next = list;
    list = entry;
    unlock();
    return &entry->article;
}

void Bucket::release_article(Article *article)
{
    Article_list **entry_ptr, *entry;

    lock();

    /* Find the article in the list. */
    for (entry_ptr = &list; *entry_ptr; entry_ptr = &(*entry_ptr)->next) {
	entry = *entry_ptr;
	if (&entry->article == article)
	    break;
    }
    if (!*entry_ptr) {
	crash("Released article <%d, %d, %l> is not in the article list.",
	      bloc.partition, bloc.bucket_num, article->offset());
    }

    entry->refs--;
    if (entry->refs == 0) {
	*entry_ptr = entry->next;
	unlock();
	delete entry;
    } else {
	unlock();
    }
}

int Bucket::find_header(long offset, String name, Stringbuf *buf)
{
    Stringbuf line;
    String str;
    FILE *copy;

    copy = fopen(filename.cstring(), "r");
    if (!copy) {
	error("Can't re-open bucket file: %m");
	return -1;
    }
    if (fseek(copy, offset, SEEK_SET) < 0) {
	error("Can't seek to location %l in partition %d bucket %d: %m",
	      offset, bloc.partition, bloc.bucket_num);
	return -1;
    }

    while (read_line_clear(&line, copy) >= 0) {
	str = line.str();
	if (str.len() == 0 || (str.len() == 1 && *str == '.')) {
	    fclose(copy);
	    return -1;
	}
	if (compare_case(str.word(), name) != 0)
	    continue;
	str.skip_whitespace();
	if (str.len() && *str == ':') {
	    /* We found a match.  Look for continuations. */
	    str.shift(1);
	    str.skip_whitespace();
	    buf->append(str);
	    while (read_line_clear(&line, copy) >= 0) {
		if (line.len() == 0 || !isspace(*line.str()))
		    break;
		buf->append(' ');
		buf->append(line.str());
	    }
	    fclose(copy);
	    return 0;
	}
    }

    fclose(copy);
    error("Corrupt article in partition %d bucket %d at offset %l.",
	  bloc.partition, bloc.bucket_num, offset);
    return -1;
}

long Bucket::current_offset() const
{
    return ftell(fp);
}

/* Scan in the article at the given offset, determining its number of
 * lines, number of bytes, and headers.  Returns 0 on success, -1 on
 * failure, 1 if there isn't a complete article to read.  This
 * function is used during a consistency check, so if we read an
 * incomplete article, truncate it and return 1. */
int Bucket::scan_article(long offset, int *num_lines, int *num_bytes,
			 Headers *headers)
{
    FILE *copy;
    Stringbuf line;

    /* Check if nothing's left, or if we're a bucket whose time of
     * accumulation has passed. */
    if (bucket_type != BUCKET_NORMAL || offset == ftell(fp))
	return 1;

    /* Open a file pointing to the bucket. */
    copy = fopen(filename.cstring(), "r");
    if (!copy) {
	error("Consistency check: can't re-open bucket file: %m");
	return -1;
    }

    /* Seek to the consistency point (before that, all articles are known to
     * be filed). */
    if (fseek(copy, offset, SEEK_SET) < 0) {
	error("Consistency check: can't seek to %l: %m", offset);
	fclose(copy);
	return -1;
    }

    /* Try reading the article, starting with the headers. */
    headers->clear();
    if (headers->add_headers(copy) >= 0) {
	/* Read until the terminating line. */
	*num_lines = 0;
	while (read_line_clear(&line, copy) == 0) {
	    (*num_lines)++;
	    if (line.len() == 1 && *line.str() == '.') {
		*num_bytes = ftell(copy) - offset;
		fclose(copy);
		return 0;
	    }
	}
    }

    /* It looks like we got a partial article. */
    fclose(copy);
    ftruncate(fileno(fp), (off_t) offset);
    return 1;
}

int Bucket::mark_article_cancelled(long offset)
{
    int status;

    lock();
    status = fseek(fp, offset, SEEK_SET);
    status = status || ((putc('\n', fp) == EOF) ? -1 : 0);
    fseek(fp, 0, SEEK_END);
    unlock();
    return status;
}

#if 0
void Bucket::drain(Spool *spool)
{
    Stringbuf article, line;
    Artloc loc = { bloc.partition, bloc.bucket_num, 0 };

    /* Mark bucket as draining. */
    lock();
    bucket_type = BUCKET_DRAINING;
    fseek(fp, 0, SEEK_SET);
    putc('d', fp);
    fseek(fp, 0, SEEK_END);
    unlock();

    while (1) {
	loc.offset = find_last_article_offset();
	if (loc.offset < 0)
	    break;

	fseek(fp, loc.offset, SEEK_SET);
	while (read_line_clear(&line, fp)) {
	    if (line.len() && *line.str() == '.') {
		if (line.len() == 1)
		    break;
		article.append(line.str().after(1));
	    } else {
		article.append(line.str());
	    }
	    article.append('\n');
	}

	/* XXX Need to check article list and deal between these two steps. */
	if (spool->refile_article(article.str(), &loc) < 0)
	    break;
	fseek(fp, loc.offset, SEEK_SET);
	if (ftruncate(fileno(fp), loc.offset) < 0)
	    break;
    }

    fseek(fp, 0, SEEK_END);
    if (ftell(fp) == 0) {
	/* Figure out what to do with empty bucket. */
    } else {
	error("Aborting drain of partition %d bucket %d: failed to "
	      "truncate bucket: %m", bloc.partition, bloc.bucket_num);
    }
}
#endif

/* Assumes bucket or pool locked. */
long Bucket::write_article(Headers *headers, String body)
{
    String s;
    Stringbuf buf;
    long offset;

    if (bucket_type == BUCKET_DRAINING)
	crash("Tried to write an article to a draining bucket.");

    offset = ftell(fp);
    for (s = headers->first_header(); s.valid(); s = headers->next_header()) {
	if (s.fwrite(fp) == 0 || putc('\n', fp) == EOF) {
	    unlock();
	    return -1;
	}
    }
    if (putc('\n', fp) == EOF || body.fwrite(fp) == 0) {
	unlock();
	return -1;
    }
    if (fputs(".\n", fp) == EOF || fflush(fp) == EOF
	|| fsync(fileno(fp)) < 0) {
	unlock();
	return -1;
    }
    return offset;
}

/* Assumes draining state.  Assumes fp is currently at end of file. */
long Bucket::find_last_article_offset()
{
    long size = ftell(fp), end, start, pos;
    int reach = 1, block_size = blocksize(fileno(fp)), i;
    enum { NEUTRAL, SEENNEWLINE, SEENDOT, FOUND } state, statemap[3][256];

    /* Is the file empty? */
    if (size == 0)
	return -1;

    /* Initialize simple state machine to search for "\n.\n". */
    for (i = 0; i < 3 * 256; i++)
	statemap[i / 256][i % 256] = NEUTRAL;
    statemap[NEUTRAL]['\n'] = SEENNEWLINE;
    statemap[SEENNEWLINE]['.'] = SEENDOT;
    statemap[SEENNEWLINE]['\n'] = SEENNEWLINE;
    statemap[SEENDOT]['\n'] = FOUND;

    /* Scan backwards exponentially until there aren't three characters left
     * to look at. */
    end = size - 3; /* Don't get trailing "\n.\n". */
    while (end > 2) {
	start = size - (reach * block_size);
	start = (start < 0) ? 0 : start;
	start -= (start % block_size);
	fseek(fp, start, SEEK_SET);

	pos = start;
	state = NEUTRAL;
	while (pos < end) {
	    state = statemap[state][(unsigned char) getc(fp)];
	    pos++;
	    if (state == FOUND)
		return pos;
	}

	end = start + 2; /* +2 in case "\n.\n" crosses block boundary */
	reach *= 2;
    }

    /* "\n.\n" doesn't appear in the file; there's only one article. */
    return 0;
}

Article_list::Article_list(Bucket *bucket, FILE *fp, long offset)
: article(bucket, fp, offset)
{
    refs = 1;
}
