/* Config.cc: The configuration file. */

#include "threads.h"
#include "util.h"
#include "platform.h"
#include "error.h"

#include "Config.h"
#include "String.h"
#include "Wildcard_list.h"
#include "Newsfeed.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <netinet/in.h>
int gethostname();
}

Config::Config(String config_filename)
: admin_access(2), post_access(3), read_access(3), transfer_access(3),
  expire_times(NULL, 1), feed_groups(2), feed_dists(2), feed_paths(2),
  feed_buffer_sizes(NULL, 1), dists(1), moderators(NULL, 1), newgroup_allow(2)
{
    char buf[128];

    /* Set reasonable defaults. */
    readpost = 0;
    if (gethostname(buf, sizeof(buf)) == 0)
	server_name.store(String(buf, strlen(buf)));
    add_mime_headers = 0;
    nntp_port = service_port("nntp", htons(119));
    admin_port = service_port("", htons(20119));
    path_sendmail.store(SENDMAIL_CMD);
    path_newsgroups_file.store(ETCDIR "/newsgroups");
    path_index_dir.store("/var/spool/news/indices/");
    path_groups_dir.store("/var/spool/news/groups/");
    path_pool_dir.store("/var/spool/news/pool/");
    path_feeds_dir.store("/var/spool/news/feeds/");
    expire_incr = 86400;
    expire_record_for = 14;
    cancel_record_for = 14;
    default_expire_time.min = 1;
    default_expire_time.def = 10;
    default_expire_time.max = 28;
    default_moderator.store("%s@uunet.uu.net");
    feeds_bidirectional = 1;

    /* Read the config file. */
    read_config_file(config_filename);
}

Config::~Config()
{
}

void Config::get_server_name(Stringbuf *name) const
{
    name->append(server_name.str());
}

int Config::any_distribution_accepted(String dist) const
{
    String tmp;
    int pos;

    /* If no distributions were specified, assume we don't limit. */
    if (dists.num_patterns() == 0)
	return 1;

    /* Check list of distributions against our table. */
    while ((pos = dist.find(',')) >= 0) {
	if (dists.match1(tmp.left(pos), 0) == 1)
	    return 1;
	dist.shift(pos + 1);
	dist.skip_whitespace();
    }
    return dists.match1(dist, 0);
}

void Config::get_default_organization(Stringbuf *buf) const
{
    buf->append(organization.str());
}

bool Config::get_add_mime_headers_flag() const
{
    return add_mime_headers;
}

unsigned short Config::get_nntp_port() const
{
    return nntp_port;
}

unsigned short Config::get_admin_port() const
{
    return admin_port;
}

void Config::get_sendmail_file(Stringbuf *filename) const
{
    filename->append(path_sendmail.str());
}

void Config::get_newsgroups_file(Stringbuf *filename) const
{
    filename->append(path_newsgroups_file.str());
}

void Config::get_index_directory(Stringbuf *dirname) const
{
    dirname->append(path_index_dir.str());
}

void Config::get_groups_directory(Stringbuf *dirname) const
{
    dirname->append(path_groups_dir.str());
}

void Config::get_pool_directory(Stringbuf *dirname) const
{
    dirname->append(path_pool_dir.str());
}

void Config::get_feeds_directory(Stringbuf *dirname) const
{
    dirname->append(path_feeds_dir.str());
}

time_t Config::decide_expire_time(String newsgroups, time_t expires_header)
     const
{
    String name;
    Expire_spec spec;
    time_t result, t, now;

    /* Check each newsgroup, and select the maximum result. */
    time(&now);
    result = constrain_time(expires_header, now, &default_expire_time);
    while (newsgroups.len()) {
	name = get_group_name(&newsgroups);
	spec = expire_times.match1(name, default_expire_time);
	t = constrain_time(expires_header, now, &spec);
	if (t > result)
	    result = t;
    }
    return result;
}

int Config::get_expire_record_for() const
{
    return expire_record_for * expire_incr;
}

int Config::get_cancel_record_for() const
{
    return cancel_record_for * expire_incr;
}

int Config::find_in_schema(String name) const
{
    int i;

    for (i = 0; i < schema.len(); i++) {
	if (compare_case(name, schema[i].str()) == 0)
	    return i;
    }
    return -1;
}

const Stringlist *Config::get_schema() const
{
    return &schema;
}

const Stringlist *Config::get_feed_sites() const
{
    return &feed_sites;
}

Wildcard_decision_list *Config::get_feed_groups(String name) const
{
    const String *line = &name;

    return feed_groups.sublist(&line, 1, 1);
}

Wildcard_decision_list *Config::get_feed_dists(String name) const
{
    const String *line = &name;

    return feed_dists.sublist(&line, 1, 1);
}

Wildcard_decision_list *Config::get_feed_paths(String name) const
{
    const String *line = &name;

    return feed_paths.sublist(&line, 1, 1);
}

int Config::get_feed_buffer_size(String name) const
{
    return feed_buffer_sizes.match1(name, 10000);
}

void Config::get_moderator(String group, Stringbuf *buf) const
{
    Stringbuf mod, dash_group(group);
    Stored_string sstr;

    dash_group.substitute(".", "-");
    sstr = moderators.match1(group, default_moderator);
    mod.append(sstr.str());
    mod.substitute("%s", dash_group.str());
    buf->append(mod.str());
}

bool Config::get_newgroup_allow(String group, String from) const
{
    String strings[] = { group, from };

    return newgroup_allow.match(strings, 0);
}

Wildcard_decision_list *Config::get_post_access(String name, String ip) const
{
    const String hostfields[] = { String("hostname"), name };
    const String ipfields[] = { String("ip"), ip };
    const String *const lines[] = { hostfields, ipfields };

    if (readpost)
	return read_access.sublist(lines, 2, 2);
    else
	return post_access.sublist(lines, 2, 2);
}

Wildcard_decision_list *Config::get_read_access(String name, String ip) const
{
    const String hostfields[] = { String("hostname"), name };
    const String ipfields[] = { String("ip"), ip };
    const String *const lines[] = { hostfields, ipfields };

    return read_access.sublist(lines, 2, 2);
}

Wildcard_decision_list *Config::get_transfer_access(String name, String ip)
    const
{
    const String hostfields[] = { String("hostname"), name };
    const String ipfields[] = { String("ip"), ip };
    const String *const lines[] = { hostfields, ipfields };

    return transfer_access.sublist(lines, 2, 2);
}

bool Config::get_admin_access(String name, String ip) const
{
    const String hostfields[] = { String("hostname"), name };
    const String ipfields[] = { String("ip"), ip };
    const String *const lines[] = { hostfields, ipfields };

    return admin_access.multimatch(&lines[0], 2, 0);
}

void Config::read_config_file(String filename)
{
    static struct {
	String name;
	void (Config::*func)(String s);
    } sections[] = {
	{ String("access"),		Config::section_access },
	{ String("expire"),		Config::section_expire },
	{ String("feeds"),		Config::section_feeds },
	{ String("general"),		Config::section_general },
	{ String("moderators"),		Config::section_moderators },
	{ String("newgroup"),		Config::section_newgroup },
	{ String("paths"),		Config::section_paths },
	{ String("schema"),		Config::section_schema },
    };
    const int NUM_SECTIONS = (sizeof(sections) / sizeof(*sections));

    Stringbuf cfilename(filename), linebuf;
    String line;
    FILE *fp;
    int section = -1, i;

    /* Open the config file; if there is none, just run with the defaults. */
    fp = fopen(cfilename.cstring(), "r");
    if (!fp) {
	warn("Can't read config file %s", filename);
	return;
    }

    while (read_line_clear(&linebuf, fp) == 0) {
	line = linebuf.str();
	line.strip_whitespace();

	/* If it's not a blank line or a comment, process it. */
	if (line.len() && *line != '#') {

	    if (line.offset_from(linebuf.str()) == 0) {
		/* There's no leading whitespace; it's a new section header. */
		for (i = 0; i < NUM_SECTIONS; i++) {
		    if (compare_case(line, sections[i].name) == 0)
			break;
		}
		if (i < NUM_SECTIONS)
		    section = i;
		else
		    warn("Invalid section header %s in config file.", line);
	    } else {
		/* There's leading whitespace; call the section function. */
		if (section != -1)
		    (this->*sections[section].func)(line);
		else
		    warn("No initial section header in config file.");
	    }
	}
    }

    fclose(fp);
}

/* Commands:	read=post
 * 		[dis]allow <type> <pattern> admin
 *		[dis]allow <type> <pattern> post <newsgroup-pattern> ...
 *		[dis]allow <type> <pattern> read <newsgroup-pattern> ...
 *		[dis]allow <type> <pattern> transfer <newsgroup-pattern> ... */
void Config::section_access(String s)
{
    String w[4], rest, group;
    bool whether;
    Wildcard_decision_list *list;

    rest = s.get_words(w, 4);

    /* Check if one word, "read=post". */
    if (compare_case(w[0], "read=post") == 0 && w[1].len() == 0) {
	readpost = 1;
	return;
    }

    /* At least four words, first word is "allow" or "disallow". */
    if (compare_case(w[0], "allow") == 0 && w[2].len() != 0) {
	whether = 1;
    } else if (compare_case(w[0], "diallow") == 0 && w[2].len() != 0) {
	whether = 0;
    } else {
	warn("Invalid reader-access line in config file: %s", s);
	return;
    }

    if (compare_case(w[3], "admin") == 0 && rest.len() == 0) {
	admin_access.add_pattern(&w[1], whether);
	return;
    } else if (compare_case(w[2], "post") == 0) {
	list = &post_access;
    } else if (compare_case(w[2], "read") == 0) {
	list = &read_access;
    } else if (compare_case(w[2], "transfer") == 0) {
	list = &transfer_access;
    } else {
	warn("Invalid reader-access line in config file: %s", s);
	return;
    }

    /* Add an entry for each word starting at the fourth. */
    for (group = rest.word(); group.len(); group = rest.word()) {
	String fields[] = { w[1], w[2], group };

	list->add_pattern(fields, whether);
    }
}

/* Commands:	<pattern> {never | <min> <default> <max>}
 *		default {never | <min> <default> <max>}
 *		increment <seconds>
 *		record cancel <incrs>
 *		record expire <incrs> */
void Config::section_expire(String s)
{
    Expire_spec spec;
    String w[4], rest;

    rest = s.get_words(w, 4);

    /* Check if two words "increment <seconds>". */
    if (compare_case(w[0], "increment") == 0 && w[1].len() && !w[2].len()) {
	expire_incr = w[2].tonum();
	return;
    }

    /* Check if three words beginning with "record". */
    if (compare_case(w[0], "record") == 0 && w[2].len() && !w[3].len()) {
	if (compare_case(w[1], "cancel") == 0)
	    cancel_record_for = w[2].tonum();
	else if (compare_case(w[1], "expire") == 0)
	    expire_record_for = w[2].tonum();
	else
	    warn("Invalid expires line in config file: %s", s);
	return;
    }

    /* Otherwise it's an expire time specification for <pattern> or default. */
    if (compare_case(w[1], "never") == 0 && w[2].len() == 0) {
	spec.min = spec.def = spec.max = -1;
    } else if (w[3].len() > 0 && rest.len() == 0) {
	spec.min = w[1].tonum();
	spec.def = w[2].tonum();
	spec.max = w[3].tonum();
    } else {
	warn("Invalid expires line in config file: %s", s);
	return;
    }

    /* Set the default specification or add one to the wildcard list. */
    if (compare_case(w[0], "default") == 0)
	default_expire_time = spec;
    else
	expire_times.add_pattern(&w[0], spec);
}

/* Commands:	outgoing
 *		bidirectional
 *		site <hostname>
 *		buffer <site-pattern> for <articles>
 *		[don't|only] feed <site-pattern> group <group-pattern> ...
 *		[don't|only] feed <site-pattern> dist <dist-pattern> ...
 *		[don't|only] feed <site-pattern> path <path-pattern> ...
 */
void Config::section_feeds(String s)
{
    String w[3], groups, group, rest, line = s;
    bool whether = 1, only = 0;
    Wildcard_decision_list *list;

    /* Get first word, check for "don't" or "only" followed by "feed". */
    rest = s.get_words(w, 1);
    if (compare_case(w[0], "don't") == 0) {
	whether = 0;
	s = rest;
    } else if (compare_case(w[0], "only") == 0) {
	only = 1;
	s = rest;
    }
    s.get_words(w, 1);
    if ((only || !whether) && compare_case(w[0], "feed") != 0) {
	warn("Invalid feeds line: %s", line);
	return;
    }

    rest = s.get_words(w, 3);

    /* Check if one word "outgoing" or "bidirectional". */
    if (compare_case(w[0], "outgoing") == 0 && w[1].len() == 0) {
	feeds_bidirectional = 0;
	return;
    } else if (compare_case(w[0], "bidirectional") == 0 && w[1].len() == 0) {
	feeds_bidirectional = 1;
	return;
    }

    /* Check if two words "site <hostname>". */
    if (compare_case(w[0], "site") == 0 && w[1].len() && !w[2].len()) {

	if (feed_sites.find_case(w[1]) >= 0) {
	    warn("Duplicate feed site \"%s\"", w[1]);
	} else {
	    String fields[] = { w[1], w[1] };

	    /* Add a site to the feed sites list and add a pattern to
	     * the feed paths to exlude a path component equal to the machine
	     * name. */
	    feed_sites.add(w[1]);
	    feed_paths.add_pattern(fields, 0);
	}

	/* If the user hasn't set "outgoing", add this host to the transfer
	 * access list as both a hostname and an IP address. */
	if (feeds_bidirectional) {
	    String as_name[] = { w[1], "*", "*" };
	    String as_ip[] = { "*", w[1], "*" };

	    transfer_access.add_pattern(as_name, 1);
	    transfer_access.add_pattern(as_ip, 1);
	}
	return;
    }

    /* Check if four words "buffer <site-pattern> for <articles>". */
    if (compare_case(w[0], "buffer") == 0 && compare_case(w[2], "for") == 0)
	feed_buffer_sizes.add_pattern(&w[1], rest.tonum());

    if (compare_case(w[0], "feed") != 0) {
	warn("Invalid feeds line: %s", line);
	return;
    }

    if (compare_case(w[2], "group") == 0) {
	list = &feed_groups;
    } else if (compare_case(w[2], "dist") == 0) {
	list = &feed_dists;
    } else if (compare_case(w[2], "path") == 0) {
	list = &feed_paths;
    } else {
	warn("Invalid feed line: %s", line);
	return;
    }

    /* Add an entry for each group (each word starting from the fourth). */
    groups = rest;
    for (group = groups.word(); group.len(); group = groups.word()) {
	String fields[] = { w[1], group };

	list->add_pattern(fields, whether);
    }

    if (only) {
	/* Go back and add a negative entry for site "*" for each group. */
	for (group = rest.word(); group.len(); group = rest.word()) {
	    String fields[] = { "*", group };

	    list->add_pattern(fields, 0);
	}
    }
}

/* Commands:	name <server name>
 *		mime-headers {yes|no}
 *		distributions <pattern> ...
 *		organization <organization>
 *		nntp_port <port>
 *		admin_port <port>
 */
void Config::section_general(String s)
{
    String w[2], rest;

    rest = s.get_words(w, 2);

    if (compare_case(w[0], "name") == 0 && w[1].len() && !rest.len()) {
	/* "Name" followed by one word; set the server name. */
	server_name.store(w[1]);
    } else if (compare_case(w[0], "mime-headers") == 0 &&
	       w[1].len() && !rest.len()) {
	add_mime_headers = (compare_case(w[1], "y") == 0 ||
			    compare_case(w[1], "yes") == 0);
    } else if (compare_case(w[0], "distributions") == 0 && w[1].len()) {
	for (; w[1].len(); w[1] = rest.word())
	    dists.add_pattern(&w[1], 1);
    } else if (compare_case(w[0], "organization") == 0 &&
	       w[1].len() && !rest.len()) {
	organization.store(w[1]);
    } else if (compare_case(w[0], "nntp_port") == 0 && w[1].len() &&
	       !rest.len()) {
	nntp_port = htons(w[1].tonum());
    } else if (compare_case(w[0], "admin_port") == 0 && w[1].len() &&
	       !rest.len()) {
	admin_port = htons(w[1].tonum());
    } else {
	warn("Unknown general command in config file: %s", s);
    }
}

/* Commands:	<pattern> <moderator>
 *		default <moderator> */
void Config::section_moderators(String s)
{
    String w[2], rest;

    /* Must be two words. */
    rest = s.get_words(w, 2);
    if (w[1].len() == 0 || rest.len() > 0) {
	warn("Invalid moderators line in config file: %s", s);
	return;
    }

    if (compare_case(w[0], "default") == 0) {
	default_moderator.store(w[1]);
    } else {
	moderators.add_pattern(&w[0], Stored_string(w[1]));
    }
}

/* Commands:	allow <email-pattern> <group-pattern> ...
 *		disallow <email-pattern> <group-pattern> ... */
void Config::section_newgroup(String s)
{
    String w[2], rest, group;
    bool whether;

    rest = s.get_words(w, 2);

    /* Figure out whether we want to allow or disallow. */
    if (compare_case(w[0], "allow") == 0 && rest.len() != 0) {
	whether = 1;
    } else if (compare_case(w[0], "disallow") == 0 && rest.len() != 0) {
	whether = 0;
    } else {
	warn("Invalid newgroup line in config file: %s", s);
	return;
    }

    /* Add an entry for each word starting at the third. */
    for (group = rest.word(); group.len(); group = rest.word()) {
	String fields[] = { w[1], group };

	newgroup_allow.add_pattern(fields, whether);
    }
}

/* Commands:	index <pathname>
 *		groups <pathname>
 *		pool <pathname>
 *		grouplist <pathname> */
void Config::section_paths(String s)
{
    struct {
	String name;
	Stored_string *str;
	int dir_flag;
    } paths[] = {
	{ String("sendmail"),	&path_sendmail,		0 },
	{ String("newsgroups"),	&path_newsgroups_file,	0 },
	{ String("index"),	&path_index_dir,	1 },
	{ String("groups"),	&path_groups_dir,	1 },
	{ String("pool"),	&path_pool_dir,		1 },
	{ String("feeds"),	&path_feeds_dir,	1 }
    };
    const int NUM_PATHS = (sizeof(paths) / sizeof(*paths));

    Stringbuf path;
    String w[2], rest;
    int i;

    /* The line must be two words. */
    rest = s.get_words(w, 2);
    if (w[1].len() == 0 || rest.len() != 0) {
	warn("Invalid paths line in config file: %s", s);
	return;
    }

    for (i = 0; i < NUM_PATHS; i++) {
	if (compare_case(w[0], paths[i].name) == 0)
	    break;
    }
    if (i < NUM_PATHS) {
	path.store(w[1]);
	if (paths[i].dir_flag && path.str()[path.len() - 1] != '/')
	    path.append('/');
	paths[i].str->store(path.str());
    } else {
	warn("Unknown path in config file: %s", s);
    }
}

/* Commands:	<header>
 *		<header> full */
void Config::section_schema(String s)
{
    String w[2], rest;
    Stringbuf entry;

    rest = s.get_words(w, 2);

    if (compare_case(w[2], "full") == 0 && rest.len() == 0) {
	/* Insert a colon at the beginning of the header entry. */
	entry.append(':');
    } else if (w[2].len() != 0) {
	warn("Unknown schema command in config file: %s", s);
	return;
    }

    /* Add the new schema entry to the existing schemas. */
    entry.append(w[1]);
    schema.add(entry.str());
}

time_t Config::constrain_time(time_t t, time_t now, const Expire_spec *spec)
    const
{
    time_t min, def, max;

    min = ((now / expire_incr) + spec->min) * expire_incr;
    def = ((now / expire_incr) + spec->def) * expire_incr;
    max = ((now / expire_incr) + spec->max) * expire_incr;
    return (t == -1) ? def : (t < min) ? min : (t > max) ? max : t;
}

int Expire_spec::operator ==(const Expire_spec &spec)
{
    return (min == spec.min && def == spec.def && max == spec.max);
}

