/* connection.cc: Handle client connections. */

#include "threads.h"
#include "error.h"

#include "Connection.h"
#include "Stringbuf.h"
#include "String.h"
#include "util.h"

extern "C" {
#include <ctype.h>
#include <string.h>
#include <stdarg.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <netdb.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>

int getpeername();
}

const int OUTBUF_MAX_LEN = 4000;

Connection::Connection(int fd)
{
    struct hostent *hostent;
    struct hostent_answer buf;
    struct sockaddr_in peer_addr;
    int len = sizeof(peer_addr);
    char ipbuf[64];

    client_socket = fd;
    flags = fcntl(fd, F_GETFL);
    if (getpeername(fd, (struct sockaddr *) &peer_addr, &len) >= 0) {
	inet_ntoa_r(peer_addr.sin_addr, ipbuf, sizeof(ipbuf));
	peer_ip.store(String(ipbuf));
	hostent = gethostbyaddr_r((char *) &peer_addr.sin_addr,
				  sizeof(struct in_addr), AF_INET, &buf);
	if (hostent) {
	    peer_name.store(String(hostent->h_name));
	    hostent = gethostbyname_r(peer_name.cstring(), &buf);
	    if (hostent && memcmp(&hostent->h_addr, &peer_addr, len) == 0)
		peer_name_status = NAME_OK;
	    else
		peer_name_status = NAME_FORGED;
	} else {
	    peer_name.store(peer_ip.str());
	    peer_name_status = NAME_NONE;
	}
    } else {
	peer_name_status = NAME_NOTSOCKET;
	peer_name.store("<not socket>");
	peer_ip.store("<not socket>");
    }
    /* Temporary. */
    log("Connection received from %s (%s).", peer_name.str(), peer_ip.str());
}

Connection::~Connection()
{
    flush_outbuf();
    close(client_socket);
}

int Connection::read_line(Stringbuf *line)
{
    int pos, start = 0;

    while (1) {
	/* Scan any unscanned text in inbuf for newlines. */
	pos = inbuf.str().after(start).find('\n');
	if (pos != -1)
	    break;

	/* Read more text into inbuf and try again. */
	flush_outbuf();
	start = inbuf.len();
	if (get_data() <= 0)
	    return -1;
    }

    line->append(inbuf.str().left(start + pos));
    if (line->str()[line->len() - 1] == '\r')
	line->trunc(line->len() - 1);
    inbuf.shift(start + pos + 1);
    return 0;
}

int Connection::read_line_clear(Stringbuf *line)
{
    line->clear();
    return read_line(line);
}

void Connection::write(String str)
{
    int space_left;

    /* Make sure the buffer length doesn't go above BUFFER_MAX_LEN. */
    while (outbuf.len() + str.len() > OUTBUF_MAX_LEN) {
	space_left = OUTBUF_MAX_LEN - outbuf.len();
	outbuf.append(str.left(space_left));
	flush_outbuf();
	str.shift(space_left);
    }

    outbuf.append(str);
}

void Connection::writef(String fmt, ...)
{
    va_list args;
    Stringbuf buf;

    va_start(args, fmt);
    buf.vformat(fmt, args);
    va_end(args);
    write(buf.str());
}

void Connection::flush_outbuf()
{
    int count;
    String str = outbuf.str();

    while (str.len()) {
	count = str.write(client_socket);
	if (count < 0)
	    break;
	str.shift(count);
    }

    outbuf.clear();
}

int Connection::get_data()
{
    char buf[4096];
    int count;

    /* Set the file descriptor non-blocking to allow fast streaming. */
    if (outbuf.len()) {
	fcntl(client_socket, F_SETFL, flags & O_NONBLOCK);
	count = read(client_socket, buf, sizeof(buf));
	fcntl(client_socket, F_SETFL, flags);
	if (count != -1 || errno != EAGAIN)
	    return count;
	flush_outbuf();
    }
    count = ::read(client_socket, buf, sizeof(buf));
    if (count > 0)
	inbuf.append(String(buf, count));
    return count;
}

