/* dbz.h: Declarations for dbz database index. */

#ifndef DBZ_H
#define DBZ_H

extern "C" {
#include <stdio.h>
#include <sys/types.h>
}

/*
 * ANSI C says an offset into a file is a long, not an off_t, for some
 * reason.  This actually does simplify life a bit, but it's still nice
 * to have a distinctive name for it.  Beware, this is just for readability,
 * don't try to change this.
 */
#define	of_t	long
#define	SOF	(sizeof(of_t))

/* for dbm and dbz */
typedef struct {
	char *dptr;
	int dsize;
} datum;

/*
 * Stdio buffer for .pag reads.  Buffering more than about 16 does not help
 * significantly at the densities we try to maintain, and the much larger
 * buffers that most stdios default to are much more expensive to fill.
 * With small buffers, stdio is performance-competitive with raw read(),
 * and it's much more portable.
 */
#ifndef NPAGBUF
#define	NPAGBUF	16
#endif

/*
 * Stdio buffer for base-file reads.  Message-IDs (all news ever needs to
 * read) are essentially never longer than 64 bytes, and the typical stdio
 * buffer is so much larger that it is much more expensive to fill.
 */
#ifndef SHISTBUF
#define	SHISTBUF	64
#endif

/*
 * We read configuration info from the .dir file into this structure,
 * so we can avoid wired-in assumptions for an existing database.
 *
 * Among the info is a record of recent peak usages, so that a new table
 * size can be chosen intelligently when rebuilding.  10 is a good
 * number of usages to keep, since news displays marked fluctuations
 * in volume on a 7-day cycle.
 */
struct dbzconfig {
	int olddbz;		/* .dir file empty but .pag not? */
	of_t tsize;		/* table size */
#	ifndef NMEMORY
#	define	NMEMORY	10	/* # days of use info to remember */
#	endif
#	define	NUSEDS	(1+NMEMORY)
	of_t used[NUSEDS];	/* entries used today, yesterday, ... */
	int valuesize;		/* size of table values, == SOF */
	int bytemap[SOF];	/* byte-order map */
	char casemap;		/* case-mapping algorithm (see cipoint()) */
	char fieldsep;		/* field separator in base file, if any */
	of_t tagenb;		/* unshifted tag-enable bit */
	of_t tagmask;		/* unshifted tag mask */
	int tagshift;		/* shift count for tagmask and tagenb */
};

/*
 * Data structure for recording info about searches.
 */
struct searcher {
	of_t place;		/* current location in file */
	int tabno;		/* which table we're in */
	int run;		/* how long we'll stay in this table */
#		ifndef MAXRUN
#		define	MAXRUN	100
#		endif
	long hash;		/* the key's hash code (for optimization) */
	of_t tag;		/* tag we are looking for */
	int seen;		/* have we examined current location? */
	int aborted;		/* has i/o error aborted search? */
};

class Dbz {
    of_t tagbits;		/* pre-shifted tag mask */
    of_t taghere;		/* pre-shifted tag-enable bit */
    of_t tagboth;		/* tagbits|taghere */
    struct dbzconfig conf;
    int incore;
    of_t pagbuf[NPAGBUF];	/* only needed if !NOBUFFER && _IOFBF */
    char basebuf[SHISTBUF];	/* only needed if _IOFBF exists */
    struct searcher srch;
    struct searcher *prevp;
    int mybmap[SOF];		/* my byte order (see mybytemap()) */
    int bytesame;		/* is database order same as mine? */
    FILE *basef;		/* descriptor for base file */
    char *basefname;		/* name for not-yet-opened base file */
    FILE *dirf;			/* descriptor for .dir file */
    int dirronly;		/* dirf open read-only? */
    FILE *pagf;			/* descriptor for .pag file */
    of_t pagpos;		/* posn in pagf; only search may set != -1 */
    int pagronly;		/* pagf open read-only? */
    of_t *corepag;		/* incore version of .pag file, if any */
    FILE *bufpagf;		/* well-buffered pagf, for incore rewrite */
    int written;		/* has a store() been done? */
    of_t key_ptr;		/* return value of fetch points here */

    int getconf(FILE *df, FILE *pf, struct dbzconfig *cp);
    int putconf(FILE *f, struct dbzconfig *cp);
    void start(struct searcher *sp, datum *kp, struct searcher *osp);
    of_t search(struct searcher *sp);
    int set(struct searcher *sp, of_t value);
    char *mapcase(char *dst, char *src, size_t siz);
    char *cipoint(char *s, size_t size);
    FILE *latebase();
    of_t *getcore(FILE *f);
    int putcore(of_t *tab, FILE *f);
    int okayvalue(of_t value);

  public:
    Dbz();
    int set_incore(int value);
    int init(const char *name);
    int fresh(const char *name, long size, int fs, int cmap, of_t tagmask);
    int again(const char *name, const char *oldname);
    int sync();
    int close();
    int cancel();
    datum fetch(datum key);
    datum fetch_case(datum key);
    int store(datum key, datum data);
    int store_case(datum key, datum value);
};

long dbzsize(long contents);
int dbzdebug(int value);

/*
 * In principle we could handle unlimited-length keys by operating a chunk
 * at a time, but it's not worth it in practice.  Setting a nice large
 * bound on them simplifies the code and doesn't hurt anything.
 */
#define DBZMAXKEY	255

#endif

