/* Groups.cc: Managing the hierarchy of group files. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Groups.h"
#include "Atomic_file.h"
#include "Newsgroup.h"
#include "String.h"
#include "Config.h"
#include "Connection.h"
#include "Headers.h"
#include "Wildcard_list.h"
#include "Array.h"
#include "Spool.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
}

/* Locking policy: the hash table and all of the ->next and ->in_file fields
 * of the Newsgroup_list structures are under control of the quiescing
 * read/write lock on the Groups object.  All other fields on the
 * Newsgroup_list objects are under control of the lock on that structure.
 * To write the bounds file or newsgroups file, you need a lock on update_lock
 * as well as a read lock on the Groups object, OR a write lock on the Groups
 * object.  (This prevents deadlock while ensuring that only one thread tries
 * to update files at a time.) */

const int MIN_POWER = 9;
const int DELTA = 2;
const int AVERAGE_LINE_LENGTH = 60;

static int check_file(String group_filename, int number, const Artloc *loc,
		      Newsgroup_list *entry, FILE **fp);
static int eliminate_bucket_in_file(const Bucket_loc *bloc, String name,
				    Newsgroup_list *entry);
static int match_distribution(String name, String dist);
static FILE *open_group_file(String name, int check);
static void delete_group_file(String name, unsigned int dirnamelen);
static void *periodic_check_newsgroups(void *arg);

Groups::Groups(const Config *config_arg)
{
    Stringbuf buf;
    unsigned int i;
    FILE *fp;

    config = config_arg;
    config->get_groups_directory(&buf);
    dirname.store(buf.str());
    buf.clear();
    config->get_newsgroups_file(&buf);
    newsgroups_filename.store(buf.str());
    newsgroups_file = new Atomic_file(newsgroups_filename.str());
    buf.format_clear("%s/bounds.list", dirname.str());
    make_directories_in_path(buf.str());
    bounds_file = new Atomic_file(buf.str());

    make_directories_in_path(dirname.str());

    size = adjust_size(0, MIN_POWER, DELTA);
    hashtab = new Newsgroup_list *[size];
    for (i = 0; i < size; i++)
	hashtab[i] = NULL;
    num_groups = 0;

    fp = newsgroups_file->open_read();
    if (fp) {
	read_newsgroups_file(fp);
	fclose(fp);
    } else {
	die("Can't read newsgroups file.");
    }
    read_bounds_file();

    start_thread(periodic_check_newsgroups, this);
}

Groups::~Groups()
{
    unsigned int i;
    Newsgroup_list *entry, *next;

    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = next) {
	    next = entry->next;
	    if (entry->newsgroup)
		delete entry->newsgroup;
	    delete entry;
	}
    }
}

Newsgroup *Groups::find_group(String name)
{
    FILE *fp;
    Newsgroup_list *entry;
    Newsgroup *group;
    Stringbuf filename;

    /* Find the entry for the group in the hash table. */
    entry = follow_aliases(name);
    if (!entry)
	return NULL;

    /* If there's already a newsgroup object for this entry, bump up its
     * reference count and return it. */
    if (entry->newsgroup) {
	group = entry->newsgroup;
	entry->refs++;
	entry->unlock();
	return group;
    }

    /* We have to create the newsgroup object.  Construct its name. */
    construct_group_name_clear(&filename, name);

    /* Open the newsgroup file.  If it's not there, the group is empty. */
    fp = fopen(filename.cstring(), "r");

    /* Create a new newsgroup, note it here, and return it. */
    group = new Newsgroup(name, fp, config, entry->last);
    entry->newsgroup = group;
    entry->refs = 1;
    entry->unlock();
    return group;
}

void Groups::release_group(Newsgroup *group)
{
    Newsgroup_list *entry;

    /* Find the newsgroup in the hash table. */
    entry = follow_aliases(group->name());
    if (!entry)
	crash("Released newsgroup %s not found.", group->name());

    entry->refs--;
    if (entry->refs == 0) {
	entry->newsgroup = NULL;
	delete group;
    }
    entry->unlock();
}

void Groups::reserve_positions(String groups,
			       const Wildcard_decision_list *access,
			       Stringbuf *xref)
{
    Stringbuf xref_buf;
    String name;
    Newsgroup_list *entry;
    int pos;

    while (groups.len()) {

	/* Get the next group which accepts articles. */
	name = get_group_name(&groups);
	if (access && !access->match1(name, 0))
	    continue;
	entry = follow_aliases(name);
	if (!entry)
	    continue;
	if (*entry->flags.str() == 'x' || *entry->flags.str() == 'j' ||
	    (access && !access->match1(entry->name.str(), 0))) {
	    entry->unlock();
	    continue;
	}

	/* Check that it's not a duplicate group. */
	pos = xref_buf.str().find(name);
	if (pos <= 0 || (unsigned int)pos + name.len() >= xref_buf.len() 
	    || xref_buf.str()[pos] != ' '
	    || xref_buf.str()[pos + name.len()] != ':') {
	    /* Reserve an article number in the group and add an Xref entry. */
	    if (entry->first_unreserved <= entry->last)
		entry->first_unreserved = entry->last + 1;
	    xref_buf.format(" %s:%d", name, entry->first_unreserved++);
	}
	entry->unlock();
    }

    xref->append(xref_buf.str());
}

/* check is set to 1 if we're refiling an article, 0 if we're adding one.
 * Refiling takes into account various possible inconsistencies in the
 * grouplist file and is much slower. */
int Groups::add_to_groups(const Headers *headers, int num_lines, int len,
			  const Artloc *loc, int check)
{
    FILE *fp;
    Newsgroup_list *entry;
    String group, field, value, xrefs;
    Stringbuf over_buf, number_buf, filename;
    const Stringlist *schema;
    int number, i, pos;

    /* Build group line. */
    schema = config->get_schema();
    for (i = 0; i < schema->len(); i++) {
	field = (*schema)[i].str();
	if (field.len() && *field == ':') {
	    field.shift(1);
	    over_buf.append(field);
	    over_buf.append(": ");
	}
	if (compare_case(field, "Lines") == 0) {
	    over_buf.format("%d", num_lines);
	} else if (compare_case(field, "Bytes") == 0) {
	    over_buf.format("%d", len);
	} else {
	    value = (*headers)[field];
	    if (value.valid()) {
		over_buf.append(value);
		over_buf.substitute("\n\t", "  ",
				    over_buf.len() - value.len());
	    }
	}
	over_buf.append('\t');
    }

    /* Iterate through the xref header created by reserve_positions(). */
    xrefs = (*headers)["Xref"];
    xrefs.word();
    while (xrefs.len()) {

	/* Get a group name and find the hash table entry for it.  If it
	 * is not in our list, then it might have been rmgrouped, so just
	 * ignore it. */
	pos = xrefs.find(':');
	group = xrefs.left(pos);
	number = xrefs.after(pos + 1).tonum();
	entry = follow_aliases(group);
	if (!entry)
	    continue;
	xrefs.word();

	/* Open the group file.  If re-indexing an article, check to make sure
	 * we don't already have an entry for this article. */
	construct_group_name_clear(&filename, group);
	if (!check) {
	    fp = open_group_file(filename.str(), 0);
	} else if (check_file(filename.str(), number, loc, entry, &fp)) {
	    entry->unlock();
	    continue;
	}
	if (!fp) {
	    error("Can't open group file %s: %m", filename.str());
	    entry->unlock();
	    return -1;
	}

	/* Update the last number and number fields for this article. */
	if (number > entry->last)
	    entry->last = number;
	entry->num++;

	/* Write out a line to the file. */
	number_buf.format_clear("%l\t%d\t%l\t%l\t", number, loc->partition,
				loc->bucket_num, loc->offset);
	number_buf.str().fwrite(fp);
	over_buf.str().fwrite(fp);
	putc('\n', fp);
	fclose(fp);

        /* If someone has the newsgroup open, update that list. */
	if (entry->newsgroup)
	    entry->newsgroup->add_entry(number, loc, over_buf.str());

	entry->unlock();
    }

    return 0;
}

int Groups::write_new_bounds_file()
{
    FILE *fp;
    unsigned int i, count;
    int status;
    Newsgroup_list *entry;
    Stringbuf buf;

    fp = bounds_file->open_new();
    if (!fp)
	return -1;

    /* Write out the new file. */
    read_lock();
    update_lock.lock();
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next) {
	    buf.format_clear("%s %d %d %d\n", entry->name.str(), entry->first,
			     entry->last, entry->num);
	    count = buf.str().fwrite(fp);
	    if (count != buf.len()) {
		error("Can't write data to bounds file: %s");
		fclose(fp);
		update_lock.unlock();
		return -1;
	    }
	}
    }
    read_done();
    fclose(fp);

    status = bounds_file->update();
    update_lock.unlock();
    return status;
}

void Groups::send_grouplist(Connection *conn,
			    const Wildcard_decision_list *access)
{
    unsigned int i;
    Newsgroup_list *entry;
    String flags;

    read_lock();
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next) {
	    if (!access || access->match1(entry->name.str(), 0)) {
		/* For moderated group, omit name of moderator. */
		flags = (*entry->flags.str() == 'm') ? String("m")
		    : entry->flags.str();
		conn->writef("%s %N %N %s\r\n", entry->name.str(), entry->last,
			     entry->first, flags);
	    }
	}
    }
    read_done();
}

void Groups::send_group_descriptions(Connection *conn,
				     const Wildcard_decision_list *access)
{
    unsigned int i;
    Newsgroup_list *entry;

    read_lock();
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next) {
	    if (!access || access->match1(entry->name.str(), 0))
		conn->writef("%s %s\n", entry->name.str(), entry->desc.str());
	}
    }
    read_done();
}

void Groups::send_new_groups(Connection *conn, time_t mintime,
			     String dist, const Wildcard_decision_list *access)
{
    unsigned int i;
    Newsgroup_list *entry;
    String flags;

    read_lock();
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next) {
	    if (entry->creation < mintime)
		continue;
	    if (*dist && !match_distribution(entry->name.str(), dist))
		continue;
	    if (access && !access->match1(entry->name.str(), 0))
		continue;
	    flags = (*entry->flags.str() == 'm') ? String("m")
		: entry->flags.str();
	    conn->writef("%s %N %N %s\r\n", entry->name.str(), entry->last,
			 entry->first, flags);
	}
    }
    read_done();
}

int Groups::eliminate_bucket(const Bucket_loc *bloc)
{
    unsigned int i;
    Newsgroup_list *entry;
    Stringbuf buf;

    read_lock();
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next) {
	    entry->lock();

	    /* Update the newsgroup, if there is one. */
	    if (entry->newsgroup)
		entry->newsgroup->eliminate_bucket(bloc);

	    /* Update the file. */
	    construct_group_name_clear(&buf, entry->name.str());
	    if (eliminate_bucket_in_file(bloc, buf.str(), entry) < 0) {
		entry->unlock();
		read_done();
		return -1;
	    }
	    entry->unlock();
	}
    }
    read_done();

    return write_new_bounds_file();
}

int Groups::verify_post_access(String groups,
			       const Wildcard_decision_list *access,
			       Stringbuf *which, Stringbuf *mod)
{
    String group_name;
    Newsgroup_list *entry;

    while (groups.len()) {
	group_name = get_group_name(&groups);
	entry = follow_aliases(group_name);
	if (!entry || !strchr("ymj", *entry->flags.str()) ||
	    !access->match1(group_name, 0) ||
	    !access->match1(entry->name.str(), 0)) {
	    if (entry)
		entry->unlock();
	    which->append(group_name);
	    return -1;
	}
	if (*entry->flags.str() == 'm' && mod) {
	    if (entry->flags.str()[1])
		mod->append(entry->flags.str().after(1));
	    else
		config->get_moderator(group_name, mod);
	}
	entry->unlock();
    }
    return 0;
}

int Groups::any_groups_moderated(String groups)
{
    String group_name;
    Newsgroup_list *entry;
    bool moderated;

    while (groups.len()) {
	group_name = get_group_name(&groups);
	entry = follow_aliases(group_name);
	moderated = (entry && *entry->flags.str() == 'm');
	if (entry)
	    entry->unlock();
	if (moderated)
	    return 1;
    }
    return 0;
}

void Groups::add_group(String group, String flags, String desc,
		       time_t creation)
{
    Stringbuf buf;
    FILE *fp;

    check_newsgroups_file();
    write_lock();
    update_lock.lock();
    if (add_group_hashtable_entry(group, flags, desc, creation)) {
	/* Entry already existed, update the long way. */
	update_newsgroups_file();
    } else {
	/* Append the new entry to the end. */
	fp = fopen(newsgroups_filename.cstring(), "a");
	if (fp) {
	    buf.format("%s %l %s %s\n", group, (long) creation, flags, desc);
	    buf.str().fwrite(fp);
	    fclose(fp);
	} else {
	    error("Can't open newsgroups file to append group %s.\r\n", group);
	}
    }
    update_lock.unlock();
    write_done();
}

void Groups::merge_groups(FILE *fp)
{
    check_newsgroups_file();
    write_lock();
    read_newsgroups_file(fp);
    update_newsgroups_file();
    write_done();
}

int Groups::cancel_article(String xref)
{
    String xref_entry, group_name;
    Stringbuf filename, line;
    int pos, number, found = 0;
    FILE *fp;
    fpos_t filepos;
    Newsgroup_list *entry;

    xref.strip_whitespace();
    while (xref.len()) {
	/* Get a group name and find the hash table entry for it.  If there's
	 * no hash table, the group might have been rmgrouped, so ignore it. */
	xref_entry = get_delimited_element(&xref, ' ');
	pos = xref_entry.find(':');
	if (pos == -1)
	    break;
	group_name = xref_entry.left(pos);
	number = xref_entry.after(pos + 1).tonum();
	entry = follow_aliases(group_name);
	if (!entry)
	    continue;

	/* Construct the group file name and open it. */
	construct_group_name_clear(&filename, group_name);
	fp = fopen(filename.cstring(), "r+");
	if (!fp) {
	    error("Can't open group file to cancel article %s/%d", group_name,
		  number);
	    entry->unlock();
	    return -1;
	}

	fgetpos(fp, &filepos);
	while (read_line_clear(&line, fp) >= 0) {
	    if (line.len() && line.str().tonum() == number) {
		fsetpos(fp, &filepos);
		putc('C', fp);
		found = 1;
		break;
	    }
	    fgetpos(fp, &filepos);
	}
	fclose(fp);

	if (found) {
	    if (entry->newsgroup)
		entry->newsgroup->eliminate_article(number);
	    entry->num--;
	}

	entry->unlock();
    }

    return 0;
}

/* Don't call this with a read or write lock; it gets a write lock. */
void Groups::check_newsgroups_file()
{
    struct stat statbuf;
    Newsgroup_list *entry;
    FILE *fp;
    unsigned int i;

    /* Get information about the newsgroups file. */
    if (stat(newsgroups_filename.cstring(), &statbuf) < 0) {
	error("Can't stat newsgroups file.");
	return;
    }

    /* If the newsgroups file hasn't been modified since we last read it,
     * we don't need to read it. */
    if (statbuf.st_mtime < newsgroups_up_to_date)
	return;

    write_lock();

    /* Set in_file marks to 0. */
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next)
	    entry->in_file = 0;
    }

    fp = newsgroups_file->open_read();
    if (fp) {
	/* Read the newsgroups file, setting in_file marks to 1. */
	read_newsgroups_file(fp);
	fclose(fp);

	/* Delete any groups which disappeared. */
	for (i = 0; i < size; i++) {
	    for (entry = hashtab[i]; entry; entry = entry->next) {
		if (entry->in_file == 0)
		    ;
	    }
	}
    }

    write_done();
}

/* Assumes write lock or initialization phase. */
void Groups::read_newsgroups_file(FILE *fp)
{
    Stringbuf buf;
    String w[3], desc;
    time_t now;

    time(&now);
    while (read_line_clear(&buf, fp) >= 0) {

	/* Ignore comments and blank lines. */
	if (!buf.len() || *buf.str() == '#')
	    continue;

	desc = buf.str().get_words(w, 3);

	if (check_flags(w[2]) < 0) {
	    warn("Invalid newsgroups line: %s\n", buf.str());
	    continue;
	}
	add_group_hashtable_entry(w[0], w[2], desc, w[1].tonum());
    }

    newsgroups_up_to_date = now;
}

void Groups::read_bounds_file()
{
    FILE *fp;
    Newsgroup_list *entry;
    Stringbuf buf, name;
    String w[4];

    fp = bounds_file->open_read();
    if (!fp)
	return;

    while (read_line_clear(&buf, fp) >= 0) {
	buf.str().get_words(w, 4);

	entry = find_table_entry(w[0]);
	if (!entry) {
	    construct_group_name_clear(&name, w[0]);
	    delete_group_file(name.str(), dirname.len());
	    continue;
	}

	/* Read in the first and last indices and the number of articles. */
	entry->first = w[1].tonum();
	entry->last = w[2].tonum();
	entry->num = w[3].tonum();
	entry->unlock();
    }

    fclose(fp);
}

/* Internal method--assumes write lock or initialization phase. */
int Groups::add_group_hashtable_entry(String name, String flags, String desc,
				      time_t creation)
{
    Newsgroup_list *entry;
    int hashval;

    entry = find_table_entry(name);
    if (entry) {

	/* Replace fields in the current entry. */
	entry->name.store(name);
	entry->flags.store(flags);
	entry->desc.store(desc);
	entry->in_file = 1;
	entry->unlock();
	return 1;

    } else {

	/* Expand the hash table if it's getting too big. */
	if (num_groups >= size * 4)
	    expand_hash_table();

	/* Make a new newsgroup entry. */
	entry = new Newsgroup_list;
	entry->name.store(name);
	entry->first = 1;
	entry->last = 0;
	entry->num = 0;
	entry->first_unreserved = 1;
	entry->creation = creation;
	entry->flags.store(flags);
	entry->desc.store(desc);
	entry->in_file = 1;
	entry->newsgroup = NULL;

	/* Insert it into the hash table. */
	hashval = entry->name.str().hash_case() % size;
	entry->next = hashtab[hashval];
	hashtab[hashval] = entry;

	num_groups++;
	return 0;
    }
}

/* Internal method--assumes lock or initialization phase. */
void Groups::expand_hash_table()
{
    unsigned int new_size, i, hashval;
    Newsgroup_list **new_tab, *entry, *next;

    /* Make space for a new hash table. */
    new_size = adjust_size(size + 1, MIN_POWER, DELTA);
    new_tab = new Newsgroup_list *[new_size];
    for (i = 0; i < new_size; i++)
	new_tab[i] = NULL;

    /* Rehash each entry in the old table. */
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = next) {
	    next = entry->next;
	    hashval = entry->name.str().hash_case() % new_size;
	    entry->next = new_tab[hashval];
	    new_tab[hashval] = entry;
	}
    }

    /* Replace the old table with the new one. */
    delete[] hashtab;
    hashtab = new_tab;
    size = new_size;
}

Newsgroup_list *Groups::find_table_entry(String name)
{
    Newsgroup_list *entry;
    int hashval;

    hashval = name.hash_case() % size;
    read_lock();
    for (entry = hashtab[hashval]; entry; entry = entry->next) {
	if (compare_case(entry->name.str(), name) == 0)
	    break;
    }
    if (entry)
	entry->lock();
    read_done();
    return entry;
}

Newsgroup_list *Groups::follow_aliases(String name)
{
    Newsgroup_list *entry, *alias;

    /* Find the table entry.  If the flags don't start with '=', we're done. */
    entry = find_table_entry(name);
    if (!entry || *entry->flags.str() != '=')
	return entry;

    /* We have an alias; look up the real group name. */
    alias = follow_aliases(entry->flags.str().after(1));
    entry->unlock();
    return alias;
}

/* Assumes update lock and write lock. */
void Groups::update_newsgroups_file()
{
    FILE *old_fp, *new_fp;
    Stringbuf line, buf;
    String w[3], whitespace[3], desc;
    Newsgroup_list *entry;
    unsigned int i;

    /* Open file handles for the old and new newsgroups files. */
    old_fp = newsgroups_file->open_read();
    if (!old_fp) {
	error("Can't open old newsgroups file for update.");
	return;
    }
    new_fp = newsgroups_file->open_new();
    if (!new_fp) {
	fclose(old_fp);
	error("Can't open new newsgroups file for update.");
	return;
    }

    /* Set all in_file fields to 0. */
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next)
	    entry->in_file = 0;
    }

    /* For each newsgroup in old_fp, write a line in new_fp and set the
     * newsgroups's in_file field to 1. */
    while (read_line_clear(&line, old_fp) >= 0) {

	if (!line.len() || *line.str() == '#') {
	    line.str().fwrite(new_fp);
	    putc('\n', new_fp);
	    continue;
	}

	desc = line.str().get_words(w, 3, whitespace);

	/* Write out the entry, using the same white space as in the original
	 * newsgroups file. */
	entry = find_table_entry(w[0]);
	if (entry) {
	    entry->in_file = 1;
	    if (!whitespace[0].len())
		whitespace[0] = " ";
	    if (!whitespace[1].len())
		whitespace[1] = " ";
	    if (!whitespace[2].len() && entry->desc.len())
		whitespace[2] = " ";
	    else if (!entry->desc.len())
		whitespace[2] = "";
	    buf.format_clear("%s%s%l%s%s%s%s\n", entry->name.str(),
			     whitespace[0], (long) entry->creation,
			     whitespace[1], entry->flags.str(), whitespace[2],
			     entry->desc.str());
	    buf.str().fwrite(new_fp);
	}
    }

    /* Look for new newsgroups and write them out. */
    for (i = 0; i < size; i++) {
	for (entry = hashtab[i]; entry; entry = entry->next) {
	    if (!entry->in_file) {
		whitespace[0] = (entry->desc.len()) ? " " : "";
		buf.format_clear("%s %l %s%s%s\n", entry->name.str(),
				 (long) entry->creation, entry->flags.str(),
				 whitespace[0], entry->desc.str());
		buf.str().fwrite(new_fp);
	    }
	}
    }

    fclose(old_fp);
    fclose(new_fp);
    newsgroups_file->update();
}

void Groups::construct_group_name_clear(Stringbuf *buf, String name)
{
    int pos;

    buf->store(dirname.str());
    while ((pos = name.find('.')) >= 0) {
	buf->format("%s/", name.left(pos));
	name.shift(pos + 1);
    }
    buf->format("%s.info", name);
}

/* Read through a newsgroup file checking for an article number.  This happens
 * during re-indexing and refiling of articles.  During re-indexing, we don't
 * know whether the number of articles or the last article number in the bounds
 * file are correct, so we need to update the entry (which has been passed to
 * us, locked). */
static int check_file(String group_filename, int number, const Artloc *loc,
		      Newsgroup_list *entry, FILE **fp)
{
    Stringbuf line, number_buf;
    String w[4], over;
    int line_number, found = 0, rewrite = 0;
    FILE *newfp;

    *fp = open_group_file(group_filename, 1);
    if (!*fp)
	return 0;

    /* Find out whether the article is already in the file; update entry->num
     * and entry->last while we're here. */
    entry->num = entry->last = 0;
    while (read_line_clear(&line, *fp) >= NULL) {
	line.str().get_words(w, 4);
	line_number = w[0].tonum();
	if (line_number == number) {
	    found = 1;
	    rewrite = (w[1].tonum() != loc->partition
		       || w[2].tonum() != loc->bucket_num
		       || w[3].tonum() != loc->offset);
	}
	if (line_number > entry->last)
	    entry->last = line_number;
	entry->num++;
    }

    /* Rewrite the group file if we're refiling. */
    if (rewrite) {
	Atomic_file file(group_filename);

	rewind(*fp);
	newfp = file.open_read();
	if (!newfp) {
	    fclose(*fp);
	    *fp = NULL;
	    return 0;
	}
	newfp = file.open_new();
	while (read_line_clear(&line, *fp) >= NULL) {
	    over = line.str().get_words(w, 4);
	    if (w[0].tonum() == number) {
		number_buf.format_clear("%l\t%d\t%l\t%l\t", loc->partition,
					loc->bucket_num, loc->offset);
	    } else {
		number_buf.format_clear("%l\t%d\t%l\t%l\t", w[1].tonum(),
					w[2].tonum(), w[3].tonum());
	    }
	    number_buf.str().fwrite(newfp);
	    over.fwrite(newfp);
	    putc('\n', newfp);
	}
	fclose(newfp);
	fclose(*fp);
	*fp = NULL;
	if (file.update() < 0)
	    return 0;
    }

    return found;
}

/* Assumes lock on entry. */
static int eliminate_bucket_in_file(const Bucket_loc *bloc, String name,
				    Newsgroup_list *entry)
{
    Atomic_file groupfile(name);
    FILE *in_fp, *out_fp;
    Stringbuf line;
    String w[3];
    int number, num = 0, first = 0;

    /* Open the current and new group files. */
    in_fp = groupfile.open_read();
    if (!in_fp)
	return 0;
    out_fp = groupfile.open_new();
    if (!out_fp) {
	fclose(in_fp);
	return -1;
    }

    /* Copy the group file, filtering lines with the bucket in question. */
    while (read_line_clear(&line, in_fp) >= 0) {
	line.str().get_words(w, 3);
	number = w[0].tonum();
	if (w[1].tonum() == bloc->partition
	    && w[2].tonum() == bloc->bucket_num)
	    continue;
	if (line.str().fwrite(out_fp) < line.len()
	    || putc('\n', out_fp) == EOF) {
	    fclose(in_fp);
	    fclose(out_fp);
	    error("Can't write new group file for %s: %m", name);
	    return -1;
	}
	num++;
	if (!first || number < first)
	    first = number;
    }
    fclose(in_fp);
    fclose(out_fp);

    /* Atomically update the group file. */
    if (groupfile.update() < 0)
	return -1;

    /* Update the fields in entry and return success. */
    entry->num = num;
    entry->first = (first > 0) ? first : entry->last + 1;
    return 0;
}

/* Dist is guaranteed to be of the form "<distributions>". */
static int match_distribution(String name, String dist)
{
    int name_dist_len, pos;

    /* Find length of first component in name. */
    pos = name.find('.');
    name_dist_len = (pos >= 0) ? pos : name.len();

    /* Move past '<' in distributions, check for empty list. */
    dist.shift(1);
    if (*dist == '>')
	return 0;

    /* Look for matches. */
    while (dist.len()) {
	pos = dist.find_set(",>");
	if (compare_case(dist.left(pos), name) == 0)
	    return 1;
	dist.shift(pos + 1);
    }

    /* Didn't find a match. */
    return 0;
}

static FILE *open_group_file(String name, int check)
{
    Stringbuf namebuf(name);
    FILE *fp;

    /* Try common cases (existing file, or file in existing directory). */
    fp = fopen(namebuf.cstring(), (check) ? "r+" : "a+");
    if (fp)
	return fp;
    fp = fopen(namebuf.cstring(), (check) ? "w+" : "w");
    if (fp)
	return fp;

    /* Make all the directories leading up to the file. */
    make_directories_in_path(name);
    return fopen(namebuf.cstring(), (check) ? "w+" : "w");
}

static void delete_group_file(String name, unsigned int dirnamelen)
{
    Stringbuf buf;
    int pos;
    DIR *dir;
    struct dirent d, *dp;
    int found;

    /* Remove the file itself. */
    buf.store(name);
    if (unlink(buf.cstring()) < 0 && errno != ENOENT) {
	error("Can't unlink group file \"%s\": %m", name);
	return;
    }

    /* Remove any empty directories created by removing the file. */
    while (1) {
	/* Lop off the trailing path component. */
	pos = buf.str().rfind('/');
	buf.trunc(pos);

	/* Quit if we're down to dirname. */
	if (buf.len() <= dirnamelen)
	    break;

	dir = opendir(buf.cstring());
	if (!dir)
	    break;
	while (readdir_r(dir, &d, &dp) >= 0 && dp) {
	    if (strcmp(d.d_name, ".") != 0 && strcmp(d.d_name, "..") != 0) {
		found = 1;
		break;
	    }
	}
	closedir(dir);
	if (found)
	    break;
	rmdir(buf.cstring());
    }
}

static void *periodic_check_newsgroups(void *arg)
{
    while (1) {
	sleep(60);
	((Groups *) arg)->check_newsgroups_file();
    }
}

