/* Headers.cc: Data structure for storing headers. */

#include "threads.h"
#include "util.h"

#include "Headers.h"
#include "String.h"
#include "Stringbuf.h"

extern "C" {
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
}

struct Field {
    Field *hash_next;
    Field *next;
    String name;
    String value;
    Stored_string line;
};

Headers::Headers()
{
    int i;

    for (i = 0; i < HEADER_HASHTAB_SIZE; i++)
	hashtab[i] = NULL;
    first = NULL;
    last = NULL;
    current = NULL;
}

Headers::~Headers()
{
    Field *next;

    for (; first; first = next) {
	next = first->next;
	delete first;
    }
}

/* Returns -1 if header is malformatted, 0 if a new header, 1 if replacing
 * an existing header. */
int Headers::add_header(String line)
{
    int hashval, pos, replacing = 1;
    Field *field;
    String name, value;

    /* Find name and value. */
    pos = line.find(':');
    name = line.left(pos);
    name.trunc_whitespace();
    value = line.after(pos + 1);
    value.skip_whitespace();
    if (!name.len())
	return -1;

    field = find_field(name);
    if (field == NULL) {
	field = new Field;
	field->next = NULL;
	if (last)
	    last->next = field;
	else
	    first = field;
	last = field;
	hashval = name.hash_case() % HEADER_HASHTAB_SIZE;
	field->hash_next = hashtab[hashval];
	hashtab[hashval] = field;
	replacing = 0;
    }

    field->line.store(line);
    field->name = name.equivalent_substr(line, field->line.str());
    field->value = value.equivalent_substr(line, field->line.str());
    return replacing;
}

int Headers::add_headers(FILE *fp)
{
    Stringbuf line;
    int c, num_lines = 0;

    while (read_line_clear(&line, fp) == 0) {
	num_lines++;
	if (line.len() == 0)
	    return num_lines;
	while (1) {
	    c = getc(fp);
	    if (c == EOF)
		return -1;
	    ungetc(c, fp);
	    if (c == ' ' || c == '\t') {
		line.append('\n');
		if (read_line(&line, fp) < 0)
		    return -1;
	    } else {
		break;
	    }
	}
	if (line.len() == 1 && *line.str() == '.')
	    return -1;
	if (add_header(line.str()) != 0)
	    return -1;
    }
    return -1;
}

String Headers::add_headers(String lines)
{
    int pos, newline;

    while (1) {
	if (lines.len() && *lines == '\n')
	    return lines.after(1);
	pos = 1;
	while (1) {
	    newline = lines.after(pos).find('\n');
	    if (newline == -1)
		return String(NULL);
	    pos += newline + 1;
	    if (lines.len() < pos || (lines[pos] != ' ' && lines[pos] != '\t'))
		break;
	}
	if (add_header(lines.left(pos - 1)) != 0)
	    return String(NULL);
	lines.shift(pos);
    }
}

String Headers::find_value(String name) const
{
    Field *field;

    field = find_field(name);
    return (field) ? field->value : String(NULL);
}

String Headers::first_header()
{
    current = first;
    return (current) ? current->line.str() : String(NULL);
}

String Headers::next_header()
{
    if (!current)
	return NULL;
    current = current->next;
    return (current) ? current->line.str() : String(NULL);
}

void Headers::clear()
{
    Field *next;
    int i;

    for (; first; first = next) {
	next = first->next;
	delete first;
    }
    for (i = 0; i < HEADER_HASHTAB_SIZE; i++)
	hashtab[i] = NULL;
    last = NULL;
    current = NULL;
}

Field *Headers::find_field(String name) const
{
    int hashval;
    Field *field;

    hashval = name.hash_case() % HEADER_HASHTAB_SIZE;
    for (field = hashtab[hashval]; field; field = field->hash_next) {
	if (compare_case(name, field->name) == 0)
	    return field;
    }
    return NULL;
}

