/* Index.cc: Maintaining the message-ID index. */

#include "threads.h"
#include "error.h"
#include "util.h"

#include "Index.h"
#include "Config.h"
#include "Dbz.h"
#include "Spool.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <unistd.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
}

/* Make configurable. */
const int MODE = S_IRUSR | S_IWUSR | S_IROTH;

Index::Index(const Config *config_arg)
{
    struct stat statbuf;
    Stringbuf buf;

    config = config_arg;

    /* Construct the name of the index. */
    config->get_index_directory(&buf);
    buf.append("articles");
    index_name.store(buf.str());

    make_directories_in_path(buf.str());

    /* Open the database. */
    db = new Dbz();
    db->set_incore(1);
    if (db->init(index_name.cstring()) < 0) {
	if (db->fresh(index_name.cstring(), 0, '\t', 'C', 1))
	    die("Can't open %s for article index database.", index_name.str());
    }

    /* Open the flat file. */
    flatfile_writer = open(index_name.cstring(), O_WRONLY | O_APPEND | O_CREAT,
			   MODE);
    if (flatfile_writer < 0)
	die("Can't open index flat file: %m");
    flatfile_rover = open(index_name.cstring(), O_RDWR, MODE);
    if (flatfile_rover < 0)
	die("Can't open index flat file: %m");

    /* stat the flat file to determine its initial size. */
    if (fstat(flatfile_rover, &statbuf) < 0)
	die("Can't stat flat file: %m");
    position = statbuf.st_size;
}

Index::~Index()
{
    db->close();
    close(flatfile_rover);
    close(flatfile_writer);
}

/* Returns -1 if id maps to a dead entry. */
int Index::lookup(String id, Artloc *loc)
{
    char buf[100];
    String w[3];

    if (get_line(id, buf, sizeof(buf), NULL) < 0)
	return -1;

    if (*buf == 'd')
	return -1;

    /* Read in the values from the buffer. */
    String(buf).get_words(w, 3);
    loc->partition = w[0].tonum();
    loc->bucket_num = w[1].tonum();
    loc->offset = w[2].tonum();
    return 0;
}

/* Returns true if any entry exists for id, even if it's a dead entry. */
int Index::present(String id)
{
    return (find_entry(id) >= 0);
}

int Index::store(String id, const Artloc *loc)
{
    Stringbuf line;
    int result;

    /* The widths here are to allow overwriting during a "drain".  I'm not
     * too pleased about the overhead this causes. */
    line.format("%s\t%3d %10d %10l\n", id, loc->partition, loc->bucket_num,
		loc->offset);
    lock();
    result = store_line(id, line.str());
    unlock();
    return result;
}

int Index::overwrite(String id, const Artloc *loc)
{
    char line[100];
    long offset;

    lock();
    if (get_line(id, line, sizeof(line), &offset) < 0 || strlen(line) != 26) {
	unlock();
	return -1;
    }

    sprintf(line, "%3d %10d %10ld\n", loc->partition, loc->bucket_num,
	    loc->offset);

    /* Write the new line to the flatfile. */
    if (lseek(flatfile_rover, offset, SEEK_SET) < 0) {
	unlock();
	error("Can't seek to %l for Message-ID %s in index flat file: %m",
	      offset, id);
	return -1;
    }
    if (write(flatfile_rover, line, 26) != 26) {
	unlock();
	error("Can't rewrite index entry for Message-ID %s: %m", id);
	return -1;
    }

    unlock();
    return 0;
}

int Index::store_dead_entry(String id)
{
    Stringbuf line;
    int result;

    line.format("%s\td%l\n", id,
		(long)(time(NULL) + config->get_cancel_record_for()));
    lock();
    result = store_line(id, line.str());
    unlock();
    return result;
}

int Index::overwrite_with_dead_entry(String id)
{
    char line[100], buf[20];
    long offset;
    int buflen, linelen;

    /* Get the existing line for id. */
    lock();
    if (get_line(id, line, sizeof(line), &offset) < 0) {
	unlock();
	return -1;
    }

    /* If it's already a dead entry, leave it alone. */
    if (*line == 'd') {
	unlock();
	return 0;
    }

    /* Overwrite the line with a dead entry. */
    linelen = strlen(line);
    sprintf(buf, "d%ld", (long)(time(NULL) + config->get_cancel_record_for()));
    buflen = strlen(buf);
    memcpy(line, buf, buflen);
    memset(line + buflen, ' ', linelen - buflen - 1);

    /* Write the new line to the flatfile. */
    if (lseek(flatfile_rover, offset, SEEK_SET) < 0) {
	unlock();
	error("Can't seek to %l for Message-ID %s in index flat file: %m",
	      offset, id);
	return -1;
    }
    if (write(flatfile_rover, line, linelen) != linelen) {
	unlock();
	error("Can't rewrite index entry for Message-ID %s: %m", id);
	return -1;
    }

    unlock();
    return 0;
}

int Index::sync()
{
    int result;

    lock();
    result = db->sync();
    unlock();
    return result;
}

/* No threads should be performing write operations while this is taking place.
 * Read operations are okay. */
int Index::eliminate_bucket(const Bucket_loc *bloc, time_t bucket_expire_time)
{
    Stringbuf new_name, old_name;
    Stringbuf dirname, new_dirname, old_dirname;
    Stringbuf pagname, new_pagname, old_pagname;
    Stringbuf line;
    String str, w[2];
    Dbz *new_db;
    FILE *in_fp = NULL, *out_fp = NULL;
    time_t now = time(NULL);
    int rover = -1, writer = -1, result, pos, space;
    datum key, value;
    long offset;

    /* Open file handles to the old and new flat files. */
    new_name.format_clear("%s.new", index_name.str());
    in_fp = fopen(index_name.cstring(), "r");
    if (!in_fp) {
	error("Can't reopen index flatfile %s: %m", index_name.str());
	return -1;
    }
    out_fp = fopen(new_name.cstring(), "w");
    if (!out_fp) {
	error("Can't open new index flatfile %s: %m", new_name.str());
	fclose(in_fp);
	return -1;
    }

    /* Copy from in_fp to out_fp, leaving out lines in the expired bucket. */
    while (read_line_clear(&line, in_fp) >= 0) {
	str = line.str();
	pos = str.find('\t');
	if (pos < 0)
	    continue;
	if (str.len() >= pos + 1 && str[pos + 1] == 'd') {
	    /* Dead entry.  Expire it if its expiration time is earlier than
	     * the current time.  If not, chop it off after the expiration
	     * time, since it might have some extra stuff on the end from
	     * overwrite_with_dead_entry(). */
	    if (str.after(pos + 2).tonum() < now)
		continue;
	    space = str.after(pos + 2).find(' ');
	    if (space >= 0)
		line.trunc(pos + 2 + space);
	} else {
	    str.after(pos + 1).get_words(w, 2);
	    if (w[0].tonum() == bloc->partition
		&& w[1].tonum() == bloc->bucket_num) {
		if (expire_record_for == 0)
		    continue;
		line.trunc(pos + 1);
		line.format("d%l",
			    (long)(bucket_expire_time +
				   config->get_expire_record_for()));
	    }
	}
	if (line.str().fwrite(out_fp) < line.len()
	    || putc('\n', out_fp) == EOF) {
	    fclose(in_fp);
	    fclose(out_fp);
	    return -1;
	}
    }
    fclose(in_fp);
    fclose(out_fp);

    /* Open a file handle to the new flatfile and create a new database. */
    in_fp = fopen(new_name.cstring(), "r");
    if (!in_fp) {
	error("Can't reopen new flatfile %s: %m", new_name.cstring());
	return -1;
    }
    new_db = new Dbz();
    new_db->set_incore(1);
    if (new_db->again(new_name.cstring(), index_name.cstring()) < 0) {
	error("Can't open new DBZ database: %m");
	fclose(in_fp);
	delete new_db;
	return -1;
    }

    /* Build the DBZ index. */
    offset = 0;
    while (read_line_clear(&line, in_fp) >= 0) {
	str = line.str();
	pos = str.find('\t');
	if (pos < 0)
	    continue;
	key.dptr = (char *) line.cstring();
	key.dsize = pos;
	value.dptr = (char *) &offset;
	value.dsize = sizeof(offset);
	if (new_db->store(key, value) < 0) {
	    warn("Can't write new DBZ database: %m");
	    fclose(in_fp);
	    new_db->close();
	    delete new_db;
	    return -1;
	}
	offset += line.len() + 1;
    }
    fclose(in_fp);

    /* Sync the new database before switching over to it, out of paranoia. */
    if (new_db->sync() < 0) {
	warn("Can't sync new DBZ database: %m");
	new_db->close();
	delete new_db;
	return -1;
    }

    /* Open rover and writer descriptors to the new flatfile. */
    writer = open(new_name.cstring(), O_WRONLY | O_APPEND | O_CREAT, MODE);
    rover = open(new_name.cstring(), O_RDWR, MODE);
    if (writer < 0 || rover < 0) {
	error("Can't reopen new index flatfile %s: %m", new_name.str());
	new_db->close();
	delete new_db;
	if (writer >= 0)
	    close(writer);
	if (rover >= 0)
	    close(rover);
	return -1;
    }

    /* Rename the database files.  This isn't totally atomic. */
    old_name.format_clear("%s.old", index_name.str());
    dirname.format_clear("%s.dir", index_name.str());
    old_dirname.format_clear("%s.old.dir", index_name.str());
    new_dirname.format_clear("%s.new.dir", index_name.str());
    pagname.format_clear("%s.pag", index_name.str());
    old_pagname.format_clear("%s.old.pag", index_name.str());
    new_pagname.format_clear("%s.new.pag", index_name.str());
    unlink(old_name.cstring());
    unlink(old_dirname.cstring());
    unlink(old_pagname.cstring());
    result = link(index_name.cstring(), old_name.cstring());
    result |= link(dirname.cstring(), old_dirname.cstring());
    result |= link(pagname.cstring(), old_pagname.cstring());
    result |= rename(new_name.cstring(), index_name.cstring());
    result |= rename(new_dirname.cstring(), dirname.cstring());
    result |= rename(new_pagname.cstring(), pagname.cstring());
    unlink(old_name.cstring());
    unlink(old_dirname.cstring());
    unlink(old_pagname.cstring());
    if (result != 0)
	die("Couldn't rename article index files: %m");

    /* Switch over to the new database. */
    lock();
    close(flatfile_rover);
    close(flatfile_writer);
    db->close();
    delete db;
    flatfile_rover = rover;
    flatfile_writer = writer;
    position = offset;
    db = new_db;
    unlock();

    return 0;
}

/* Assumes lock. */
int Index::get_line(String id, char *buf, int bufsize, long *offset)
{
    long off;
    char *p;

    if (!offset)
	offset = &off;

    /* Look up the offset in the index. */
    *offset = find_entry(id);
    if (*offset < 0)
	return -1;

    /* Skip past the message ID and read the data. */
    *offset += id.len() + 1;
    if (lseek(flatfile_rover, *offset, SEEK_SET) < 0) {
	error("Can't seek to %d for Message-ID %s in index flat file: %m",
	      *offset, id);
	return -1;
    }
    if (read(flatfile_rover, buf, bufsize - 1) <= 0) {
	error("Can't read at %d for Message-ID %s in index flat file: %m",
	      *offset, id);
	return -1;
    }
    unlock();

    buf[bufsize - 1] = 0;
    p = strchr(buf, '\n');
    if (p)
	p[1] = 0;
    return 0;
}

/* Assumes lock. */
int Index::store_line(String id, String line)
{
    Stringbuf buf(id);
    datum key, value;
    int result;

    /* Write the bucket and offset into the flat file. */
    result = line.write(flatfile_writer);
    if (result != line.len()) {
	warn("Can't write index entry for Message-ID %s: %m", id);
	return -1;
    }

    /* Store the association in the dbz database. */
    key.dptr = (char *) buf.cstring();
    key.dsize = buf.len();
    if (key.dsize > DBZMAXKEY)
	key.dsize = DBZMAXKEY;
    value.dptr = (char *) &position;
    value.dsize = sizeof(position);
    if (db->store_case(key, value) < 0) {
	warn("Can't store pair for Message-ID %s.", id);
	return -1;
    }

    /* Update the position and return. */
    position += line.len();
    return 0;
}

/* Assumes lock. */
long Index::find_entry(String id)
{
    Stringbuf buf(id);
    datum key;
    datum value;
    long offset;

    /* Look up the message ID in the database. */
    key.dptr = (char *) buf.cstring();
    key.dsize = buf.len();
    if (key.dsize > DBZMAXKEY)
	key.dsize = DBZMAXKEY;
    value = db->fetch_case(key);
    if (value.dptr == NULL || value.dsize != sizeof(offset))
	return -1;

    /* Copy the value to an aligned spot and return it. */
    memcpy(&offset, value.dptr, sizeof(offset));
    return offset;
}

