/* Newsfeed.cc: Handler for feeding articles to another site. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Newsfeed.h"
#include "Config.h"
#include "Headers.h"
#include "Connection.h"
#include "Article.h"
#include "Spool.h"
#include "Stringbuf.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
}

/* A word on locking: normally, control within the TNS spool only
 * flows outward, from the spool interface to the spool components to
 * the spool items.  We are special in that we use the spool to
 * retrieve the articles we're about to feed.  To avoid possible
 * deadlocks, we must not make calls to the spool when we are holding
 * internal locks, and we must only make calls to the spool from
 * within the feeder thread, not from within calls from outside.  This
 * constraint is easy to obey in the current design; we just have to
 * unlock the queue lock before doing feed_article() calls, and we
 * want to do that anyway. */

static void *run_feed(void *arg);

Newsfeed::Newsfeed(String name_arg, Spool *spool_arg, const Config *config)
{
    struct stat statbuf;
    Stringbuf buf;

    name.store(name_arg);
    spool = spool_arg;
    groups_list = config->get_feed_groups(name_arg);
    dists_list = config->get_feed_dists(name_arg);
    paths_list = config->get_feed_paths(name_arg);
    config->get_feeds_directory(&buf);
    dirname.store(buf.str());
    buf.format_clear("%s/%s.queue", dirname.str(), name_arg);
    queue_filename.store(buf.str());
    buf.format_clear("%s/%s.work", dirname.str(), name_arg);
    work_filename.store(buf.str());
    buffer_size = config->get_feed_buffer_size(name_arg);
    buffer_pos = 0;
    queue_buffer = new Artloc[buffer_size];
    queue_file = fopen(queue_filename.cstring(), "a+");
    if (!queue_file)
	die("Can't open queue file %s for output.", queue_filename.str());

    if (stat(queue_filename.cstring(), &statbuf) < 0)
	die("Can't stat queue file %s.", queue_filename.str());
    file_flag = (statbuf.st_size > 0);
    work_flag = file_flag;

    start_thread(run_feed, this);
}

Newsfeed::~Newsfeed()
{
    delete groups_list;
    delete dists_list;
    delete paths_list;
    delete queue_buffer;
}

void Newsfeed::feed_articles()
{
    FILE *fp;
    Connection *conn;
    struct timeval tv;
    struct timezone tz;
    struct timespec ts;

    /* Check for a work file left by a crash. */
    fp = fopen(work_filename.cstring(), "r");
    if (fp) {
	queue_articles_from_file(fp);
	fclose(fp);
	sync_queue();
	unlink(work_filename.cstring());
    }

    /* Loop, waiting for work and processing the queue file. */
    while (1) {

	/* Wait until the queue file is non-empty. */
	queue_lock.lock();
	while (!work_flag)
	    work.wait(&queue_lock);
	queue_lock.unlock();

	/* Connect to the remote server. */
	conn = make_connection();
	while (!conn) {
	    sleep(300);
	    conn = make_connection();
	}

	while (1) {
	    if (process_queue(conn) < 0)
		break;

	    /* Hold connection open for five minutes waiting for more work. */
	    gettimeofday(&tv, &tz);
	    TIMEVAL_TO_TIMESPEC(&tv, &ts);
	    ts.tv_sec += 300;
	    queue_lock.lock();
	    while (!work_flag) {
		if (work.timedwait(&queue_lock, &ts) == ETIMEDOUT)
		    break;
	    }
	    queue_lock.unlock();
	    if (!work_flag)
		break;
	}

	delete conn;
    }
}

int Newsfeed::sync_queue()
{
    int ok = 1;

    queue_lock.lock();
    if (flush_buffer_to_queue_file() < 0)
	ok = 0;
    if (ok && fflush(queue_file) == EOF)
	ok = 0;
    if (ok && fsync(fileno(queue_file)) < 0)
	ok = 0;
    queue_lock.unlock();
    return (ok) ? 0 : -1;
}

/* XXX error-handling needs more attention. */
int Newsfeed::process_queue(Connection *conn)
{
    FILE *fp = NULL;
    Artloc *buf, loc;
    Stringbuf line;
    String w[3];
    int i, buffered;

    /* Snapshot the queue buffer and queue file. */
    queue_lock.lock();
    buffered = buffer_pos;
    buf = new Artloc[buffered];
    memcpy(buf, queue_buffer, buffered * sizeof(queue_buffer));
    buffer_pos = 0;

    if (file_flag) {
	fclose(queue_file);
	rename(queue_filename.cstring(), work_filename.cstring());
	queue_file = fopen(queue_filename.cstring(), "a+");
	file_flag = 0;
	fp = fopen(work_filename.cstring(), "r");
	if (!fp)
	    die("Can't open the work file we just created.");
    }

    work_flag = 0;
    queue_lock.unlock();

    /* Process our copy of the buffer. */
    for (i = 0; i < buffered; i++) {
	if (feed_article(conn, &buf[i]) < 0) {
	    queue_article(&buf[i]);
	    for (; i < buffered; i++)
		queue_article(&buf[i]);
	    if (fp)
		queue_articles_from_file(fp);
	    break;
	}
    }

    /* Process the work file if we have one. */
    if (fp) {
	while (read_line_clear(&line, fp) >= 0) {
	    line.str().get_words(w, 3);
	    loc.partition = w[0].tonum();
	    loc.bucket_num = w[1].tonum();
	    loc.offset = w[2].tonum();
	    if (feed_article(conn, &loc) < 0) {
		queue_article(&loc);
		queue_articles_from_file(fp);
		break;
	    }
	}
	fclose(fp);
    }

    delete[] buf;
    return 0;
}

void Newsfeed::queue_articles_from_file(FILE *fp)
{
    Stringbuf line;
    String w[3];
    Artloc loc;

    while (read_line_clear(&line, fp)) {
	line.str().get_words(w, 3);
	loc.partition = w[0].tonum();
	loc.bucket_num = w[1].tonum();
	loc.offset = w[2].tonum();
	queue_article(&loc);
    }
}

int Newsfeed::propagate_article(const Artloc *loc, const Headers *headers,
				String)
{
    if (!wanted(headers))
	return 0;

    return queue_article(loc);
}

int Newsfeed::feed_article(Connection *conn, const Artloc *loc)
{
    Article *art;
    Stringbuf buf;

    /* Get a handle to the article from the spool. */
    art = spool->get_article(loc);
    if (!art)
	return 0;

    /* Ask to send the article with an IHAVE. */
    art->get_message_id(&buf);
    conn->writef("IHAVE %S\r\n", &buf);
    if (conn->read_line_clear(&buf) < 0) {
	spool->release_article(art);
	return -1;
    }

    /* See if the remote server wants the article. */
    if (buf.len() > 0 && *buf.str() == '4') {
	spool->release_article(art);
	return 0;
    }
    if (buf.len() == 0 || *buf.str() != '3') {
	spool->release_article(art);
	return -1;
    }

    /* Send the article. */
    art->write_head(conn);
    conn->write("\r\n");
    art->write_body(conn);
    conn->write(".\r\n");

    /* See if we need to requeue the article. */
    if (conn->read_line_clear(&buf) < 0) {
	spool->release_article(art);
	return -1;
    }
    if (compare(buf.str().left(3), "436") == 0)
	queue_article(loc);

    spool->release_article(art);
    return 0;
}

Connection *Newsfeed::make_connection()
{
    Stringbuf buf;
    int fd;
    Connection *conn;

    /* Make a TCP connection to the remote host. */
    fd = tcp_connect(name.str(), "nntp", 119, &buf);
    if (fd < 0) {
	warn("%s", buf.str());
	return NULL;
    }

    conn = new Connection(fd);

    /* Read the first line from the remote host. */
    if (conn->read_line_clear(&buf) < 0) {
	warn("Couldn't read initial line from %s.", name.str());
	delete conn;
	return NULL;
    }

    /* See if we have read access to the host. */
    if (compare(buf.str().left(3), "200") != 0) {
	warn("No transfer permission to %s: %s", name.str(), buf.str());
	if (buf.len() == 1 && *buf.str() == '2') {
	    /* Shut down the connection gracefully. */
	    conn->write("QUIT\r\n");
	    conn->read_line_clear(&buf);
	}
	delete conn;
	return NULL;
    }

    return conn;
}

int Newsfeed::wanted(const Headers *headers)
{
    String path, groups, dists, name;

    /* Don't propagate if there is an excluded path component. */
    path = (*headers)["Path"];
    while (path.len()) {
	name = get_delimited_element(&path, '!');
	if (paths_list->match1(name, 1) == 0)
	    return 0;
    }

    /* Don't propagate if none of the newsgroups are wanted. */
    groups = (*headers)["Newsgroups"];
    while ((name = get_group_name(&groups)).len()) {
	if (groups_list->match1(name, 0) == 0)
	    break;
    }
    if (!name.len())
	return 0;

    /* Don't propagate if distributions are specified and none are wanted. */
    dists = (*headers)["Distributions"];
    if (dists.valid()) {
	while ((name = get_delimited_element(&dists, ',')).len()) {
	    if (dists_list->match1(name, 1) == 0)
		break;
	}
	if (!name.len())
	    return 0;
    }

    /* Propagate the article. */
    return 1;
}

int Newsfeed::queue_article(const Artloc *loc)
{
    queue_lock.lock();

    /* If the in-memory buffer is full, flush the buffer to the queue file. */
    if (buffer_pos == buffer_size) {
	if (flush_buffer_to_queue_file() < 0) {
	    queue_lock.unlock();
	    return -1;
	}
    }

    /* Insert the article location into the queue buffer. */
    queue_buffer[buffer_pos] = *loc;
    buffer_pos++;

    work_flag = 1;
    work.signal();
    queue_lock.unlock();
    return 0;
}

/* Assumes lock. */
int Newsfeed::flush_buffer_to_queue_file()
{
    int i;
    Stringbuf buf;

    for (i = 0; i < buffer_pos; i++) {
	buf.format_clear("%d\t%d\t%l\n", queue_buffer[i].partition,
			 queue_buffer[i].bucket_num,
			 queue_buffer[i].offset);
	if (buf.str().fwrite(queue_file) != buf.len()) {
	    queue_lock.unlock();
	    return -1;
	}
    }
    buffer_pos = 0;
    file_flag = 1;
    return 0;
}

static void *run_feed(void *arg)
{
    ((Newsfeed *) arg)->feed_articles();
    return NULL;
}

