/* Newsgroup.cc: A newsgroup file. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Newsgroup.h"
#include "Connection.h"
#include "Config.h"
#include "Spool.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
}

const int MIN_POWER = 5;
const int DELTA = 2;
const int AVERAGE_LINE_LENGTH = 225; /* A heuristic. */

static String find_overview_field(String overdata, int fnum);
static void send_header(Connection *conn, Article_entry *entry, int fnum,
			Spool *spool, String name);

/* fp may be NULL, in which case the newsgroup is assumed to be empty. */
Newsgroup::Newsgroup(String name, FILE *fp, const Config *config_arg,
		     int last_arg)
{
    struct stat statbuf;
    int approx_num_entries = 0;
    Stringbuf buf;
    String overdata, w[4];
    Artloc loc;

    group_name.store(name);
    config = config_arg;
    last = last_arg;

    /* Estimate the number of entries in the group file. */
    if (fp && fstat(fileno(fp), &statbuf) >= 0)
	approx_num_entries = statbuf.st_size / AVERAGE_LINE_LENGTH;
    else if (fp)
	warn("Can't stat newsgroup file %s: %m", name);

    /* Initialize the article entries table. */
    entries_size = adjust_size(approx_num_entries, MIN_POWER, DELTA);
    entries = new Article_entry[entries_size];
    num_entries = 0;

    /* Read in the group file and close it, if we have one. */
    if (fp) {
	while (read_line_clear(&buf, fp) >= NULL) {
	    /* Check if it's a cancelled line. */
	    if (buf.len() && *buf.str() == 'C')
		continue;

	    /* Extract the number and location of the article and add an entry
	     * to the list. */
	    overdata = buf.str().get_words(w, 4);
	    loc.partition = w[1].tonum();
	    loc.bucket_num = w[2].tonum();
	    loc.offset = w[3].tonum();
	    add_article_entry(w[0].tonum(), &loc, overdata);
	}
	fclose(fp);
    }
}

Newsgroup::~Newsgroup()
{
    delete[] entries;
}

int Newsgroup::num_articles()
{
    return num_entries;
}

int Newsgroup::first_number()
{
    int ind;

    lock();
    ind = (num_entries > 0) ? entries[0].number : last + 1;
    unlock();
    return ind;
}

int Newsgroup::last_number()
{
    return last;
}

int Newsgroup::lookup(int number, Artloc *loc)
{
    int i;

    /* Look for the number in the table. */
    lock();
    i = find_entry(number);
    if (i < 0) {
	unlock();
	return -1;
    }

    *loc = entries[i].loc;
    unlock();
    return 0;
}

int Newsgroup::next(int *number, Artloc *loc)
{
    int i;

    /* Look for the number in the table. */ 
    lock();
    i = find_entry(*number);
    if (i < 0 || i == num_entries - 1) {
	unlock();
	return -1;
    }

    /* Copy the next entry into the arguments, and cache the entry
     * number. */
    i++;
    *number = entries[i].number;
    *loc = entries[i].loc;
    unlock();
    return 0;
}

int Newsgroup::prev(int *number, Artloc *loc)
{
    int i;

    /* Look for the number in the table. */
    lock();
    i = find_entry(*number);
    if (i <= 0) {
	unlock();
	return -1;
    }

    /* Copy the previous entry into the arguments, and cache the entry
     * number. */
    i--;
    *number = entries[i].number;
    *loc = entries[i].loc;
    unlock();
    return 0;
}

int Newsgroup::first(int *number, Artloc *loc)
{
    /* Return the first entry, unless the group is empty. */
    lock();
    if (!num_entries) {
	unlock();
	return -1;
    }
    *number = entries[0].number;
    *loc = entries[0].loc;
    unlock();
    return 0;
}

void Newsgroup::add_entry(int number, const Artloc *loc, String overdata)
{
    lock();
    add_article_entry(number, loc, overdata);
    unlock();
}

void Newsgroup::eliminate_bucket(const Bucket_loc *bloc)
{
    int i, j;

    /* Eliminate the entries in the newsgroup whose bucket is equal to
     * bucket_num. */
    lock();
    j = 0;
    for (i = 0; i < num_entries; i++) {
	if (entries[i].loc.partition != bloc->partition
	    || entries[i].loc.bucket_num != bloc->bucket_num)
	    entries[j++] = entries[i];
    }
    num_entries = j;
    unlock();
}

void Newsgroup::eliminate_article(int number)
{
    int i;

    /* Eliminate the entry in the newsgroup whose number is number. */
    lock();
    for (i = 0; i < num_entries; i++) {
	if (entries[i].number == number)
	    break;
    }
    for (; i < num_entries - 1; i++)
	entries[i] = entries[i + 1];
    num_entries--;
    unlock();
}

void Newsgroup::send_overview(Connection *conn, int low, int high)
{
    int i, ind;

    lock();
    if (low == high) {
	/* If it's a single article, use a binary search to do it quickly. */
	i = find_entry(low);
	if (i >= 0)
	    conn->writef("%d\t%s\r\n", low, entries[i].overdata.str());
    } else {
	/* If it's anything else, just iterate through the whole list. */
	for (i = 0; i < num_entries; i++) {
	    ind = entries[i].number;
	    if (ind >= low && (high == -1 || ind <= high))
		conn->writef("%d\t%s\r\n", ind, entries[i].overdata.str());
	}
    }
    unlock();
}

void Newsgroup::send_headers(Connection *conn, String name, int low, int high,
			     Spool *spool)
{
    int fnum, ind, i;

    lock();
    fnum = config->find_in_schema(name);
    if (low == high) {
	/* Find a single article quickly with a binary search. */
	i = find_entry(low);
	if (i >= 0)
	    send_header(conn, &entries[i], fnum, spool, name);
    } else {
	/* Otherwise, just iterate through all of the entries. */
	for (i = 0; i < num_entries; i++) {
	    ind = entries[i].number;
	    if (ind >= low && (high == -1 || ind <= high))
		send_header(conn, &entries[i], fnum, spool, name);
	}
    }
    unlock();
}

/* Assumes lock. */
void Newsgroup::add_article_entry(int number, const Artloc *loc,
				  String overdata)
{
    Article_entry *new_entries, *entry;
    int new_size;

    /* Expand the table if we have to. */
    if (num_entries >= entries_size) {
	new_size = adjust_size(num_entries + 1, MIN_POWER, DELTA);
	new_entries = new Article_entry[new_size];
	memcpy(new_entries, entries, num_entries * sizeof(Article_entry));
	delete entries;
	entries = new_entries;
	entries_size = new_size;
    }

    /* Entries will usually be added sequentially, but we need to make sure. */
    entry = entries + num_entries;
    while (entry > entries && (entry - 1)->number > number) {
	*entry = *(entry - 1);
	entry--;
    }

    entry->number = number;
    entry->loc = *loc;
    entry->overdata.store(overdata);
    num_entries++;
    if (number > last)
	last = number;
}

/* Assumes read lock. */
int Newsgroup::find_entry(int number)
{
    int low, high, mid;

    /* Perform a simple binary search.  Note that there's no equality
     * test in the loop; this is somewhat faster on average.  Be
     * careful of infinite loops if you modify this search, though;
     * make sure to consider the case where high == low + 1. */
    low = 0;
    high = num_entries - 1;
    while (high > low) {
	mid = (low + high) / 2;
	if (number > entries[mid].number)
	    low = mid + 1;
	else
	    high = mid;
    }

    return (entries[low].number == number) ? low : -1;
}

static String find_overview_field(String overdata, int fnum)
{
    int pos;

    for (; fnum; fnum--) {
	pos = overdata.find('\t');
	if (pos < 0)
	    return "";
	overdata.shift(pos + 1);
    }
    pos = overdata.find('\t');
    return (pos >= 0) ? overdata.left(pos) : overdata;
}

static void send_header(Connection *conn, Article_entry *entry, int fnum,
			Spool *spool, String name)
{
    Stringbuf buf;
    String field;

    if (fnum >= 0) {
	field = find_overview_field(entry->overdata.str(), fnum);
	conn->writef("%d %s\r\n", entry->number, field);
    } else if (spool->find_header(&entry->loc, name, &buf) >= 0) {
	conn->writef("%d %s\r\n", entry->number, buf.str());
    }
}

