/* Nntp.c: Implementation of the Nntp protocol-handler class. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Nntp.h"
#include "Connection.h"
#include "String.h"
#include "Spool.h"
#include "Config.h"
#include "Wildcard_list.h"
#include "Newsgroup.h"

extern "C" {
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
}

typedef void (Nntp::*Handler)(const char *s);

static int get_range(String s, int *low, int *high);

Nntp::Nntp(Spool *spool_arg, Connection *conn_arg, Config *config_arg)
{
    spool = spool_arg;
    conn = conn_arg;
    config = config_arg;
    cur_group = NULL;
    cur_article = NULL;
    cur_number = -1;
    running = 1;
    post_access = config->get_post_access(conn->name(), conn->ip());
    read_access = config->get_read_access(conn->name(), conn->ip());
    transfer_access = config->get_transfer_access(conn->name(), conn->ip());
}

Nntp::~Nntp()
{
    if (cur_group)
	spool->release_group(cur_group);
    if (cur_article)
	spool->release_article(cur_article);
    delete conn;
    delete post_access;
    delete read_access;
    delete transfer_access;
}

void Nntp::run()
{
    Stringbuf line;
    String str, cmd, rest;
    bool whether;

    /* If you're not a cop, you're little people. */
    if (read_access->is_degenerate(0, &whether) && !whether &&
	transfer_access->is_degenerate(0, &whether) && !whether) {
	conn->write("502 You have no permission to talk.  Goodbye.\r\n");
	return;
    }

    if (post_access->is_degenerate(0, &whether) && !whether)
	conn->write("201 Threaded news server ready (no post access)\r\n");
    else
	conn->write("200 Threaded news server ready\r\n");
    while (running && conn->read_line_clear(&line) >= 0) {
	log("NNTP %s: %s", conn->name(), line.str());

	/* Ignore blank lines. */
	str = line.str();
	str.strip_whitespace();
	if (str.len() == 0)
	    continue;

	/* Run the command associated with the starting word. */
	rest = str.get_words(&cmd, 1);
	if (handle_command(cmd, rest) < 0)
	    conn->write("500 Command not recognized\r\n");
    }
}

int Nntp::handle_command(String cmd, String rest)
{
    static struct entry {
	String name;
	void (Nntp::*func)(String s);
    } tab[] = {
	{ String("article"),	Nntp::article },
	{ String("body"),	Nntp::body },
	{ String("group"),	Nntp::group },
	{ String("head"),	Nntp::head },
	{ String("help"),	Nntp::help },
	{ String("ihave"),	Nntp::ihave },
	{ String("last"),	Nntp::last },
	{ String("list"),	Nntp::list },
	{ String("newgroups"),	Nntp::newgroups },
	{ String("newnews"),	Nntp::newnews },
	{ String("next"),	Nntp::next },
	{ String("post"),	Nntp::post },
	{ String("quit"),	Nntp::quit },
	{ String("slave"),	Nntp::slave },
	{ String("stat"),	Nntp::stat },
	{ String("xhdr"),	Nntp::xhdr },
	{ String("xover"),	Nntp::xover }
    };
    unsigned int i;

    for (i = 0; i < sizeof(tab) / sizeof(struct entry); i++) {
	if (compare_case(tab[i].name, cmd) == 0) {
	    (this->*tab[i].func)(rest);
	    return 0;
	}
    }

    return -1;
}

void Nntp::article(String s)
{
    fetch_article(s, 1, 1);
}

void Nntp::body(String s)
{
    fetch_article(s, 0, 1);
}

void Nntp::group(String s)
{
    Newsgroup *group;
    int num, first, last;

    /* Check reader access to the group. */
    if (!read_access->match1(s, 0)) {
	conn->writef("502 No access to group \"%s\"\r\n", s);
	return;
    }

    /* Look up the group in the spool. */
    group = spool->find_group(s);
    if (!group) {
	conn->write("411 No such newsgroup\r\n");
	return;
    }

    /* Release the current article, if we have one. */
    if (cur_article) {
	spool->release_article(cur_article);
	cur_article = NULL;
	cur_number = -1;
    }

    /* Get information about the group. */
    num = spool->group_num_articles(group);
    first = spool->group_first_number(group);
    last = spool->group_last_number(group);

    /* Set the current group to this group. */
    if (cur_group)
	spool->release_group(cur_group);
    cur_group = group;

    /* Respond with the number of articles in the group and the number
     * of the first and last article. */
    conn->writef("211 %d %d %d %s group selected\r\n", num, first, last, s);
}

void Nntp::head(String s)
{
    fetch_article(s, 1, 0);
}

void Nntp::help(String)
{
    conn->write("100 Help text follows.\r\n");
    conn->write("  ARTICLE [message-id|number]\r\n");
    conn->write("  BODY [message-id|number]\r\n");
    conn->write("  GROUP newsgroup\r\n");
    conn->write("  HELP\r\n");
    conn->write("  IHAVE messsage-id\r\n");
    conn->write("  LAST\r\n");
    conn->write("  LIST {ACTIVE|NEWSGROUPS|DISTRIBUTIONS|SCHEMA}\r\n");
    conn->write("  NEWGROUPS date time [GMT] [distributions]\r\n");
    conn->write("  NEWNEWS newsgroups date time [GMT] [distributions]\r\n");
    conn->write("  NEXT\r\n");
    conn->write("  POST\r\n");
    conn->write("  QUIT\r\n");
    conn->write("  SLAVE\r\n");
    conn->write("  STAT [message-id|number]\r\n");
    conn->write("  XHDR field [message-id|range]\r\n");
    conn->write("  XOVER range\r\n");
    conn->write(".\r\n");
}

void Nntp::ihave(String s)
{
    /* NNTP status numbers for POST_OK, POST_REJECTED, and POST_FAILED. */
    static int statuses[] = { 235, 437, 437 };
    Stringbuf line, article;
    Post_status result;

    if (transfer_access->num_patterns() == 0) {
	conn->write("502 You don't have access to transfer articles.\r\n");
	return;
    }

    /* Check the message ID for syntactic correctness. */
    if (verify_message_id(s) < 0) {
	conn->write("501 message ID not correct\r\n");
	return;
    }

    if (!spool->article_wanted(s)) {
	conn->write("435 article not wanted - do not send it\r\n");
	return;
    }

    conn->write("335 send article to be transferred."
		"  End with <CR-LF>.<CR-LF>\r\n");

    /* Read the article from the client. */
    while (conn->read_line_clear(&line) >= 0) {
	/* If the line is simply a period, the article is done. */
	if (line.len() == 1 && *line.str() == '.')
	    break;

	article.append(line.str());
        article.append('\n');
    }

    spool_message.clear();
    result = spool->ihave_article(s, article.str(), transfer_access,
				  &spool_message);

    conn->writef("%d %s\r\n", statuses[result], spool_message.str());
}

void Nntp::last(String)
{
    Article *article;

    /* Make sure there's a current group and article. */
    if (!cur_group) {
	conn->write("412 No newsgroup selected\r\n");
	return;
    }
    if (!cur_article && select_first_article() < 0) {
	conn->write("420 No current article has been selected.\r\n");
	return;
    }

    /* Get the next article in the current group. */
    article = spool->prev_article(cur_group, &cur_number);
    if (!article) {
	conn->write("422 No previous article in this group.\r\n");
	return;
    }

    /* Set the current article. */
    spool->release_article(cur_article);
    cur_article = article;

    /* Return success to the client. */
    message_id.clear();
    spool->get_article_message_id(article, &message_id);
    conn->writef("223 %d %s article retrieved - request text separately.\r\n",
		 cur_number, message_id.str());
}

void Nntp::list(String s)
{
    Wildcard_decision_list *access = NULL;
    bool whether;

    /* Sites with transfer access can find out about all groups. */
    if (transfer_access->is_degenerate(0, &whether) && !whether)
	access = read_access;

    if (!s.len() || compare_case(s, "active") == 0) {
	conn->write("215 Newsgroups in form \"group high low flags\"\r\n");
	spool->send_grouplist(conn, access);
	conn->write(".\r\n");
    } else if (compare_case(s, "newsgroups") == 0) {
	conn->write("215 Descriptions in form \"group description\"\r\n");
	spool->send_group_descriptions(conn, access);
	conn->write(".\r\n");
    } else if (compare_case(s, "distributions") == 0) {
	conn->write("500 Command not implemented\r\n");
    } else if (compare_case(s, "distrib.pats") == 0) {
	conn->write("500 What's a distrib.pats file?\r\n");
    } else if (compare_case(s, "schema") == 0 ||
	       compare_case(s, "overview.fmt") == 0) {
	conn->write("215 Order of fields in overview database\r\n");
	spool->send_schema(conn);
	conn->write(".\r\n");
    } else {
	conn->write("501 Bad command use\r\n");
    }
}

void Nntp::newgroups(String s)
{
    String rest, w[4], date, time, gmt, dist;
    time_t t;
    Wildcard_decision_list *access = NULL;
    bool whether;

    /* Sites with transfer access can find out about all groups. */
    if (transfer_access->is_degenerate(0, &whether) && !whether)
	access = read_access;

    /* Date, time, optional GMT, and optional distribution. */
    rest = s.get_words(w, 4);
    date = w[0];
    time = w[1];
    gmt = (!w[3].len() && w[2].len() && *w[2] == '<') ? String("") : w[2];
    dist = (!w[3].len() && w[2].len() && *w[2] == '<') ? w[2] : w[3];

    /* First two words must be six characters each, gmt must be "GMT" if
     * specified, distributions must begin with a < and end with a >. */
    if (rest.len() || date.len() != 6 || time.len() != 6
	|| (gmt.len() && compare_case(gmt, "GMT"))
	|| (dist.len() && (*dist != '<' || dist[dist.len() - 1] != '>'))) {
	conn->write("501 Bad command use\r\n");
	return;
    }

    /* Parse date and time into Unix time. */
    t = parse_nntp_date(date, time, gmt);
    if (t == -1) {
	conn->write("501 Bad command use\r\n");
	return;
    }

    conn->write("231 Newsgroups in form \"group high low flags\"\r\n");
    spool->send_new_groups(conn, t, dist, access);
    conn->write(".\r\n");
}

void Nntp::newnews(String)
{
    conn->write("400 Command not implemented\r\n");
}

void Nntp::next(String)
{
    Article *article;

    /* Make sure there's a current group and article. */
    if (!cur_group) {
	conn->write("412 No newsgroup selected.\r\n");
	return;
    }
    if (!cur_article && select_first_article() < 0) {
	conn->write("420 No current article has been selected.\r\n");
	return;
    }

    /* Get the next article in the current group. */
    article = spool->next_article(cur_group, &cur_number);
    if (!article) {
	conn->write("421 No next article in this group.\r\n");
	return;
    }

    /* Set the current article. */
    spool->release_article(cur_article);
    cur_article = article;

    /* Return success to the client. */
    message_id.clear();
    spool->get_article_message_id(article, &message_id);
    conn->writef("223 %d %s article retrieved - request text separately.\r\n",
		 cur_number, message_id.str());
}

void Nntp::post(String)
{
    /* NNTP status numbers for POST_OK, POST_REJECTED, and POST_FAILED. */
    Stringbuf line, article;
    bool whether;

    if (post_access->is_degenerate(0, &whether) && !whether) {
	conn->write("440 Posting not allowed.\r\n");
	return;
    }

    conn->write("340 Send article to be posted."
		"  End with <CR-LF>.<CR-LF>\r\n");

    /* Read the article from the client. */
    while (conn->read_line(&line) >= 0) {
	/* If the line is simply a period, the article is done. */
	if (line.len() == 1 && *line.str() == '.')
	    break;

	article.append(line.str());
        article.append('\n');
	line.clear();
    }

    spool_message.clear();
    if (spool->post_article(article.str(), conn->name(), post_access,
			    &spool_message) < 0)
	conn->writef("441 %s\r\n", spool_message.str());
    else
	conn->write("240 Article posted okay.\r\n");
}

void Nntp::quit(String)
{
    conn->write("205 Closing connection.\r\n");
    running = 0;
}

void Nntp::slave(String)
{
    conn->write("202 Slave status noted.\r\n");
}

void Nntp::stat(String s)
{
    /* Special case stat of a message-ID. */
    if (s.len() && *s == '<') {
	if (spool->article_exists(s))
	    conn->writef("223 0 %s article exists in spool.\r\n", s);
	else
	    conn->write("430 No such article.\r\n");
    } else {
	fetch_article(s, 0, 0);
    }
}

void Nntp::xhdr(String s)
{
    int low, high;
    String header, range;

    range = s.get_words(&header, 1);
    if (range.len() && *range == '<') {
	spool_message.clear();
	if (spool->find_header(range, header, &spool_message) < 0) {
	    conn->write("430 No such article\r\n");
	} else {
	    conn->writef("221 0 %s header of article %s\r\n", header, range);
	    conn->writef("%s %s\r\n.\r\n", range, &spool_message);
	}
	return;
    }
    if (get_range(range, &low, &high) < 0) {
	conn->write("501 Incorrect range syntax\r\n");
    } else if (!cur_group) {
	conn->write("412 No newsgroup has been selected.\r\n");
    } else {
	if (low == -1 && high == -1) {
	    if (!cur_article && cur_group->num_articles() > 0)
		cur_article = spool->first_article(cur_group, &cur_number);
	    if (!cur_article) {
		conn->write("420 No current article has been selected.\r\n");
		return;
	    }
	    low = high = cur_number;
	}
	conn->writef("221 %s fields follow.\r\n", header);
	cur_group->send_headers(conn, header, low, high, spool);
	conn->write(".\r\n");
    }
}

void Nntp::xover(String s)
{
    int low, high;

    if (get_range(s, &low, &high) < 0) {
	conn->write("501 Incorrect range syntax\r\n");
    } else if (!cur_group) {
	conn->write("412 No newsgroup has been selected.\r\n");
    } else {
	if (low == -1 && high == -1) {
	    if (!cur_article && cur_group->num_articles() > 0)
		cur_article = spool->first_article(cur_group, &cur_number);
	    if (!cur_article) {
		conn->write("420 No current article has been selected.\r\n");
		return;
	    }
	    low = high = cur_number;
	}
	conn->write("224 data follows.\r\n");
	cur_group->send_overview(conn, low, high);
	conn->write(".\r\n");
    }
}

void Nntp::fetch_article(String s, int head, int body)
{
    static String responses[2][2] = {
	{ "223 %d %s article retrieved - request text separately.\r\n",
	  "222 %d %s article retrieved - body follows.\r\n" },
	{ "221 %d %s article retrieved - head follows.\r\n",
	  "220 %d %s article retrieved - head and body follow.\r\n" }
    };
    Stringbuf buf;
    Article *article;
    int number = 0;
    String message_id;
    bool whether;

    /* Sites with only transfer access can't do more than stat a message-ID. */
    if (read_access->is_degenerate(0, &whether) && !whether) {
	conn->write("502 Reading articles not allowed.\r\n");
	return;
    }

    /* Preemptively check a couple of error conditions. */
    if ((!s.len() || *s != '<') && !cur_group) {
	conn->write("412 No newsgroup has been selected.\r\n");
	return;
    }
    if (!s.len() && !cur_article && select_first_article() < 0) {
	conn->write("420 No current article has been selected.\r\n");
	return;
    }

    if (!s.len()) {
	article = cur_article;
	number = cur_number;
    } else if (*s == '<') {
	/* It's a message ID. */
	article = spool->find_article(s);
	if (!article) {
	    conn->write("430 No such article.\r\n");
	    return;
	}
    } else {
	/* It's an article number in the current group. */
	number = s.tonum();
	article = spool->find_article(cur_group, number);
	if (!article) {
	    conn->write("423 Bad article number\r\n");
	    return;
	}

	/* Set the current article and number to this article. */
	if (cur_article)
	    spool->release_article(cur_article);
	cur_article = article;
	cur_number = number;
    }

    if (!s.len() || *s != '<') {
	spool->get_article_message_id(article, &buf);
	message_id = buf.str();
    } else {
	message_id = s;
    }

    conn->writef(responses[head][body], number, message_id);
    if (head)
	spool->write_article_head(conn, article);
    if (head && body)
	conn->write("\r\n");
    if (body)
	spool->write_article_body(conn, article);
    if (head || body)
	conn->write(".\r\n");

    /* If the article was specified by message-ID, discard it. */
    if (s.len() && *s == '<')
	spool->release_article(article);
}

int Nntp::select_first_article()
{
    if (cur_group->num_articles() > 0)
	cur_article = spool->first_article(cur_group, &cur_number);
    return (cur_article) ? 0 : -1;
}

/* Parse a range, returning 0 if successful, less than zero if not.  Accept
 * N, -N, N-, N-N, or a blank string, where N is a positive integer.  In the
 * returned bounds, -1 indicates an unspecified value. */
static int get_range(String s, int *low, int *high)
{
    if (!s.len()) {
	*low = *high = -1;
	return 0;
    }

    /* Get the low bound if it's specified. */
    if (isdigit(*s)) {
	*low = s.number();
    } else {
	*low = -1;
    }

    /* Get the upper bound if it's specified. */
    s.skip_whitespace();
    if (s.len() && *s == '-') {
	s.shift(1);
	s.skip_whitespace();
	*high = (s.len() && isdigit(*s)) ? s.tonum() : -1;
    } else {
	*high = *low;
    }

    /* If *low = *high = -1, the range was syntactically incorrect. */
    return (*low == -1 && *high == -1) ? -1 : 0;
}

