/* Bucket.cc: Managing the pool of article buckets. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Pool.h"
#include "Atomic_file.h"
#include "Bucket.h"
#include "Config.h"
#include "String.h"
#include "Stringbuf.h"
#include "Connection.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <dirent.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>
}

Pool::Pool(Spool *spool, const Config *config)
{
    FILE *fp;
    DIR *dir;
    struct dirent direntry, *direntp;
    Stringbuf name, line;
    String entry_name, w[3];
    struct stat statbuf;
    Bucket_loc bloc;
    Bucket_list *entry;

    this->config = config;
    this->spool = spool;

    /* Get directory and checkpoint file pathnames. */
    config->get_pool_directory(&name);
    dirname.store(name.str());
    name.append("checkpoint");
    checkpoint = new Atomic_file(name.str());
    make_directories_in_path(dirname.str());

    /* Find all of the buckets in all of the partitions. */
    for (bloc.partition = 0; ; bloc.partition++) {
	name.format_clear("%s%d", dirname.str(), bloc.partition);
	if (stat(name.cstring(), &statbuf) < 0 || !S_ISDIR(statbuf.st_mode))
	    break;
	partitions.ensure(bloc.partition);
	partitions[bloc.partition].buckets = NULL;
	dir = opendir(name.cstring());
	while (readdir_r(dir, &direntry, &direntp) >= 0 && direntp) {
	    if (!isdigit(*direntry.d_name))
		continue;
	    entry_name = String(direntry.d_name);
	    bloc.bucket_num = entry_name.tonum();

	    /* Make sure the name is numeric. */
	    name.format_clear("%d", bloc.bucket_num);
	    if (compare(name.str(), entry_name) != 0)
		continue;

	    name.format_clear("%s%d/%s", dirname.str(), bloc.partition,
			      entry_name);
	    open_and_add_bucket(&bloc, name.str());
	}
	closedir(dir);
    }
    npart = bloc.partition;

    /* Make sure there's at least one partition. */
    if (npart == 0) {
	name.format_clear("%s0", dirname.str());
	if (mkdir(name.cstring(), S_IRWXU | S_IRWXG | S_IROTH) < 0)
	    die("Can't create initial partition %s", name.str());
	partitions.ensure(0);
	partitions[0].buckets = NULL;
	npart = 1;
    }

    /* Read in the checkpoint record and set the consistent_at offsets. */
    fp = checkpoint->open_read();
    if (fp) {
	while (read_line_clear(&line, fp)) {
	    line.str().get_words(w, 3);
	    bloc.partition = w[0].tonum();
	    bloc.bucket_num = w[1].tonum();
	    entry = search_buckets(&bloc);
	    if (entry)
		entry->consistent_at = w[2].tonum();
	}
	fclose(fp);
    }
}

/* Doesn't unlink from the filesystem; all buckets should be released
 * before this destructor is used. */
Pool::~Pool()
{
    int i;
    Bucket_list *next;

    for (i = 0; i < npart; i++) {
	while (partitions[i].buckets) {
	    next = partitions[i].buckets->next;
	    delete partitions[i].buckets;
	    partitions[i].buckets = next;
	}
    }
}

Bucket *Pool::find_bucket(const Bucket_loc *bloc)
{
    Bucket_list *entry;
    Bucket *bucket;

    /* If we find the entry, bump its refcount so we don't throw away
     * the bucket, and return its bucket.  Otherwise, return NULL. */
    lock();
    entry = search_buckets(bloc);
    if (entry) {
	bucket = &entry->bucket;
	entry->refs++;
    } else {
	bucket = NULL;
    }
    unlock();
    return bucket;
}

int Pool::store_article(Headers *headers, String body, time_t expire_time,
			Artloc *loc, int partition)
{
    Bucket_list *entry;

    /* Decide on a partition if one wasn't specified.  Do a better job of
     * this in the future. */
    if (partition == -1)
	partition = rand() % npart;

    /* Lock the pool until we have a locked bucket to put the article in. */
    lock();

    /* Find and lock a normal bucket with this expire time. */
    for (entry = partitions[partition].buckets; entry; entry = entry->next) {
	entry->bucket.lock();
	if (entry->bucket.type() == BUCKET_NORMAL
	    && entry->bucket.expires() == expire_time)
	    break;
	entry->bucket.unlock();
    }

    /* Create a new bucket and lock it if we didn't find one above. */
    if (!entry) {
	entry = create_bucket(partition, BUCKET_NORMAL, expire_time);
	if (!entry) {
	    unlock();
	    return -1;
	}
	entry->bucket.lock();
    }

    /* We now have a locked bucket to put the article in; unlock the pool while
     * we store the article. */
    unlock();

    /* Write out the article and record its location. */
    loc->partition = partition;
    loc->bucket_num = entry->bucket.number();
    loc->offset = entry->bucket.write_article(headers, body);

    /* We succeeded if Bucket::write_article() didn't return -1. */
    entry->bucket.unlock();
    return (loc->offset >= 0) ? 0 : -1;
}

void Pool::release_bucket(Bucket *bucket)
{
    Bucket_list **entry_ptr, *entry;
    int i;

    lock();

    /* Find the entry whose bucket is bucket. */
    for (i = 0; i < npart; i++) {
	entry_ptr = &partitions[i].buckets;
	for (; *entry_ptr; entry_ptr = &(*entry_ptr)->next) {
	    entry = *entry_ptr;
	    if (&entry->bucket == bucket)
		break;
	}
	if (*entry_ptr)
	    break;
    }
    if (!*entry_ptr) {
	/* Something is amiss. */
	crash("Released bucket %d in partition %d is not in the bucket list.",
	      bucket->number(), bucket->partition());
    }

    /* Release this list entry. */
    release_list_entry(entry_ptr);

    unlock();
}

/* No other thread should touch the pool while this is taking place. */
int Pool::write_checkpoint_record()
{
    FILE *fp;
    Bucket_list *entry;
    Stringbuf line;
    int i;

    fp = checkpoint->open_new();
    if (!fp)
	return -1;

    /* Write out the new file.  Remember the offsets we wrote out as the
     * new consistent_at values in case we do a consistency check between
     * now and the next checkpoint. */
    for (i = 0; i < npart; i++) {
	for (entry = partitions[i].buckets; entry; entry = entry->next) {
	    if (entry->bucket.type() != BUCKET_NORMAL)
		continue;
	    entry->consistent_at = entry->bucket.current_offset();
	    line.format_clear("%d %d %l\n", entry->bucket.partition(),
			      entry->bucket.number(), entry->consistent_at);
	    if (line.str().fwrite(fp) != line.len()) {
		warn("Can't write data to checkpoint record: %m");
		fclose(fp);
		return -1;
	    }
	}
    }
    fclose(fp);

    return checkpoint->update();
}

void Pool::begin_unfiled_articles()
{
    int i;
    Bucket_list *entry;

    for (i = 0; i < npart; i++) {
	for (entry = partitions[i].buckets; entry; entry = entry->next)
	    entry->checked_to = entry->consistent_at;
    }
}

/* No other threads should touch the pool while this is taking place. */
/* Returns 0 if successful, 1 if there are no more, -1 on failure. */
int Pool::next_unfiled_article(int *num_lines, int *num_bytes, Artloc *loc,
			       Headers *headers)
{
    Bucket_list *entry;
    int i, result;

    for (i = 0; i < npart; i++) {
	for (entry = partitions[i].buckets; entry; entry = entry->next) {
	    loc->partition = i;
	    loc->bucket_num = entry->bucket.number();
	    loc->offset = entry->checked_to;
	    result = entry->scan_article(entry->checked_to, num_lines,
					 num_bytes, headers);
	    if (result == 0)
		entry->checked_to += *num_bytes;
	    if (result != 1)
		return result;
	}
    }
    return 1;
}

int Pool::soonest_bucket_to_expire(Bucket_loc *loc, time_t *expire_time)
{
    Bucket_list *b;
    int i;
    bool found = 0;

    for (i = 0; i < npart; i++) {
	for (b = partitions[i].buckets; b; b = b->next) {
	    if (!found || (b->bucket.type() == BUCKET_NORMAL
			   && b->bucket.expires() != -1
			   && b->bucket.expires() < *expire_time)) {
		*expire_time = b->bucket.expires();
		loc->partition = b->bucket.partition();
		loc->bucket_num = b->bucket.number();
		found = 1;
	    }
	}
    }
    return (found) ? 0 : -1;
}

void Pool::expire_bucket(const Bucket_loc *bloc)
{
    Bucket_list **entry_ptr;
    int i;

    /* Release the list entries for any buckets numbered bucket_num. */
    lock();
    for (i = 0; i < npart; i++) {
	entry_ptr = &partitions[i].buckets;
	for (; *entry_ptr; entry_ptr = &(*entry_ptr)->next) {
	    if ((*entry_ptr)->bucket.partition() == bloc->partition
		&& (*entry_ptr)->bucket.number() == bloc->bucket_num) {
		release_list_entry(entry_ptr);
		break;
	    }
	}
    }
    unlock();
}

int Pool::mark_cancelled(const Artloc *loc)
{
    Bucket_loc bloc = { loc->partition, loc->bucket_num };
    Bucket_list *entry;

    /* XXX drain going on at same time? */
    lock();
    entry = search_bucket(&bloc);
    unlock();
    return 0;
}

void Pool::send_bucket_list(Connection *conn)
{
    int i;
    Bucket_list *entry;
    Stringbuf buf;

    /* Make sure we don't block writing to the client with the pool locked. */
    lock();
    for (i = 0; i < npart; i++) {
	for (entry = partitions[i].buckets; entry; entry = entry->next) {
	    buf.format("Partition %d bucket %d\r\n", i,
		       entry->bucket.number());
	}
    }
    unlock();
    conn->write(buf.str());
}

/* Assumes lock. */
void Pool::release_list_entry(Bucket_list **entry_ptr)
{
    Bucket_list *entry = *entry_ptr;

    entry->refs--;
    if (entry->refs == 0) {
	Stringbuf filename;

	/* The refcount hit zero, so this bucket is due to be removed.
	 * Unlink the bucket from the file system and delete it.  Later
	 * we need to worry about moving open articles from this bucket
	 * into temporary files. */
	filename.format("%s%d/%d", dirname.str(), entry->bucket.partition(),
			entry->bucket.number());
	unlink(filename.cstring());

	/* Remove this entry from the list. */
	*entry_ptr = entry->next;
	delete entry;
    }
}

/* Assumes initialization phase. */
void Pool::open_and_add_bucket(const Bucket_loc *bloc, String name)
{
    FILE *fp;
    Bucket_list *entry;
    Stringbuf buf(name), line;
    Bucket_type type;
    time_t expire_time;

    fp = fopen(buf.cstring(), "r+");
    if (fp == NULL)
	die("Could not open bucket %s: %m", name);

    /* Read in tag line. */
    if (read_line(&line, fp) < 0) {
	warn("Deleting untagged bucket %d in partition %d", bloc->bucket_num,
	     bloc->partition);
	unlink(buf.cstring());
	return;
    }
    if (line.len() == 0) {
	die("Corrupt tag line for bucket %d in partition %d: %s",
	    bloc->bucket_num, bloc->partition, line.str());
    }
    if (*line.str() == 't') {
	log("Deleting temporary bucket %d in partition %d", bloc->bucket_num,
	    bloc->partition);
	unlink(buf.cstring());
	return;
    }
    if (*line.str() == 'n')
	type = BUCKET_NORMAL;
    else if (*line.str() == 'd')
	type = BUCKET_DRAINING;
    expire_time = line.str().after(1).tonum();

    entry = new Bucket_list(fp, name, bloc, type, expire_time);
    entry->next = partitions[bloc->partition].buckets;
    partitions[bloc->partition].buckets = entry;
}

/* Assumes lock or initialization phase. */
Bucket_list *Pool::search_buckets(const Bucket_loc *bloc)
{
    Bucket_list *entry;

    if (bloc->partition < 0 || bloc->partition >= npart)
	return NULL;

    /* Find the entry whose bucket's number is number. */
    for (entry = partitions[bloc->partition].buckets; entry;
	 entry = entry->next) {
	if (entry->bucket.number() == bloc->bucket_num)
	    return entry;
    }
    return NULL;
}

/* Assumes lock.  type must be BUCKET_NORMAL or BUCKET_TEMPORARY. */
Bucket_list *Pool::create_bucket(int partition, Bucket_type type,
				 time_t expire_time)
{
    Bucket_list *entry;
    FILE *fp;
    Stringbuf name, line;
    Bucket_loc bloc = { partition, 1 };

    /* Find an unused bucket number. */
    for (entry = partitions[partition].buckets; entry; entry = entry->next) {
	if (entry->bucket.number() >= bloc.bucket_num)
	    bloc.bucket_num = entry->bucket.number() + 1;
    }

    /* Create the bucket file. */
    name.format("%s%d/%l", dirname.str(), partition, bloc.bucket_num);
    fp = fopen(name.cstring(), "w+");
    if (!fp) {
	error("Can't create bucket file %s: %m", name.str());
	return NULL;
    }
    if (type == BUCKET_NORMAL)
	line.format("n%l\n", (long) expire_time);
    else
	line.store("t\n");
    if (line.str().fwrite(fp) != line.len()) {
	error("Can't write tag line to bucket file %s: %m", name.str());
	fclose(fp);
	unlink(name.cstring());
	return NULL;
    }

    /* Chain the new bucket into the list for the partition. */
    entry = new Bucket_list(fp, name.str(), &bloc, BUCKET_NORMAL,
			    expire_time);
    entry->next = partitions[partition].buckets;
    partitions[partition].buckets = entry;

    return entry;
}

Bucket_list::Bucket_list(FILE *fp, String name, const Bucket_loc *bloc,
			 Bucket_type type, time_t expire_time)
    : bucket(fp, name, bloc, type, expire_time)
{
    refs = 1;
    consistent_at = 0;
}
