/* Spool.cc: Implementation of the news spool. */

/* See ../doc/INTERNALS for notes on the spool organization. */

#include "threads.h"
#include "util.h"
#include "error.h"

#include "Spool.h"
#include "Index.h"
#include "Groups.h"
#include "Pool.h"
#include "Bucket.h"
#include "Newsgroup.h"
#include "Article.h"
#include "Config.h"
#include "Connection.h"

extern "C" {
#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
}

struct Drain_data {
    Spool *spool;
    Bucket *bucket;
}

static void *periodic_checkpoints_thread(void *arg);
static void *periodic_expire_thread(void *arg);
static void *drain_thread(void *arg);

Spool::Spool(const Config *config_arg)
{
    Stringbuf buf;

    config = config_arg;
    index = new Index(config);
    groups = new Groups(config);
    pool = new Pool(this, config);
    feeds = new Feeds(this, config);
    config->get_server_name(&buf);
    server_name.store(buf.str());
    throttle_message = NULL;

    /* Make sure the spool is consistent, or disallow writing. */
    if (consistency_check() < 0)
	throttle("Consistency check failed.");

    start_thread(periodic_checkpoints_thread, this);
    start_thread(periodic_expire_thread, this);
}

Spool::~Spool()
{
    delete index;
    delete groups;
    delete pool;
    delete feeds;
}

Newsgroup *Spool::find_group(String name)
{
    return groups->find_group(name);
}

Article *Spool::find_article(String id)
{
    Artloc loc;

    if (index->lookup(id, &loc) < 0)
	return NULL;
    return get_article(&loc);
}

Article *Spool::find_article(Newsgroup *group, int number)
{
    Artloc loc;

    if (group->lookup(number, &loc) < 0)
	return NULL;
    return get_article(&loc);
}

Article *Spool::next_article(Newsgroup *group, int *number)
{
    Artloc loc;

    if (group->next(number, &loc) < 0)
	return NULL;
    return get_article(&loc);
}

Article *Spool::prev_article(Newsgroup *group, int *number)
{
    Artloc loc;

    if (group->prev(number, &loc) < 0)
	return NULL;

    return get_article(&loc);
}

Article *Spool::first_article(Newsgroup *group, int *number)
{
    Artloc loc;

    if (group->first(number, &loc) < 0)
	return NULL;

    return get_article(&loc);
}

Article *Spool::get_article(const Artloc *loc)
{
    Bucket_loc bloc = { loc->partition, loc->bucket_num };
    Bucket *bucket;
    Article *article;

    /* Find the bucket corresponding to the given bucket number. */
    bucket = pool->find_bucket(&bloc);
    if (bucket == NULL)
	return NULL;

    /* Get the article from the bucket, release the bucket, and return. */
    article = bucket->get_article(loc->offset);
    pool->release_bucket(bucket);
    return article;
}

int Spool::article_exists(String id)
{
    Artloc loc;

    /* Perform an index lookup, return true if it succeeds. */
    return (index->lookup(id, &loc) >= 0);
}

int group_num_articles(Newsgroup *group)
{
    return group->num_articles();
}

int group_first_number(Newsgroup *group)
{
    return group->first_number();
}

int group_last_number(Newsgroup *group)
{
    return group->last_number();
}

void write_article_head(Connection *conn, Article *article)
{
    return article->write_head(conn);
}

void write_article_body(Connection *conn, Article *article)
{
    return article->write_body(conn);
}

void get_article_message_id(Article *article, Stringbuf *buf)
{
    return article->get_message_id(buf);
}

void Spool::release_group(Newsgroup *group)
{
    groups->release_group(group);
}

void Spool::release_article(Article *article)
{
    Bucket *bucket = article->bucket();

    bucket->release_article(article);
}

int Spool::article_wanted(String id)
{
    return !index->present(id);
}

/* The ihave command has to:
 * - Verify that all required headers on the article exist.
 * - Verify that the Message-ID field is the same as the one in the ihave
 *   command.
 * - Verify that the Newsgroups line doesn't contain any invalid
 *   characters.
 * - Verify that we want at least one of the specified distributions.
 * - Store and propagate the article. */
Post_status Spool::ihave_article(String id, String article,
				 const Wildcard_decision_list *access,
				 Stringbuf *message)
{
    static String required[] = { "From", "Date", "Newsgroups", "Subject",
				 "Message-ID", "Path", NULL };
    Headers headers;
    int num_lines;
    String *req, body, message_id, newsgroups, dist;

    /* Fill headers with header lines. */
    body = headers.add_headers(article);
    if (!body.valid()) {
	message->append("Malformatted headers.");
	return POST_REJECTED;
    }

    /* Verify that all required headers exist. */
    for (req = required; req->valid(); req++) {
	if (!headers[*req].valid()) {
	    message->format("Missing header %s", req);
	    return POST_REJECTED;
	}
    }

    /* Verify that the message ID is equivalent to id. */
    message_id = headers["Message-ID"];
    if (!message_id_equal(message_id, id)) {
	message->append("Message-ID not equivalent to ihave ID.");
	return POST_REJECTED;
    }

    /* Verify that the newsgroup line contains no invalid characters. */
    newsgroups = headers["Newsgroups"];
    if (newsgroups.find_set(":/ \f\t\n") >= 0) {
	message->append("Invalid characters in newsgroup line.");
	return POST_REJECTED;
    }

    /* Verify that we accept at least one distribution. */
    dist = headers["Distribution"];
    if (dist.valid() && config->any_distribution_accepted(dist) < 0) {
	message->format("Server does not accept distributions \"%s\".", dist);
	return POST_REJECTED;
    }

    /* If any group is moderated, verify that there is an Approved header. */
    if (groups->any_groups_moderated(newsgroups)
	&& !headers["Approved"].valid()) {
	message->append("Unapproved article posted to moderated group.");
	return POST_REJECTED;
    }

    /* Count lines and bytes. */
    num_lines = body.count_char('\n');

    return transfer_article(&headers, body, id, num_lines, access, message);
}

/* The post command needs to:
 * - Error if From or Newsgroups doesn't exist, if neither of Subject
 *   and Control exist, or if NNTP-Posting-Host, Xref, Date-Received,
 *   Received, Posted, Posting-Version, or Relay-Version do exist.
 * - Verify the Date, Expires, Distribution, and Message-ID headers
 *   if they exist.
 * - Generate Date, Message-ID, Path, Organization, MIME-Version,
 *   Content-Type, Lines, and NNTP-Posting-Host headers if they don't
 *   exist.
 * - Override Subject with Control header if a Control header exists;
 *   otherwise, generate Control header from Subject if Subject begins
 *   with "cmsg".
 * - Verify that posting is allowed to all newsgroups in the Newsgroups
 *   header and, if it exists, the Followup-To header.
 * - Mail the article if it is to a moderated group.
 * - Otherwise, store and propagate the article. */
int Spool::post_article(String article, String posting_host,
			const Wildcard_decision_list *access,
			Stringbuf *message)
{
    Headers headers;
    String body, newsgroups, control, subject, date, expires, dist, id, follow;
    int num_lines;
    time_t t, now;
    Stringbuf buf, mod;

    body = headers.add_headers(article);
    if (!body.valid()) {
	message->append("Malformatted headers.");
	return -1;
    }
    num_lines = body.count_char('\n');

    /* Pull out some newsgroup headers we want to examine. */
    newsgroups = headers["Newsgroups"];
    control = headers["Control"];
    subject = headers["Subject"];
    date = headers["Date"];
    expires = headers["Expires"];
    dist = headers["Distribution"];
    id = headers["Message-ID"];
    follow = headers["Followup-To"];

    /* Verify that Newsgroups and From headers exist, and that either Control
     * or Subject exists. */
    if (!newsgroups.valid() || (!control.valid() && !subject.valid())
	|| !headers["From"].valid()) {
	message->append("Missing From, Newsgroups, or Subject header.");
	return -1;
    }

    /* Complain about system or obsolete headers. */
    if (headers["NNTP-Posting-Host"].valid() || headers["Xref"].valid()) {
	message->append("NNTP-Posting-Host or Xref set in headers.");
	return -1;
    }
    if (headers["Date-Received"].valid() || headers["Received"].valid()
	|| headers["Posted"].valid() || headers["Posting-Version"].valid()
	|| headers["Relay-Version"].valid()) {
	message->append("Obsolete article headers present.");
	return -1;
    }

    /* Verify the Date header if it exists, or fill it in if it doesn't. */
    if (date.valid()) {
	t = parse_header_date(date);
	time(&now);
	if (t == -1 || t > now) { /* XXX should allow clock skew */
	    message->append("Date header is invalid or in the future.");
	    return -1;
	}
    } else {
	buf.store("Date: ");
	generate_current_header_date(&buf);
	headers.add_header(buf.str());
    }

    /* Verify the Expires header if it exists. */
    if (expires.valid() && parse_header_date(expires) == -1) {
	message->append("Expires header contains invalid date.");
	return -1;
    }

    /* Verify the Distribution header if it exists. */
    if (dist.valid() && config->any_distribution_accepted(dist) < 0) {
	message->format("Server does not accept distributions \"%s\"", dist);
	return -1;
    }

    /* Verify that posting is allowed in Newsgroups and Followup-To:
     * headers. */
    if (newsgroups.find_set(":/ \f\t\n") != NULL ||
	(follow.valid() && follow.find_set(":/ \f\t\n") != NULL)) {
	message->append("Invalid Newsgroups or Followup-To header.");
	return -1;
    }
    buf.clear();
    if (groups->verify_post_access(newsgroups, access, &buf, &mod) < 0 ||
	(follow.valid() &&
	 groups->verify_post_access(follow, access, &buf, NULL) < 0)) {
	message->format("Posting not allowed to \"%s\".", buf.str());
	return -1;
    }

    /* Verify the Message-ID header if it exists, or generate one if not. */
    if (id.valid() && verify_message_id(id) < 0) {
	message->format("Invalid Message-ID \"%s\"", id);
	return -1;
    } else if (!id.valid()) {
	buf.store("Message-ID: ");
	generate_message_id(&buf);
	headers.add_header(buf.str());
	id = headers["Message-ID"];
    }

    /* Generate Path, Organization, MIME-Version, Content-Type, Lines, and
     * NNTP-Posting-Host headers if they don't exist. */
    if (!headers["Path"].valid())
	headers.add_header("Path: usenet");
    if (!headers["Organization"].valid()) {
	buf.store("Organization: ");
	config->get_default_organization(&buf);
	headers.add_header(buf.str());
    }
    if (config->get_add_mime_headers_flag()) {
	if (!headers["MIME-Version"].valid())
	    headers.add_header("MIME-Version: 1.0");
	if (!headers["Content-Type"].valid())
	    headers.add_header("Content-Type: text/plain");
    }
    buf.format_clear("Lines: %d", num_lines);
    headers.add_header(buf.str());
    buf.format_clear("NNTP-Posting-Host: %s", posting_host);
    headers.add_header(buf.str());

    /* Make sure Subject and Control headers agree. */
    if (control.valid()) {
	buf.format_clear("Subject: cmsg %s", control);
	headers.add_header(buf.str());
    } else if (subject.len() > 5 && !compare_case(subject.left(5), "cmsg ")) {
	subject.shift(5);
	subject.skip_whitespace();
	buf.format_clear("Control: %s", subject);
	headers.add_header(buf.str());
    }

    if (mod.len()) {
	return mail_article(&headers, body, mod.str(), message);
    } else {
	return (transfer_article(&headers, body, id, num_lines, access,
				 message) == POST_OK) ? 0 : -1;
    }
}

int Spool::refile_article(String article, const Artloc *oldloc)
{
    Headers headers;
    String body, newsgroups, id, xref, control, expires;
    int num_lines;
    Post_status status;
    Artloc loc;
    Id_entry *entry;
    time_t desired_expire_time, expire_time;

    /* Parse into headers and body. */
    body = headers.add_headers(article);
    if (!body.valid()) {
	error("Malformatted headers in refiled article.");
	return -1;
    }
    num_lines = body.count_char('\n');

    /* Get some necessary headers. */
    newsgroups = headers["Newsgroups"];
    id = headers["Message-ID"];
    xref = headers["Xref"];
    if (!newsgroups.valid() || !id.valid()
	|| (!control.valid() && !xref.valid())) {
	error("Missing headers in refiled article.");
	return -1;
    }
    control = headers["Control"];

    /* Determine the new expire time for the article. */
    expires = headers["Expires"];
    desired_expire_time = (expires.valid()) ? parse_header_date(expires) : -1;
    expire_time = config->decide_expire_time(newsgroups, desired_expire_time);

    transactions.beginning_refile();

    /* Store the article in a new place in the pool (same partition),
     * overwrite its index entry, overwrite its entries in the groups files,
     * and replace any references to it in newsfeeds. */
    if (pool->store_article(headers, body, expire_time, &loc,
			    oldloc.partition) < 0) {
	error("Failed to store refiled article.");
    } else if (index->overwrite(id, &loc) < 0) {
	error("Failed to overwrite index entry for refiled article.");
    } else if (!control.valid() &&
	       groups->add_to_groups(headers, num_lines, body.len(),
				     &loc, 1) < 0) {
	error("Failed to overwrite group entries for refiled article.");
    } else if (feeds->replace_article_references(oldloc, &loc) < 0) {
	error("Failed to overwrite feed references for refiled article.");
    } else {
	transactions.refile_done();
	return 0;
    }

    transactions.refile_done();
    return -1;
}

Post_status Spool::transfer_article(Headers *headers, String body, String id,
				    int num_lines,
				    const Wildcard_decision_list *access,
				    Stringbuf *message)
{
    Artloc loc;
    Bucket *bucket;
    time_t desired_expire_time, expire_time;
    Post_status status;
    Id_entry *entry;
    Stringbuf buf;
    String expires, control, newsgroups;

    control = (*headers)["Control"];
    newsgroups = (*headers)["Newsgroups"];

    /* Add the name of this server to the path. */
    buf.format_clear("Path: %s!%s", server_name.str(), (*headers)["Path"]);
    headers->add_header(buf.str());

    /* Add an Xref line. */
    if (!control.valid()) {
	buf.format_clear("Xref: %s", server_name.str());
	groups->reserve_positions(newsgroups, access, &buf);
	headers->add_header(buf.str());
    }

    /* Record that we've started the post, or return an appropriate status
     * if we collided with another thread (POST_REJECTED if it committed
     * or rejected the article, POST_FAILED if it throttled trying to post
     * the article). */
    entry = transactions.beginning_post(id, &status);
    if (!entry) {
	if (status == POST_REJECTED)
	    message->append("Already have this article.");
	else if (status == POST_FAILED)
	    message->append("Can't write article to disk.");
	return status;
    }

    /* Reject the article if we already have it. */
    if (index->present(id)) {
	transactions.post_committed(entry);
	message->append("Already have this article.");
	return POST_REJECTED;
    }

    /* Accept the article later if we are throttled. */
    throttle_lock.lock();
    if (throttle_message) {
	transactions.post_failed(entry);
	message->append(throttle_message->str());
	return POST_FAILED;
    }
    throttle_lock.unlock();

    /* Add the article to the appropriate bucket. */
    expires = (*headers)["Expires"];
    desired_expire_time = (expires.valid()) ? parse_header_date(expires) : -1;
    expire_time = config->decide_expire_time(newsgroups, desired_expire_time);
    if (pool->store_article(headers, body, expire_time, &loc) < 0) {
	throttle("Failed to store article.");
	transactions.post_failed(entry);
	message->append("Can't write article to disk.");
	return POST_FAILED;
    }

    /* Add the article to the index. */
    if (index->store(id, &loc) < 0) {
	throttle("Can't write index entries.");
	transactions.post_committed(entry);
	message->append("Article stored, but not indexed or in groups.");
	return POST_REJECTED;
    }

    /* If this is a control message, process it; otherwise add the article
     * to its newsgroups. */
    if (control.valid()) {
	handle_control_message(headers, body);
    } else if (groups->add_to_groups(headers, num_lines, body.len(), &loc,
				     0) < 0) {
	throttle("Can't write to group file.");
	transactions.post_committed(entry);
	message->append("Article stored and indexed, but not in groups.");
	return POST_REJECTED;
    }

    /* Propagate the article.  If this fails, throttle, but report success
     * in transferring the article; we will repropagate the article during
     * the next consistency check. */
    if (feeds->propagate_article(&loc, headers, body) == -1)
	throttle("Can't write propagation bookkeeping entry.");

    transactions.post_committed(entry);
    message->append("Article transferred ok.");
    return POST_OK;
}

/* Assumes no modification while this is taking place, and only one call
 * at a time. */
int Spool::reindex_article(int num_lines, int num_bytes, const Artloc *loc,
			   const Headers *headers)
{
    String groupnames, id;

    groupnames = (*headers)["Newsgroups"];
    id = (*headers)["Message-ID"];
    if (!groupnames.valid() || !id.valid())
	return -1;

    if (!index->present(id) < 0) {
	if (index->store(id, loc) < 0)
	    return -1;
    }

    if (!(*headers)["Control"].valid()) {
	if (groups->add_to_groups(headers, num_lines, num_bytes, loc, 1) < 0)
	    return -1;
    }

    if (feeds->propagate_article(loc, headers, NULL) < 0)
	return -1;

    /* XXX Unfinished control messages. */
    return 0;
}

void Spool::handle_control_message(Headers *headers, String body)
{
    String cmd, rest, groups;

    rest = body.get_words(&cmd, 1);
    groups = (*headers)["Newsgroups"];

    if (compare_case(cmd, "cancel") == 0 && rest.len())
	control_cancel(rest, (*headers)["From"], groups);
    else if (compare_case(cmd, "ihave") == 0)
	control_ihave(rest, body, groups);
    else if (compare_case(cmd, "sendme") == 0)
	control_sendme(rest, body, groups);
    else if (compare_case(cmd, "newgroup") == 0)
	control_newgroup(rest, body, (*headers)["From"]);
    else if (compare_case(cmd, "rmgroup") == 0)
	control_rmgroup(rest, (*headers)["From"]);
}

int Spool::mail_article(Headers *headers, String body, String mod,
			Stringbuf *message)
{
    FILE *mail;
    Stringbuf buf;
    String line;
    int status;

    config->get_sendmail_file(&buf);
    mail = popen(buf.cstring(), "w");
    if (!mail) {
	error("Can't start mailer %s", buf.str());
	message->append("Can't start mailer to mail article to moderator.");
	return -1;
    }

    buf.format_clear("To: %s\n", mod);
    buf.str().fwrite(mail);
    for (line = headers->first_header(); line.valid();
	 line = headers->next_header()) {
	line.fwrite(mail);
	putc('\n', mail);
    }

    putc('\n', mail);
    body.fwrite(mail);
    status = ferror(mail);
    status = (pclose(mail) == 0) ? status : -1;
    if (status < 0) {
	message->append("Can't mail article to moderator.");
	return -1;
    }

    return 0;
}

void Spool::create_group(String group, String flags, String desc)
{
    groups->add_group(group, flags, desc, time(NULL));
}

void Spool::merge_groups(FILE *fp)
{
    groups->merge_groups(fp);
}

void Spool::send_grouplist(Connection *conn,
			   const Wildcard_decision_list *access)
{
    groups->send_grouplist(conn, access);
}

void Spool::send_group_descriptions(Connection *conn,
				    const Wildcard_decision_list *access)
{
    groups->send_group_descriptions(conn, access);
}

void Spool::send_new_groups(Connection *conn, time_t mintime, String dist,
			    const Wildcard_decision_list *access)
{
    groups->send_new_groups(conn, mintime, dist, access);
}

void Spool::send_schema(Connection *conn)
{
    const Stringlist *schema;
    int i;

    schema = config->get_schema();
    for (i = 0; i < schema->len(); i++) {
	if ((*schema)[i].len() && *(*schema)[i].str() == ':')
	    conn->writef("%s:full\r\n", (*schema)[i].str().after(1));
	else
	    conn->writef("%s:\r\n", (*schema)[i].str());
    }
}

int Spool::find_header(const Artloc *loc, String name, Stringbuf *buf)
{
    Bucket_loc bloc = { loc->partition, loc->bucket_num };
    Bucket *bucket;
    int result;

    /* Find the bucket. */
    bucket = pool->find_bucket(&bloc);
    if (bucket == NULL)
	return -1;

    /* Look up the header, release the bucket, and return. */
    result = bucket->find_header(loc->offset, name, buf);
    pool->release_bucket(bucket);
    return result;
}

int Spool::find_header(String id, String name, Stringbuf *buf)
{
    Artloc loc;

    if (index->lookup(id, &loc) < 0)
	return -1;
    return find_header(&loc, name, buf);
}

void Spool::send_bucket_list(Connection *conn)
{
    pool->send_bucket_list(conn);
}

int Spool::drain_bucket(const Bucket_loc *bloc)
{
    Bucket *bucket;

    bucket = pool->find_bucket(bloc);
    if (!bucket)
	return -1;

    data = new Drain_data;
    data->spool = this;
    data->bucket = bloc;
    start_thread(drain_thread, data)
}

void Spool::throttle(String message, ...)
{
    va_list args;

    throttle_lock.lock();
    if (!throttle_message)
	throttle_message = new Stringbuf;
    va_start(args, message);
    throttle_message->format_clear(message, args);
    va_end(args);
    throttle_lock.unlock();
}

void Spool::unthrottle()
{
    if (consistency_check() < 0)
	return;

    throttle_lock.lock();
    delete throttle_message;
    throttle_message = NULL;
    throttle_lock.unlock();
}

void Spool::checkpoint()
{
    transactions.quiesce();
    if (!throttle_message) {
	if (groups->write_new_bounds_file() < 0)
	    throttle("Can't write updated grouplist.");
	else if (index->sync() < 0)
	    throttle("Can't synchronize article index.");
	else if (feeds->sync_queues() < 0)
	    throttle("Can't flush newsfeed queues to disk.");
	else if (pool->write_checkpoint_record() < 0)
	    throttle("Can't write updated consistency record.");
    }
    transactions.quiesce_done();
}

void Spool::expire()
{
    Bucket_loc bloc;
    time_t expire_time, now;

    time(&now);
    while (1) {
	if (pool->soonest_bucket_to_expire(&bloc, &expire_time) < 0)
	    break;
	if (expire_time > now)
	    break;
	log("Expiring bucket %d in partition %d", bloc.bucket_num,
	    bloc.partition);
	transactions.quiesce();
	if (index->eliminate_bucket(&bloc, expire_time) < 0)
	    throttle("Not enough space to rewrite index.");
	transactions.quiesce_done();
	if (groups->eliminate_bucket(&bloc) < 0)
	    throttle("Not enough space to rewrite group files.");
	pool->expire_bucket(&bloc);
    }
}

void Spool::drain(Bucket *bucket)
{
    Stringbuf article;
    long offset;

    bucket->drain_begin();
    while (1) {
	if (bucket->drain_last_article(&article, &offset) != 0)
	    break;
	if (refile_article(article.str()) < 0)
	    break;
	if (bucket->drain_truncate(offset) < 0)
	    break;
    }
    /* ... at this point, we should remove the bucket if successful. */
}

int Spool::consistency_check()
{
    int result, num_lines, num_bytes;
    Headers headers;
    Stringbuf article;
    Artloc loc;

    /* Quiesce all transactions.  Normally there won't be any, since the
     * server is throttled or hasn't started accepting connections.  But
     * since posting is asynchronous, there could still be some stray
     * posting operations. */
    transactions.quiesce();

    /* Reindex every potentially unindexed article. */
    pool->begin_unfiled_articles();
    while (1) {
	result = pool->next_unfiled_article(&num_lines, &num_bytes, &loc,
					    &headers);
	if (result != 0)
	    break;
	result = reindex_article(num_lines, num_bytes, &loc, &headers);
	if (result < 0)
	    break;
    }

    /* If we finished, checkpoint. */
    if (result == 1) {
	result = 0;
	if (groups->write_new_bounds_file() < 0 || index->sync() < 0 ||
	    feeds->sync_queues() < 0 || pool->write_checkpoint_record() < 0)
	    result = -1;
    }

    transactions.quiesce_done();
    return result;
}

/* Message ID format similar to INN's, representing numbers using a
 * 32-character alphabet to represent the process ID, the time, and a
 * 15-bit sequence number.  (INN doesn't use the sequence number.) */
void Spool::generate_message_id(Stringbuf *buf)
{
    const time_t OFFSET = 673416000L;
    unsigned long pid, t, seq;

    pid = getpid();
    t = time(NULL) - OFFSET;
    seq = sequence.next();
    buf->format("<%R$%R$%R@%S>", pid, t, seq, &server_name);
}

/* To cancel an article which has already been posted:
 * 	- Remove all references to the article from all group files it's in.
 *	- Rewrite the index entry to the article as a dead entry.
 *	- Mark the article as cancelled in the buckets.
 */
void Spool::control_cancel(String ids, String, String)
{
    String idword;
    Id_entry *entry;

    for (idword = ids.word(); idword.len(); idword = ids.word()) {
	entry = transactions.beginning_cancel(idword);
	cancel_article(idword);
	transactions.cancel_done(entry);
    }
}

/* Stub. */
void Spool::control_ihave(String, String, String)
{
    ;
}

/* Stub. */
void Spool::control_sendme(String, String, String)
{
    ;
}

void Spool::control_newgroup(String s, String body, String from)
{
    String w[2], rest, flags = "y", desc = "";
    int pos;

    rest = s.get_words(w, 2);

    /* If second word exists, it must be "moderated" or "unmoderated". */
    if (compare_case(w[1], "moderated") == 0 && !rest.len()) {
	flags = "m";
    } else if (compare_case(w[1], "unmoderated") == 0 && !rest.len()) {
	flags = "y";
    } else if (w[1].len()) {
	log("Bad newgroup syntax: %s", s);
    }

    /* Check the newsgroup name for bad characters. */
    if (w[0].find_set(":/\f\t\n")) {
	log("Bad newgroup group name: %s", s);
	return;
    }

    /* Check permissions on creating a new group. */
    if (!config->get_newgroup_allow(w[0], from)) {
	log("Unapproved newgroup for from %s: %s", from, s);
	return;
    }

    /* Look for a description in the body. */
    pos = body.find("\nFor your newsgroups file:\n");
    if (pos != -1) {
	desc = body.after(27);
	desc.word();
	pos = desc.find('\n');
	if (pos >= 0)
	    desc.trunc(pos);
	desc.trunc_whitespace();
    }

    /* Create the group. */
    log("Creating group requested by %s: %s", from, s);
    create_group(w[0], flags, desc);

    /* For failure atomicity, flush the grouplist file. */
    groups->write_new_bounds_file();
}

/* Stub. */
void Spool::control_rmgroup(String, String)
{
    ;
}

/* Assumes a cancel transaction on id is in effect. */
/* XXX get transaction? */
void Spool::cancel_article(String id)
{
    Stringbuf xref;
    Artloc loc;
    Id_entry *entry;

    entry = transactions.beginning_cancel(id, &status);

    if (index->lookup(id, &loc) < 0) {
	/* We don't have the article.  If there's not already a dead entry
	 * for the message ID, make one. */
	if (!index->present(id))
	    index->store_dead_entry(id);
    } else {
	if (find_header(&loc, "Xref", &xref) < 0) {
	    error("Can't get Xref header for canceled article %s.", id);
	    return;
	}
	if (groups->cancel_article(xref.str()) < 0)
	    return;
	if (index->overwrite_with_dead_entry(id) < 0)
	    return;
	pool->mark_cancelled(&loc);
    }

    transactions.cancel_done(entry);
}

static void *periodic_checkpoints_thread(void *arg)
{
    while (1) {
	sleep(60);
	((Spool *) arg)->checkpoint();
    }
}

static void *periodic_expire_thread(void *arg)
{
    while (1) {
	sleep(60);
	((Spool *) arg)->expire();
    }
}

static void *drain_thread(void *arg)
{
    Drain_data *data = (Drain_data *) arg;
    Spool *spool = data->spool;
    Bucket *bucket = data->bucket;

    delete data;
    spool->drain(bucket);
}
