/* Spool.h: Interface to the news spool (thread-safe). */

#ifndef SPOOL_H
#define SPOOL_H

struct Bucket_loc {
    int partition;
    int bucket_num;
};

struct Artloc {
    int partition;
    int bucket_num;
    long offset;
};

#include "util.h"

#include "Transactions.h"
#include "String.h"
#include "Feeds.h"

extern "C" {
#include <sys/types.h>
}

class Config;
class Article;
class Newsgroup;
class Index;
class Groups;
class Pool;
class Bucket;
class Feeds;
class Headers;
class Connection;
class Stringbuf;
class Wildcard_decision_list;

class Spool {
    const Config *config;
    Index *index;
    Groups *groups;
    Pool *pool;
    Feeds *feeds;
    Transactions transactions;
    Stringbuf *throttle_message;
    Mutex throttle_lock;
    Stored_string server_name;
    Counter sequence;

    int consistency_check();
    int decide_bucket(String expires);
    Post_status transfer_article(Headers *headers, String body,
				 String id, int num_lines,
				 const Wildcard_decision_list *access,
				 Stringbuf *message);
    int reindex_article(int num_lines, int num_bytes, const Artloc *loc,
			const Headers *headers);
    void handle_control_message(Headers *headers, String body);
    int mail_article(Headers *headers, String body, String mod,
		     Stringbuf *message);
    void generate_message_id(Stringbuf *buf);
    void control_cancel(String ids, String from, String groups);
    void control_ihave(String ids, String body, String groups);
    void control_sendme(String ids, String body, String groups);
    void control_newgroup(String s, String body, String from);
    void control_rmgroup(String group, String from);
    void cancel_article(String id);

  public:
    Spool(const Config *config_arg);
    ~Spool();

    /* Looking up newsgroups and articles. */
    Newsgroup *find_group(String name);
    Article *find_article(String id);
    Article *find_article(Newsgroup *group, int number);
    Article *next_article(Newsgroup *group, int *number);
    Article *prev_article(Newsgroup *group, int *number);
    Article *first_article(Newsgroup *group, int *number);
    Article *get_article(const Artloc *loc);
    int article_exists(String id);

    /* Operations on newsgroups and articles. */
    int group_num_articles(Newsgroup *group);
    int group_first_number(Newsgroup *group);
    int group_last_number(Newsgroup *group);
    void write_article_head(Connection *conn, Article *article);
    void write_article_body(Connection *conn, Article *article);
    void get_article_message_id(Article *article, Stringbuf *buf);
    void release_group(Newsgroup *group);
    void release_article(Article *article);

    int article_wanted(String id);
    Post_status ihave_article(String id, String article,
			      const Wildcard_decision_list *access,
			      Stringbuf *message);
    int post_article(String article, String posting_host,
		     const Wildcard_decision_list *access, Stringbuf *message);
    int refile_article(String article);
    void create_group(String group, String flags, String desc);
    void merge_groups(FILE *fp);

    void send_grouplist(Connection *conn,
			const Wildcard_decision_list *access);
    void send_group_descriptions(Connection *conn,
				 const Wildcard_decision_list *access);
    void send_new_groups(Connection *conn, time_t mintime, String dist,
			 const Wildcard_decision_list *access);
    void send_schema(Connection *conn);
    int find_header(const Artloc *loc, String name, Stringbuf *buf);
    int find_header(String id, String name, Stringbuf *buf);

    void send_bucket_list(Connection *conn);
    int drain_bucket(const Bucket_loc *bloc);

    void throttle(String message, ...);
    void unthrottle();

    void checkpoint();
    void expire();
    void drain(Bucket *bucket);
};

#endif

