/* String.cc: Implementation of nontrivial read-only string methods. */

#include "threads.h"

#include "String.h"

extern "C" {
#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
}

ssize_t String::write(int fd) const
{
    return ::write(fd, s, l);
}

size_t String::fwrite(FILE *fp) const
{
    return ::fwrite(s, sizeof(char), l, fp);
}

long String::tonum() const
{
    String copy = *this;

    return copy.number();
}

/* Assumes ASCII.  Change "(*s & 0x5f)" to "tolower(*s)" to eliminate
 * this assumption, with a corresponding speed hit. */
unsigned long String::hash_case() const
{
    unsigned long hashval = 0, g;
    const char *p;
    size_t len;

    /* Algorithm by Peter J. Weinberger. */
    for (p = s, len = l; len; p++, len--) {
	hashval = (hashval << 4) + (*p & 0x5f);
	g = hashval & 0xf0000000;
	if (g) {
	    hashval ^= g >> 24;
	    hashval ^= g;
	}
    }
    return hashval;
}

int String::count_char(char c) const
{
    const char *p;
    int len, count = 0;

    for (p = s, len = l; len; p++, len--) {
	if (*p == c)
	    count++;
    }
    return count;
}

int String::count_words() const
{
    const char *p = s;
    int len = l, count = 0;

    while (len && isspace(*p))
	p++, len--;
    while (len) {
	count++;
	while (len && !isspace(*p))
	    p++, len--;
	while (len && isspace(*p))
	    p++, len--;
    }
    return count;
}

int compare(String s1, String s2)
{
    int val;

    if (s1.l < s2.l) {
	val = memcmp(s1.s, s2.s, s1.l);
	return (val != 0) ? val : s2[s1.l];
    } else {
	val = memcmp(s1.s, s2.s, s2.l);
	return (val != 0) ? val : (s1.l > s2.l) ? s1[s2.l] : 0;
    }
}

int compare_case(String s1, String s2)
{
    size_t i, len;

    len = (s1.l < s2.l) ? s1.l : s2.l;
    for (i = 0; i < len; i++) {
	if (tolower(s1[i]) != tolower(s2[i]))
	    return tolower(s1[i]) - tolower(s2[i]);
    }
    return (i < s1.l) ? s1[i] : (i < s2.l) ? -s2[i] : 0;
}

int String::find(char c) const
{
    const char *p;

    p = (const char *) memchr(s, c, l);
    return (p) ? p - s : -1;
}

int String::find(String str) const
{
    const char *p;

    if (str.len() == 0)
	return 0;
    p = (const char *) memchr(s, *str, l - str.len() + 1);
    while (p) {
	if (memcmp(p, str.s, str.len()) == 0)
	    return p - s;;
	p = (const char *) memchr(p + 1, *str,
				  l - (p + 1 - s) - str.len() + 1);
    }
    return -1;
}

int String::find_case(char c) const
{
    const char *p;

    c = tolower(c);
    for (p = s; p < s + l; p++) {
	if (tolower(*p) == c)
	    return p - s;
    }
    return -1;
}

int String::find_case(String str) const
{
    const char *p, *end;
    int i;
    char first;

    if (str.len() == 0)
	return 0;
    end = s + l - str.len() + 1;
    first = tolower(*str);
    for (p = s; p < end; p++) {
	if (tolower(*p) == first) {
	    for (i = 1; i < str.len(); i++) {
		if (tolower(p[i]) != tolower(str[i]))
		    break;
	    }
	    if (i == str.len())
		return p - s;
	}
    }
    return - 1;
}

int String::rfind(char c) const
{
    const char *p;

    for (p = s + l - 1; p >= s; p--) {
	if (*p == c)
	    return p - s;
    }
    return -1;
}

int String::find_set(String str) const
{
    const char *p;

    for (p = s; p <= s + l; p++) {
	if (memchr(str.s, *p, str.l))
	    return(p - s);
    }
    return -1;
}

void String::trunc_whitespace()
{
    while (l && isspace(s[l - 1]))
	l--;
}

void String::skip_whitespace()
{
    while (l && isspace(*s)) {
	s++;
	l--;
    }
}

void String::strip_whitespace()
{
    trunc_whitespace();
    skip_whitespace();
}

String String::word()
{
    String w;

    *this = get_words(&w, 1);
    return w;
}

long String::number()
{
    long n = 0, sign = 1;

    if (l && *s == '-' || *s == '+') {
	sign = (*s == '-') ? -1 : 1;
	s++, l--;
    }
    while (l && isdigit(*s)) {
	n = n * 10 + (*s - '0');
	s++, l++;
    }
    return sign * n;
}

/* result must contain space for at least n entries.  Returns string
 * beginning at first word after n'th word.  Will return zero-length words
 * at the end of the string if we can't find n words before the end. */
String String::get_words(String *result, int n, String *whitespace) const
{
    String str = *this;

    /* Read words into result. */
    str.skip_whitespace();
    for (; n; n--, result++) {
	result->s = str.s;
	while (str.len() && !isspace(*str))
	    str.shift(1);
	result->l = str.s - result->s;
	str.skip_whitespace();
	if (whitespace) {
	    whitespace->s = result->s + result->l;
	    whitespace->l = str.s - whitespace->s;
	    whitespace++;
	}
    }
    return str;
}

String String::equivalent_substr(String old_backing, String new_backing) const
{
    return new_backing.substr(this->offset_from(old_backing), l);
}

Stored_string::Stored_string()
{
    s = NULL;
}

Stored_string::Stored_string(String str)
{
    s = new char[str.len() + 1];
    str.copy_into(s);
    s[str.len()] = 0;
}

Stored_string::Stored_string(const Stored_string &ss)
{
    s = new char[strlen(ss.s) + 1];
    strcpy(s, ss.s);
}

Stored_string::~Stored_string()
{
    if (s)
	delete[] s;
}

void Stored_string::store(String str)
{
    if (s)
	delete[] s;
    s = new char[str.len() + 1];
    str.copy_into(s);
    s[str.len()] = 0;
}

