/* Stringbuf.cc: Implementation of string buffer methods. */

#include "threads.h"
#include "util.h"

#include "Stringbuf.h"

extern "C" {
#include <string.h>
#include <errno.h>
}

Stringbuf::Stringbuf()
{
    l = 0;
}

Stringbuf::Stringbuf(size_t anticipated) : buf(anticipated)
{
    l = 0;
}

Stringbuf::Stringbuf(String str) : buf(str.len())
{
    str.copy_into(buf.base());
    l = str.len();
}

const char *Stringbuf::cstring()
{
    buf.ensure(l);
    buf[l] = 0;
    return buf.base();
}

void Stringbuf::append(char c)
{
    buf.ensure(l);
    buf[l] = c;
    l++;
}

void Stringbuf::append(String str)
{
    buf.ensure(l + str.len() - 1);
    str.copy_into(buf.base() + l);
    l += str.len();
}

void Stringbuf::store(String str)
{
    clear();
    append(str);
}

void Stringbuf::trunc(size_t len)
{
    l = len;
}

void Stringbuf::shift(size_t start)
{
    l -= start;
    memmove(buf.base(), buf.base() + start, l);
}

void Stringbuf::clear()
{
    l = 0;
}

void Stringbuf::reverse()
{
    char t, *p, *q;

    for (p = buf.base(), q = p + l; p < q; p++, q--) {
	t = *p;
	*p = *q;
	*q = t;
    }
}

void Stringbuf::substitute(String set1, String set2, size_t start)
{
    int pos;
    unsigned int i;

    for (i = start; i < l; i++) {
	pos = set1.find(buf[i]);
	if (pos >= 0)
	    buf[i] = set2[pos];
    }
}

void Stringbuf::format(String fmt, ...)
{
    va_list args;

    va_start(args, fmt);
    vformat(fmt, args);
    va_end(args);
}

void Stringbuf::format_clear(String fmt, ...)
{
    va_list args;

    clear();
    va_start(args, fmt);
    vformat(fmt, args);
    va_end(args);
}

void Stringbuf::vformat(String fmt, va_list args)
{
    String zero_filler = "00000000000";
    Number_buf nbuf;
    String str;
    int len, pos, width;

    while (1) {

	/* Find % or end of string. */
	pos = fmt.find('%');
	if (pos < 0) {
	    /* No more percents; copy rest and stop. */
	    append(fmt);
	    break;
	}
	append(fmt.left(pos));
	fmt.shift(pos + 1);

	width = 0;
	while (fmt.len() && isdigit(*fmt)) {
	    width = width * 10 + *fmt;
	    fmt.shift(1);
	}
	if (!fmt.len())
	    break;

	switch (*fmt) {

	  case '%':
	    append('%');
	    break;

	  case 's':
	    append(va_arg(args, String));
	    break;

	  case 'd':
	    append(long_to_ascii(va_arg(args, int), nbuf, width));
	    break;

	  case 'l':
	    append(long_to_ascii(va_arg(args, long), nbuf, width));
	    break;

	  case 'c':
	    append(va_arg(args, int));
	    break;

	  case 'm':
	    append(String(strerror(errno)));
	    break;

	  case 'N':
	    str = long_to_ascii(va_arg(args, int), nbuf);
	    len = NNTP_NUMBER_LEN - str.len();
	    if (*str == '-')
		append('-');
	    append(zero_filler.left(len));
	    append((*str == '-') ? str.after(1) : str);
	    break;

	  case 'R':
	    append(long_to_radix32(va_arg(args, unsigned long), nbuf));
	    break;

	  default:
	    append('%');
	    append(*fmt);
	}

	fmt.shift(1);
    }
}

