/* Wildcard_list.cc: Matching against a list of wildcard patterns. */

#include "threads.h"
#include "util.h"

#include "Wildcard_list.h"

extern "C" {
#include <assert.h>
#include <string.h>
}

const int DELTA = 4;
const int MIN_POWER = 5;

template<class Value>
Wildcard_list<Value>::Wildcard_list(void (*destructor)(Value value),
				    int nfields)
{
    dest = destructor;
    num_fields = nfields;
}

template<class Value> Wildcard_list<Value>::~Wildcard_list()
{
    int i;

    for (i = 0; i < patterns.len(); i++) {
	delete[] patterns[i].fields;
	if (dest)
	    dest(patterns[i].value);
    }
}

template<class Value>
void Wildcard_list<Value>::add_pattern(const String *fields, Value value)
{
    Pattern<Value> *pattern;
    int i;

    pattern = patterns.next();
    pattern->fields = new Stored_string[num_fields];
    for (i = 0; i < num_fields; i++)
	pattern->fields[i].store(fields[i]);
    pattern->value = value;
}

template<class Value>
Value Wildcard_list<Value>::match(const String *fields,
				  Value default_val) const
{
    return multimatch(&fields, 1, default_val);
}

template<class Value>
Value Wildcard_list<Value>::multimatch(const String *const *lines,
				       int num_lines, Value default_val) const
{
    int i, j, k;
    Stored_string *pattern_fields;

    /* For each pattern, try to match the lines against it. */
    for (i = 0; i < patterns.len(); i++) {
	pattern_fields = patterns[i].fields;
	/* Try to match each line against the pattern. */
	for (j = 0; j < num_lines; j++) {
	    /* For each field, try to match the pattern's field against the
	     * line's field. */
	    for (k = 0; k < num_fields; k++) {
		if (!wildcard_match(pattern_fields[k].str(), lines[j][k]))
		    break;
	    }
	    if (k == num_fields)
		return patterns[i].value;
	}
    }
    return default_val;
}

template<class Value> int Wildcard_list<Value>::num_patterns() const
{
    return patterns.len();
}

Wildcard_decision_list::Wildcard_decision_list(int nfields)
: Wildcard_list<bool>(NULL, nfields)
{
    ;
}

bool Wildcard_decision_list::is_degenerate(bool default_value,
					   bool *always_is) const
{
    int i;

    /* If there are no patterns, the list always returns the default value. */
    if (patterns.len() == 0) {
	*always_is = default_value;
	return 1;
    }

    /* See if we always return the first value. */
    *always_is = patterns[0].value;
    if (fields_always_match(patterns[0].fields))
	return 1;
    for (i = 1; i < patterns.len(); i++) {
	if (!(patterns[i].value == *always_is))
	    return 0;
	if (fields_always_match(patterns[i].fields))
	    return 1;
    }

    /* Degenerate if all the values in the table are the default value. */
    return (*always_is == default_value);
}

Wildcard_decision_list *
Wildcard_decision_list::sublist(const String *const *partial_lines,
				int num_lines, int num_known_fields) const
{
    int i, j, k;
    Stored_string *pattern_fields;
    Wildcard_decision_list *sub =
	new Wildcard_decision_list(num_fields - num_known_fields);

    /* Selectively add patterns which could match one of the partial lines. */
    for (i = 0; i < patterns.len(); i++) {
	pattern_fields = patterns[i].fields;
	/* See if this pattern could match each of the partial lines. */
	for (j = 0; j < num_lines; j++) {
	    /* The pattern must match all the known fields. */
	    for (k = 0; k < num_known_fields; k++) {
		if (!wildcard_match(pattern_fields[k].str(),
				    partial_lines[j][k]))
		    break;
	    }
	    if (k == num_known_fields)
		sub->add_subpattern(pattern_fields + k, patterns[i].value);
	}
    }

    return sub;
}

int
Wildcard_decision_list::fields_always_match(const Stored_string *fields) const
{
    String str;
    int i;

    for (i = 0; i < num_fields; i++) {
	/* A pattern always matches if it is a nonzero sequence of '*'s. */
	str = fields[i].str();
	if (!str.len())
	    return 0;
	while (str.len() && *str == '*')
	    str.shift(1);
	if (str.len())
	    return 0;
    }
    return 1;
}

void Wildcard_decision_list::add_subpattern(const Stored_string *fields,
					    bool value)
{
    Pattern<bool> *pattern;
    int i;

    pattern = patterns.next();
    pattern->fields = new Stored_string[num_fields];
    for (i = 0; i < num_fields; i++)
	pattern->fields[i].store(fields[i].str());
    pattern->value = value;
}

/* Kludge to force instantiations of template. */
#include "Config.h"
template class Wildcard_list<Expire_spec>;
template class Wildcard_list<bool>;
template class Wildcard_list<int>;
template class Wildcard_list<char *>;
template class Wildcard_list<Stored_string>;
template struct Pattern<Expire_spec>;
template struct Pattern<bool>;
template struct Pattern<int>;
template struct Pattern<char *>;
template struct Pattern<Stored_string>;

