/* Wildcard_list.h: Declarations for matching a list of wildcard patterns. */

#ifndef WILDCARD_LIST_H
#define WILDCARD_LIST_H

#include "String.h"
#include "Array.h"

/* This is the most complicated tool in the program.  It's intended to
 * represent a list of wildcard patterns such as are used in the
 * configuration file to decide various values based on parameters.
 * The table consists of a list of patterns; each pattern is a list of
 * wildcard-pattern fields and a value:
 *
 *           Field 1     Field 2     ...     Field m     Value
 *  Line 1
 *  Line 2
 *  ...
 *  Line n
 *
 * To get a value out of the list, you feed in a line consisting of n
 * string fields.  The wildcard list will run through each pattern, from
 * start to end, and return the value of the first pattern whose
 * wildcard-pattern fields match the line's string fields.  The caller of
 * match() is required to pass in a default value to return if no patterns
 * match the input line.
 *
 * The caller can also call multimatch() to try several lines at a time
 * (this is intended to support authentication, where the client may have
 * several credentials of different types, and should be recognized by
 * any one of them).
 *
 * Wildcard_decision_list is a subtype of Wildcard_list for boolean values.
 * For efficiency, Wildcard_decision_list supports taking a sublist: you
 * pass in a group list of lines with only some of the fields specified,
 * and sublist() returns a new Wildcard_decision_list containing the
 * patterns (pared down to their remaining fields) which might match one
 * of the input lines if the remaining fields were filled in.
 *
 * Wildcard_decision_list also supports an is_degenerate() operation, to
 * determine whether the decision list will always return the same value,
 * and what value that is.  (Wildcard patterns are sufficiently simple
 * that we can answer this question with absolute precision.) */

template<class Value> struct Pattern {
    Stored_string *fields;
    Value value;
};

template<class Value> class Wildcard_list {
    void (*dest)(Value value);

    /* Using children to implement template features specific to a certain
     * type is a mild corruption of the inheritance model; thus we have some
     * protected members for the internal representation. */
  protected:
    List<Pattern<Value>, 3, 3> patterns;
    int num_fields;

  public:
    Wildcard_list(void (*destructor)(Value value), int nfields);
    ~Wildcard_list();
    void add_pattern(const String *fields, Value value);
    Value match(const String *fields, Value default_val) const;
    Value multimatch(const String *const *lines, int num_lines,
		     Value default_val) const;
    int num_patterns() const;

    /* Convenience for single-field wildcard lists. */
    Value match1(String field, Value default_val) const
	{ return match(&field, default_val); }
};

class Wildcard_decision_list : public Wildcard_list<bool> {
    int fields_always_match(const Stored_string *fields) const;
    void add_subpattern(const Stored_string *fields, bool value);

  public:
    Wildcard_decision_list(int nfields);

    bool is_degenerate(bool default_value, bool *always_is) const;
    Wildcard_decision_list *sublist(const String *const *partial_lines,
				    int num_lines, int num_known_fields) const;
};

#endif

