/* util.cc: Miscellaneous utilities. */

#include "threads.h"
#include "util.h"
#include "platform.h"

#include "String.h"
#include "Stringbuf.h"

extern "C" {
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>
#include <resolv.h>
#include <errno.h>
}

/* Timezone error, must not be a valid differential. */
#define TZ_ERROR -10000

static String months[] = { "Jan", "Feb", "Mar", "Apr", "May", "Jun",
			   "Jul", "Aug", "Sep", "Oct", "Nov", "Dec" };

static int find_weekday(String s);
static int find_month(String s);
static int parse_timezone(String s);
#ifndef HAVE_TIMEGM
static inline bool leap_year(int y);
static time_t timegm(struct tm *tm);
#endif

int wildcard_match(String pattern, String s)
{
    int pos, lastglob = -1;

    /* Scan the input string updating pattern, backtracking pattern to lastglob
     * whenever we hit a mismatch. */
    while (pos < pattern.len() && s.len()) {
	if (pattern[pos] == '*') {
	    while (pos < pattern.len() && pattern[pos] == '*')
		pos++;
	    if (pos == pattern.len())
		return 1;
	    lastglob = pos - 1;
	    if (pattern[pos] != '?') {
		while (s.len() && tolower(*s) != tolower(pattern[pos]))
		    s.shift(1);
	    }
	} else if (tolower(pattern[pos]) == tolower(*s) ||
		   pattern[pos] == '?') {
	    pos++;
	    s.shift(1);
	} else if (lastglob == -1) {
	    return 0;
	} else {
	    pos = lastglob;
	}
    }

    /* Return true if we exhausted the pattern and string at the same time. */
    return (pos == pattern.len() && !s.len());
}

String long_to_ascii(long num, Number_buf nbuf, int width)
{
    char *p = nbuf + NUMBER_BUF_SIZE - 1;
    int sign = 0;

    if (num < 0) {
	sign = 1;
	num = -num;
    } else if (!num) {
	*p-- = '0';
    }
    while (num) {
	*p-- = num % 10 + '0';
	num /= 10;
    }
    while (nbuf + NUMBER_BUF_SIZE - (p + 1) < width - sign)
	*p-- = '0';
    if (sign)
	*p-- = '-';
    return String(p + 1, nbuf + NUMBER_BUF_SIZE - (p + 1));
}

String long_to_radix32(unsigned long num, Number_buf nbuf)
{
    static char *alphabet = "0123456789abcdefghijklmnopqrstuv";
    char *p = nbuf + NUMBER_BUF_SIZE - 1;

    if (num) {
	while (num) {
	    *p-- = alphabet[num % 32];
	    num /= 32;
	}
    } else {
	*p-- = alphabet[0];
    }
    return String(p + 1, nbuf + NUMBER_BUF_SIZE - (p + 1));
}

int read_line(Stringbuf *line, FILE *fp)
{
    char buf[1024];
    int len;

    /* Read in the first chunk, and handle a trivial cases. */
    if (!fgets(buf, sizeof(buf), fp) || !*buf)
	return -1;
    if (*buf == '\n')
	return 0;

    /* Until we hit a newline, keep reading in text. */
    while (1) {
	len = strlen(buf);
	if (buf[len - 1] == '\n') {
	    line->append(String(buf, len - ((buf[len - 2] == '\r') ? 2 : 1)));
	    return 0;
	}
	line->append(String(buf, len));
	if (!fgets(buf, sizeof(buf), fp))
	    return -1;
    }
}

int read_line_clear(Stringbuf *line, FILE *fp)
{
    line->clear();
    return read_line(line, fp);
}

int tcp_connect(String name, String service, unsigned short fallback,
		Stringbuf *error)
{
    Stringbuf namebuf(name), servicebuf(service);
    struct hostent_answer hanswer;
    struct servent_answer sanswer;
    struct hostent *hostent;
    struct servent *servent;
    struct sockaddr_in sin;
    int fd;

    sin.sin_family = AF_INET;

    hostent = gethostbyname_r(namebuf.cstring(), &hanswer);
    if (!hostent) {
	error->format("Can't resolve host %s: %s", name,
		      String(hstrerror(h_errno)));
	return -1;
    }
    memcpy(&sin.sin_addr, hostent->h_addr, sizeof(struct in_addr));

    servent = getservbyname_r(servicebuf.cstring(), "tcp", &sanswer);
    sin.sin_port = (servent) ? servent->s_port : fallback;

    fd = socket(AF_INET, SOCK_STREAM, 0);
    if (fd < 0) {
	error->format("Can't create socket for connection to %s: %s",
		      name, String(strerror(errno)));
	return -1;
    }

    if (connect(fd, (struct sockaddr *) &sin, sizeof(sin)) < 0) {
	error->format("Can't connect to %s: %s", name,
		      String(strerror(errno)));
	close(fd);
	return -1;
    }

    return fd;
}

unsigned short service_port(String service, unsigned short fallback)
{
    Stringbuf service_buf(service);
    struct servent *servent = NULL;
    struct servent_answer buf;

    if (service.len())
	servent = getservbyname_r(service_buf.cstring(), "tcp", &buf);
    return (servent) ? servent->s_port : fallback;
}

int blocksize(int fd)
{
    const int guess = 4096;

#ifdef HAVE_ST_BLKSIZE
    struct stat statbuf;

    if (fstat(fd, &statbuf) == -1) {
	warn("Can't fstat fd %d: %m", fd);
	return guess;
    }
    return statbuf.st_blksize;
#else
    return guess;
#endif
}

/* Compare two Message-IDs.  A robust comparison would require more
 * parsing (see RFCs 822 and 1036), but this works fine in news
 * articles. */
int message_id_equal(String id1, String id2)
{
    if (id1.len() != id2.len())
	return 0;
    while (id1.len() && *id1 == *id2 && *id1 != '@')
	id1.shift(1), id2.shift(1);
    if (id1.len() && *id1 == '@') {
	while (id1.len() && tolower(*id1) == tolower(*id2))
	    id1.shift(1), id2.shift(1);
    }
    return (id1.len() == 0);
}

/* Verify that a Message-ID is valid.  A full check for RFC 822 compliance
 * would require more parsing, but this routine only checks whether the
 * Message-ID is composed of a string of non-blank characters other than
 * '>' between '<' and '>', and that it contains at least one '@'.  (RFC
 * 822 allows blanks in Message-IDs, but RFC 1036 doesn't.)  I don't see
 * any need to verify full RFC822 compliance. */
int verify_message_id(String id)
{
    if (id.len() == 0 || *id != '<')
	return 0;
    id.shift(1);
    while (id.len() && *id != '>' && !isspace(*id))
	id.shift(1);
    return (id.len() == 1 && *id == '>') ? 0 : -1;
}

/* Parse a date according to RFC 822, allowing more than two characters
 * for the year.  For compatibility with gnus in bnews mode, assume local
 * time zone if time zone is omitted.  Return -1 on rejection.  Grammar is:
 *
 * date = [ weekday "," ] D[D] month Y* HH:MM[:SS] [zone]
 * zone = ("+" / "-")HHMM / Military code / Supported zone name
 *
 * Example: Fri, 23 Sep 1994 14:42:50 GMT */
time_t parse_header_date(String date)
{
    struct tm tm;
    int differential;

    /* Look for and discard a leading weekday and comma. */
    date.skip_whitespace();
    if (date.len() >= 4 && find_weekday(date.left(3)) >= 0 && date[3] == ',') {
	date.shift(4);
	date.skip_whitespace();
    }

    /* One or two digits giving the day of the month. */
    if (date.len() == 0 || !isdigit(*date))
	return -1;
    tm.tm_mday = *date - '0';
    date.shift(1);
    if (date.len() != 0 && isdigit(*date)) {
	tm.tm_mday = (tm.tm_mday * 10) + (*date - '0');
	date.shift(1);
    }
    date.skip_whitespace();

    /* Name of month. */
    tm.tm_mon = (date.len() >= 3) ? find_month(date.left(3)) : -1;
    if (tm.tm_mon < 0)
	return -1;
    date.shift(3);
    date.skip_whitespace();

    /* One or more digits giving the year. */
    if (date.len() == 0 || !isdigit(*date))
	return -1;
    tm.tm_year = date.number();
    tm.tm_year = (tm.tm_year < 1000) ? tm.tm_year + 1900 : tm.tm_year;
    if (tm.tm_year < 1900)
	return -1;
    date.skip_whitespace();

    /* HH:MM */
    if (date.len() < 5 || !isdigit(*date) || !isdigit(date[1]) ||
	date[2] != ':' || !isdigit(date[3]) || !isdigit(date[4]))
	return -1;
    tm.tm_hour = (*date - '0') * 10 + (date[1] - '0');
    tm.tm_min = (date[3] - '0') * 10 + (date[4] - '0');
    tm.tm_sec = 0;
    date.shift(5);

    /* Optional :SS */
    if (date.len() >= 3 && *date == ':' && isdigit(date[1]) &&
	isdigit(date[2])) {
	tm.tm_sec = (date[1] - '0') * 10 + (date[2] - '0');
	date.shift(3);
    }
    date.skip_whitespace();

    if (date.len()) {
	/* Get time zone differential in minutes. */
	differential = parse_timezone(date);
	if (differential == TZ_ERROR)
	    return -1;
	return timegm(&tm) - (differential * 60);
    } else {
	return mktime(&tm);
    }
}

static int find_weekday(String s)
{
    static String weekdays[] = { "Mon", "Tue", "Wed", "Thu",
				 "Fri", "Sat", "Sun" };
    int i;

    for (i = 0; i < 7; i++) {
	if (compare_case(weekdays[i], s) == 0)
	    return i;
    }
    return -1;
}

static int find_month(String s)
{
    int i;

    for (i = 0; i < 12; i++) {
	if (compare_case(months[i], s) == 0)
	    return i;
    }
    return -1;
}

/* Returns differential in minutes, or TZ_ERROR for error. */
static int parse_timezone(String s)
{
    static struct {
	String name;
	int differential;
    } named[] = {
	{ String("UT"),   0 },		{ String("GMT"),  0 },
	{ String("EST"), -5 * 60 },	{ String("EDT"), -4 * 60 },
	{ String("CST"), -6 * 60 },	{ String("CDT"), -5 * 60 },
	{ String("MST"), -7 * 60 },	{ String("MDT"), -6 * 60 },
	{ String("PST"), -8 * 60 },	{ String("PDT"), -7 * 60 }
    };
    const int NUM_NAMED_ZONES = sizeof(named) / sizeof(*named);
    static int milzones[] = { -1, -2, -3, -4, -5, -6, -7, -8, -9, 666,
			      -10, -11, -12, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10,
			      11, 12, 0 };
    int i;

    if (s.len() && (*s == '+' || *s == '-'))
	return s.tonum();

    if (s.len() >= 3) {
	s.trunc(3);
	for (i = 0; i < NUM_NAMED_ZONES; i++) {
	    if (compare_case(named[i].name, s) == 0)
		return named[i].differential;
	}
    }

    if (s.len() && isalpha(*s) && *s != 'j' && *s != 'J' &&
	!(s.len() >= 2 && isalpha(s[1])))
	return milzones[tolower(*s) - 'a'] * 60;

    return TZ_ERROR;
}

time_t parse_nntp_date(String date, String timest, String gmt)
{
    struct tm tm, current;
    time_t t;
    int century, i;

    /* Verify that date and time are each six digits, and that gmt is
     * either empty or "GMT". */
    if (date.len() != 6 || timest.len() != 6)
	return -1;
    for (i = 0; i < 6; i++) {
	if (!isdigit(date[i]))
	    return -1;
    }
    for (i = 0; i < 6; i++) {
	if (!isdigit(timest[i]))
	    return -1;
    }
    if (gmt.len() && compare_case(gmt, "gmt") != 0)
	return -1;

    /* Convert digits into numbers, reject invalid numbers. */
    tm.tm_year = (date[0] - '0') * 10 + (date[1] - '0');
    tm.tm_mon =  (date[2] - '0') * 10 + (date[3] - '0') - 1;
    tm.tm_mday = (date[4] - '0') * 10 + (date[5] - '0');
    tm.tm_hour = (timest[0] - '0') * 10 + (timest[1] - '0');
    tm.tm_min =  (timest[2] - '0') * 10 + (timest[3] - '0');
    tm.tm_sec =  (timest[4] - '0') * 10 + (timest[5] - '0');
    tm.tm_isdst = -1;
    if (tm.tm_mon > 11 || tm.tm_mday > 31 || tm.tm_hour > 23 ||
	tm.tm_min > 59 || tm.tm_sec > 59 || tm.tm_mon < 0 || tm.tm_mday < 1)
	return -1;

    /* Get current time to find century. */
    time(&t);
    if (localtime_r(&t, &current)) {
	century = (current.tm_year + 1900) / 100 * 100;
	if (current.tm_year - tm.tm_year >= 50)
	    century += 100;
	else if (tm.tm_year - current.tm_year >= 50)
	    century -= 100;
	tm.tm_year += century - 1900;
    }

    return gmt.len() ? timegm(&tm) : mktime(&tm);
}

#ifndef HAVE_TIMEGM
static inline bool leap_year(int y)
{
    return (y % 4 == 0 && y % 100 != 0) || y % 400 == 0;
}

static time_t timegm(struct tm *tm)
{
    static int mstart[] = { 0,  31,  59,  90, 120, 151,
			    181, 212, 243, 273, 304, 334, 365 };
    time_t t;

    /* Convert tm fields into seconds since midnight, Jan 1, 1970 GMT. */
    t = (tm->tm_year - 70) * (time_t)(365 * 86400);
    t += ((tm->tm_year - 69) / 4) * 86400;
    t += (mstart[tm->tm_mon] + tm->tm_mday - 1) * 86400;
    t += tm->tm_hour * 3600 + tm->tm_min * 60 + tm->tm_sec;
    if (leap_year(tm->tm_year) && tm->tm_mon > 1)
	t += 86400;
    return t;
}
#endif

int generate_current_header_date(Stringbuf *buf)
{
    char tmp[40];
    time_t now;
    struct tm tm;

    time(&now);
    if (gmtime_r(&now, &tm) == NULL)
	return -1;
    buf->format("%d %s %d ", tm.tm_mday, months[tm.tm_mon], 1900 + tm.tm_year);
    sprintf(tmp, "%02d:%02d:%02d GMT", tm.tm_hour, tm.tm_min, tm.tm_sec);
    buf->append(tmp);
    return 0;
}

String get_delimited_element(String *str, char delimiter)
{
    int pos;
    String name;

    pos = str->find(delimiter);
    pos = (pos == -1) ? str->len() : pos;
    name = str->left(pos);
    str->shift(pos + 1);
    name.strip_whitespace();
    str->skip_whitespace();
    return name;
}

String get_group_name(String *groups)
{
    return get_delimited_element(groups, ',');
}

void make_directories_in_path(String name)
{
    Stringbuf partial;
    int slash;

    while ((slash = name.find('/')) >= 0) {
	partial.append(name.left(slash));
	mkdir(partial.cstring(), S_IRWXU | S_IRWXG | S_IROTH);
	partial.append('/');
	name.shift(slash + 1);
    }
}

int check_flags(String flags)
{
    if (!flags.len() || !strchr("xjmyn=", *flags))
	return -1;
    if (flags.len() != 1 && !(*flags == 'm' || *flags == '='))
	return -1;
    return 0;
}

